/* C++ includes */
#include <stream.h>
#include <String.h>

/* C includes */
extern "C"
{
#include "options.h"		// Octtools
#include "utility.h"		// Octtools
#include <signal.h>
}

/* Local includes */
#include "vovlib.h"
#include "trace.h"
#include "place.h"
#include "retrace.h"
#include "voverror.h"
#include "dialog.h"
#include "path.h"
#include "ddsplus.h"
#include "assist.h"
#include "asshelp.h"

extern int unix_exists( char* );

// GLOBAL VARIABLES:
int VAAutomatic = 0;			// automatic processing flag.
int VAdebug = 0;

#define VA_MAX_VARIABLES 5

int VA_getExampleTraces( String& filename, String traceNames[],  int arrayLength )
{
    istream traceListFile( (char*)filename, io_readonly, a_useonly );
    int numberOfTraces = 0;
    String traceName;
    while ( traceListFile.readable() ) {
	traceListFile  >> traceName;
	if ( traceListFile.eof() ) break;
	traceNames[numberOfTraces] = VOV_MODEL_TRACE_ABSOLUTE_PATH_DEFAULT + traceName;
	numberOfTraces++;
    }
    if ( !numberOfTraces ) {
	vovFatalError( "No example traces in file " + filename );
    } 
    return numberOfTraces;
}

void VA_getScriptSet( VovTrace& trace, VovNodeSet& scriptset )
{
    int source_mask = VOV_SOURCE_MASK | VOV_PLACE_MASK;
    int sink_mask   = VOV_SINK_MASK | VOV_PLACE_MASK;
    VovSelectRule selSources( source_mask, VOV_ALL_PLACE_TYPE_MASK, VOV_ALL_STATUS_MASK, SELECT_ALL );
    VovSelectRule selSinks( sink_mask, VOV_ALL_PLACE_TYPE_MASK, VOV_ALL_STATUS_MASK, SELECT_ALL );
    VovNodeSet sources( "SOURCES", selSources );
    VovNodeSet sinks( "SINKS", selSinks );
    VOV_ASSERT( trace.findOrCreate( sources ));
    VOV_ASSERT( trace.findOrCreate( sinks ) );
    VOV_ASSERT( trace.findOrCreate( scriptset ) );
    VovSelectRule sel(VOV_PLACE_MASK | VOV_TRANSITION_MASK, VOV_ALL_PLACE_TYPE_MASK,
		      VOV_ALL_STATUS_MASK, SELECT_ALL );
    VOV_ASSERT( trace.getIntermediateNodes( sources, sinks, sel, scriptset ) );
    return;			// VOID.
}

String VA_getRootName( VovPlace& place )
{
    String name = place.getName();
    switch ( place.getPlaceType() ) {
    case VOV_UNIX_FILE:
	name = name.after( "/" , -1 );
	name = name.before( ".", -1 );
	return name;
    case VOV_OCT_FACET:
	{
	    extern octStatus vovUnpackFacetName( VovObject* facet, char* name);
	    VovObject facet( OCT_FACET );
	    vovUnpackFacetName( &facet,  (char*)name );
	    String s = facet.contents.facet.cell;
	    s = s.after( "/", -1 );
	    return s;
	}
	break;
    default:
	return 0;
    }
}

String VA_getDirectory( VovPlace& place ) 
{
    String name = place.getName();
    switch ( place.getPlaceType() ) {
    case VOV_UNIX_FILE:
    case VOV_OCT_FACET:
	name = name.before( "/" , -1 );
	return name;
    default:
	return 0;
    }
}

VovStatus VA_guessTraceRoot( VovTrace& trace, String& root )
{
    root = "<old root could not be determined>";

    int source_mask = VOV_SOURCE_MASK | VOV_PLACE_MASK;
    VovSelectRule selSources( source_mask, VOV_ALL_PLACE_TYPE_MASK, VOV_ALL_STATUS_MASK, SELECT_ALL );
    VovNodeSet sources( "SOURCES", selSources );
    VOV_ASSERT( trace.findOrCreate( sources ));
    
    VovGenerator genPlace;
    VOV_ASSERT( trace.initGenContents( sources, VOV_PLACE_OCTMASK, genPlace ) ); 
    
    VovPlace examplePlace;
    while ( trace.generate( genPlace, examplePlace ) == VOV_OK ) {
	VOV_ASSERT( trace.getById( examplePlace ) );

	VovPlaceType type = examplePlace.getPlaceType();
	if ( type != VOV_UNIX_FILE &&  type != VOV_OCT_FACET ) {
	    continue;
	} else {
	    root = VA_getRootName( examplePlace );
	    break;
	}
    }
    // vovMessage( "Root seems to be " + root );
    return VOV_OK;
}


// -------------------------------------------------------------------
// -------------------------------------------------------------------
// -------------------------------------------------------------------
// -------------------------------------------------------------------
// -------------------------------------------------------------------
// -------------------------------------------------------------------

typedef struct {
    VovTrace* mainTracePtr;
    VovTrace* exampleTracePtr;
    VovNodeSet* scriptSetPtr;
    VovDialogEditText* scriptDlgPtr;
    VovDialogEditText** newDlgPtr;
    VovDialogEditText** oldDlgPtr;
    ddsHandle setNameDlgHandle; 	// Still not fully used.
    int scriptflags;
} VovLocalAssistData;

void vov_assist_update( VovLocalAssistData& data )
{
    VovDialogEditText& scriptDlg = *data.scriptDlgPtr;
    VovTrace& exampleTrace = *data.exampleTracePtr;
    VovNodeSet& scriptSet = *data.scriptSetPtr;

    String script;
    VOV_ASSERT( exampleTrace.prepareScript( scriptSet, data.scriptflags, script ) );
    scriptDlg.modifyText( script );

}


void vov_assist_toggle_notes_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalAssistData& data = *(VovLocalAssistData*)ctrl.getData();
    data.scriptflags ^= VOV_ANNOTATION_MASK;
    vov_assist_update( data );
}

void vov_assist_proceed_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalAssistData& data = *(VovLocalAssistData*)ctrl.getData();

    VovTrace& mainTrace = *data.mainTracePtr;
    VovTrace& exampleTrace = *data.exampleTracePtr;
    VovDialogEditText** oldDlgPtrArray = data.oldDlgPtr;
    VovDialogEditText** newDlgPtrArray = data.newDlgPtr;
    VovDialogEditText setNameDlg( data.setNameDlgHandle );

    VovVarSet varSet;
    for ( int i = 0 ; i < VA_MAX_VARIABLES ; i++ ) {
	String oldrootname = oldDlgPtrArray[i]->getTextItem( );
	if ( oldrootname == 0 ) {
	    continue;
	}
	String rootname =    newDlgPtrArray[i]->getTextItem( );
	if ( rootname == 0 ) {
	    vovShortMessage( "ERROR", "NEWROOT cannot be empty" );
	    continue;
	}

	VovVar var_root( form("ROOT-%d",i), oldrootname, rootname );
	varSet.add( var_root );
	oldDlgPtrArray[i]->modifyText( rootname ); // Update oldroot, so it is ready for next time.
    }

	
    
    // var_root.newValue() = rootname;
    String setName = setNameDlg.getTextItem() ;
    VovNodeSet alltransitions( setName );
    if ( exampleTrace.find( alltransitions ) != VOV_OK ) {
	vovConfirm( "ERROR", "set "+setName+ " does not exist.", "Continue", 0 );
	return;
    }
    if ( VA_prepareTrace( mainTrace, exampleTrace, alltransitions, varSet ) == VOV_OK ) {
	vovLongMessage( "VOV ASSISTANT: WHAT NEXT?", VHA_WHAT_NEXT, 10, 80 );
    } else {
	vovConfirm( "WARNING", "\nAssistance aborted. You can continue.\n", "Continue", 0 );
    }
    vov_assist_update( data );
}

void vov_assist_click_me_ccb( ddsHandle item )
{
    vovLongMessage( "ADVICE FOR THE NOVICE", VHA_NOVICE, 20, 70 );
}

VovStatus VA_editTrace( VovTrace& mainTrace, VovTrace& exampleTrace, String& setName, String& rootname ) 
{
    VovNodeSet scriptSet( setName );
    if ( exampleTrace.find( scriptSet ) != VOV_OK ) {
	vovFatalError( "Set not found:" + setName );
    }
    // VA_getScriptSet( exampleTrace, scriptSet );
    String oldrootname ;
    VOV_ASSERT( VA_guessTraceRoot( exampleTrace, oldrootname ) );
    if ( rootname == 0 ) rootname = oldrootname;

    
    VovDialog  top_dialog( "VOV ASSISTANT", DDS_VERTICAL, 1000 );
    
    VovDialogEditText scriptDlg(top_dialog, "SCRIPT", "", 15, 80 );
    VovDialogEditText setNameDlg( top_dialog, "SET NAME", setName, 1, 60 );

    VovDialog varsDialog( top_dialog, "VARIABLE SUBSTITUTIONS", DDS_HORIZONTAL, 2 );
    VovDialogEditText* oldDlgPtrArray[VA_MAX_VARIABLES];
    VovDialogEditText* newDlgPtrArray[VA_MAX_VARIABLES];
    for ( int i = 0 ; i < VA_MAX_VARIABLES; i++ ) {
	String name = (i == 0 ) ? rootname : "";
	oldDlgPtrArray[i] = new VovDialogEditText( varsDialog, "OLD", name, 1 , 25 );
	newDlgPtrArray[i] = new VovDialogEditText( varsDialog, "NEW", name, 1 , 25 );
    }
	

    VovLocalAssistData data;
    data.mainTracePtr = &mainTrace;
    data.exampleTracePtr = &exampleTrace;
    data.scriptSetPtr = &scriptSet;
    data.scriptDlgPtr = &scriptDlg;
    data.newDlgPtr = newDlgPtrArray;
    data.oldDlgPtr = oldDlgPtrArray;
    data.setNameDlgHandle = setNameDlg.handle();

    data.scriptflags = VOV_TRANSITION_MASK ;

    top_dialog.addControl("CLICK ME", vov_assist_click_me_ccb, 0 );
    top_dialog.addControl("Toggle notes", vov_assist_toggle_notes_ccb, (ddsData)&data );
    top_dialog.addControl("Proceed", vov_assist_proceed_ccb, (ddsData)&data );
	
    top_dialog.addHelp( VHA_EDIT );



    vov_assist_update( data );
    top_dialog.postModed();

    return VOV_OK;
}


String shortTraceName( String name )
{
    if ( name.contains("/") ) name = name.after( "/", -1 );
    return name;
}


// -------------------------------------------------------------
// -------------------------------------------------------------
// -------------------------------------------------------------
// -------------------------------------------------------------
// -------------------------------------------------------------
// -------------------------------------------------------------

typedef struct {
    String*   traceNameArray;
    VovTrace* mainTracePtr;
    VovTrace* exampleTraceArray;
    VovTrace* selectedTracePtr;
    ddsHandle infoDlgHandle;
    String     setName;
    int index;
} VovLocalTraceSelectData;


String VA_getTraceInfo( String& traceName )
{
    String info = "No info for this trace";

    octObject facet;
    String actualTraceName = traceName + ".page.0" ;
    if ( ohOpenFacet( &facet, (char*)actualTraceName, "symbolic", "interface", "r" ) < OCT_OK ) {
	info = form( "Cannot open %s\n" , ohFormatName( &facet ) );
    } else {
	VovProp prop( "TRACE_INFO", OCT_STRING );
	if ( octGetByName( &facet, &prop ) == OCT_OK ) {
	    info = prop.string();
	}
	octCloseFacet( &facet );
    }
    return info;
}


void vov_assist_trace_index_ccb( ddsHandle item, int toggled )
{
    VovDialogList list( item );
    VovLocalTraceSelectData& data = *(VovLocalTraceSelectData*)list.getData();
    data.index = (int)list.getWhichOne();

    VovDialogEditText infoDlg( data.infoDlgHandle );

    String traceInfo = "\nPlease select a trace.\n";
    if ( list.isSelected( toggled ) ) {
	traceInfo = VA_getTraceInfo( data.traceNameArray[data.index] );
    } 
    infoDlg.modifyText( traceInfo );

    // vovMessage( form( "index is %d", data.index ) );
}

void vov_assist_select_ccb( ddsHandle item )
{
    VovDialogControl button( item );
    VovLocalTraceSelectData& data = *(VovLocalTraceSelectData*)button.getData();

    VovTrace* exampleTraceArray = data.exampleTraceArray;
    String*  traceNames = data.traceNameArray;
    String ctrl = button.name();

    
    VovTrace& trace = exampleTraceArray[data.index];
    if ( trace.disabled() ) {
	trace.open( traceNames[data.index], "<no host>", "r", VOV_TRACE_TYPE_LOCAL );
    }

    if ( ctrl == "Show trace" ) {
	String rpcApp = form( "%s/bin/vov_rpc", "$OCTTOOLS" );
	String command = form( "vem -F %s.page.0:symbolic -R %s >& /dev/null &",
			      (char*)traceNames[data.index], (char*)rpcApp );
	util_csystem( (char*)command );
    } else if ( ctrl == "Assist" ) {
	// data.selectedTracePtr = &exampleTraceArray[data.index];
	VovTrace& mainTrace = *data.mainTracePtr;
	VAAutomatic = 0;	// Reset automatic flag.
	VA_editTrace( mainTrace, exampleTraceArray[data.index], data.setName, 0 );
    } else {
	vovMessage( "What button did you click?" );
    }
}

void VA_chooseExampleTrace( String& fileName, VovTrace& mainTrace, String& setName )
{
    const int MAX_EXAMPLE_TRACES = 40; 		// Sort of arbitrary.
    static String* traceNames = 0; 		// 
    static VovTrace* exampleTraceArray = 0;	// Static because if automatic g++ screws up upon destruction.
    
    // If necessary, allocate space for all the example traces.
    if ( traceNames == 0 ) {
	traceNames = new String[MAX_EXAMPLE_TRACES]; 		// Names of the example traces.
	exampleTraceArray = new VovTrace[MAX_EXAMPLE_TRACES];	// Space for the traces themselves.
    }
    
    int numberOfTraces = 0;
    if ( numberOfTraces == 0 ) {
	numberOfTraces = VA_getExampleTraces( fileName, traceNames, MAX_EXAMPLE_TRACES );
    }


    VovLocalTraceSelectData data;
    data.mainTracePtr = &mainTrace;
    data.setName = setName;
    data.traceNameArray = traceNames;
    data.exampleTraceArray = exampleTraceArray;
    data.index = 0;
    data.selectedTracePtr = 0;

    VovDialog top_dialog("VOV ASSISTANT: CHOOSE EXAMPLE TRACE", DDS_VERTICAL, MAX_EXAMPLE_TRACES );

    VovDialogList chooseTraces( top_dialog, DDS_TRUE );

    for ( int i = 0 ; i < numberOfTraces; i++ ) {
	String name = shortTraceName( traceNames[i] );
	chooseTraces.addItem( name, i, 0 ); // Select none.
    }

    chooseTraces.endOfList( vov_assist_trace_index_ccb , (ddsData)&data, 6, 0 );

    VovDialogEditText infoDlg( top_dialog, "INFO", "Please select an example trace.", 4, 45 );
    data.infoDlgHandle = infoDlg.handle();

    top_dialog.addControl( "CLICK ME", vov_assist_click_me_ccb, 0 );
    top_dialog.addControl( "Show trace", vov_assist_select_ccb, (ddsData)&data );
    top_dialog.addControl( "Assist", vov_assist_select_ccb, (ddsData)&data );
    top_dialog.addHelp( VHA_MAIN );
    

    top_dialog.postModed();
    
}


optionStruct optionList[] = {
    { "i", 0, 		"Interactive mode" },
    { "d", 0,		"Print debugging information" },
    { "p", "tracename", "Name of the example trace" },
    { "O", "name",	"Old string in example trace" },
    { "N", "name",	"New string in example trace" },
    { "f", "filename",	"File containing list of example traces" },
    { "H", 0,       	"Output hardcopy of trace and annotations (must use -p)" },
    { 0, 0, 0 }
};

String VA_traceListFileName()
{
    String fileName = "vov.traces";
    if ( ! unix_exists( (char*)fileName ) ) {
	fileName =  VOV_MODEL_TRACE_ABSOLUTE_PATH_DEFAULT + fileName;
    }
    return fileName;
 
}
main(int argc, char* argv[])
{
    extern int optind;
    extern char* optarg;

    vovRegisterMainArgs( argc, argv );

    
    String fileName = VA_traceListFileName(); // File containing list of example traces.
    String exampleTraceName;		      // Name of the selected example trace.
    String setName = ";;;NODES;;;";	      // Name of the set to consider in example trace.
    String newroot, oldroot;		      // Used for variable substitution.
    
    // Flags.
    int interactive = 0;	// Interactive operation of vov_assist.
    int hardcopy = 0;		// Produce a hardcopy of example trace.

    int option = 0;
    while (( option = optGetOption( argc, argv ) ) != EOF ) {
	switch (option) {
	case 'H':
	    hardcopy = 1;
	    break;
	case 'i':
	    interactive = 1;
	    break;
	case 's':
	    setName = optarg;
	    break;
	case 'd':
	    VAdebug = 1;
	    break;
	case 'p':
	    exampleTraceName = optarg;
	    break;
	case 'O': 
	    oldroot = optarg;
	    break;
	case 'N':
	    newroot = optarg;
	    break;
	case 'f':
	    if ( unix_exists( optarg ) ) {
		fileName = optarg;
	    } else {
		vovFatalError( form( "File %s does not exists", optarg ) );
	    }
	    break;
	default:
	    break;
	}
    }
    if ( argc == 1 ) {
	optUsage();
    }


    if ( hardcopy ) {
	if ( exampleTraceName == 0 ) {
	    vovFatalError("Example trace not specified: please use the -p option." );
	}
	VA_hardcopy( exampleTraceName, setName );
	exit( 0 );
    }
    
    
    String project = utcc_getenv("VOV_PROJECT_NAME");
    String host =  utcc_getenv("VOV_HOST_NAME");
    VovTrace mainTrace( project, host, "a", VOV_TRACE_TYPE_CLIENT );
    
    if ( exampleTraceName != 0 && newroot != 0 ) {
	VovTrace exampleTrace;
	String fullTraceName = form( "%s%s", VOV_MODEL_TRACE_ABSOLUTE_PATH_DEFAULT , (char*)exampleTraceName);
	exampleTrace.open(  fullTraceName, "<no host>", "r", VOV_TRACE_TYPE_LOCAL  );
	VovNodeSet set( setName );
	if ( exampleTrace.find( set ) != VOV_OK ) {
	    vovFatalError( "Set " + setName + " not found in " + exampleTrace.getProjectName() );
	}
	if ( oldroot == 0 ) {
	    VOV_ASSERT( VA_guessTraceRoot( exampleTrace, oldroot ) );
	}
	VovVar var_root( "ROOT", oldroot, newroot );
	VovVarSet varSet;
	varSet.add( var_root );
	vovMessage( "Automatic assistance" );
	VAAutomatic = 1;
	VA_prepareTrace( mainTrace, exampleTrace, set, varSet );
    } else if ( interactive ) {
	fileName = util_file_expand( (char*)fileName );
	VA_chooseExampleTrace( fileName, mainTrace, setName );
    }
    exit( 0 );
}


