/* C++ includes */
#include "port_internal.h"

#include "options.h"		// Octtools

/* C includes */
#include <signal.h>		// System
#include <sys/types.h>
#include <sys/stat.h>


/* Local includes */
#include "trace.h"
#include "transition.h"
#include "place.h"
#include "user.h"
#include "voverror.h"
#include "path.h"
#include "placelist.h"
#include "dialog.h"
#include "ddsplus.h"
#include "vovlib.h"
#include "version.h"


extern String vov_normalizeString( const String& line );
extern String vov_stdoutRedirect( String& commandline );


optionStruct optionList[] = {
    { "I", 0,		"Interactive transaction (do not redirect stdin, stderr,stdout)" },
    { "i", "file/facet","Input" },
    { "o", "file/facet","Output" },
    { "a", "file/facet","Input and output (for tools running 'in place')" },
    { "P", "rootname",	"Set root name for program options (OPT)" },
    { "p", "option"    ,"Program option" },
    { "c", "\"command\"","Top level command line" },
    { "x", "executable","Input executable" },
    { "y", "executable","Output executable" },
    { "X", 0,		"Record exit status" },
    { "B", "name",	"Boolean input" },
    { "r", "file",	"Rename stderr" },
    { "R", "file",	"Rename stdout" },
    { "s", "file",	"Rename stdin" }, 
    { "S", "status",    "Legal exit status string (ex: \"0\", \"0-5\", \"0 1 2 4-9\")" },
    { "A", "string",    "Affinity of transition (\"none\" for no affinity)" },
    { "d", 0,		"Debug communication with server"},
    { "V", 0,		"Print version number"},
    { "n", 0,		"Do not actually run the transaction"},
/*     { OPT_RARG , "<command line>", "Complete command line" }, */
    { 0, 0, 0 }
};


void printIOplace(  char* IO, VovPlace& place )
{
    String type = vov_getPlaceTypeString( place.getPlaceType() );
    String shortName = place.getName();
    String cwd = utcc_getwd() + "/";
    if ( shortName.contains( cwd ) ) {
	shortName = shortName.after( cwd );
    }
    cout.form( "%2s:%6s: %s\n", IO, (char*)type, (char*)shortName );
}

//  cell:view	-> VOV_OCT_FACET	( contains a colon ':')
//  file	-> VOV_UNIX_FILE	( everything else )
VovPlaceType guessType( String& name )
{
    if (name.contains(":")) {
	return VOV_OCT_FACET;
    } else {
	return VOV_UNIX_FILE;
    }
}


//
// Given the name of an executable find its full path using `which`.
//
String findExecutable( String& exec )
{
    if ( exec.contains( "/" ) ) { 			// No use of the path mechanism.
	return exec;					// 
    }							// 
    String executable = utcc_popen( "which " + exec  );	// 
    if ( executable.contains( '\n' ) ) { 		// Eliminate final new line.
	executable = executable.before( "\n" );		// 
    }							// 
    if ( executable[0] != '/' && executable[0] != '.') {
	// Executable for %s not found: please check your path.", (char*)exec ) 
	return exec; // Return any way. It could be a .o or a .a file.
    }
    return executable;
}

//
// Redirect stdin, stdout and stderr as appropriate
//
String createCommand( String& cmd, String& in, String& out, String& err )
{
    String result = cmd;
    if ( in  != 0 ) result =  cmd + " < " + in;
    
    int mask = 0;
    mask |=   ( out != 0 ) ;
    mask |= ( ( err != 0 && err != out ) << 1 );
    switch ( mask ) {
    case 00:
	break;
    case 01:
	result = result  + " >& " + out;
	break;
    case 02:
	result = result + " >& " + err;
	break;
    case 03:
	result = "(" + result + " > " + out + ") >& " + err;
	break;
    }
    return result;
}

SIGNAL_FN cleanExit()
{
    vovMessage(" Signal caught " );
}
void shellSignals()
{
    struct sigvec sig, osig;
    sig.sv_handler = cleanExit;
    sig.sv_mask = 0;
    SIG_FLAGS(sig) = 0;
    sigvec( SIGHUP, &sig, &osig );
    return;
}


VovStatus agreeOnPlace( VovTrace& trace, VovPlace& p )
{
    int exist;
    int mode;
    Date ts;				// Timestamp.
    p.info( &exist, &mode, &ts, 0 );

    if ( exist && p.getStatus() ==  VOV_NONEXISTING_NODE ) {
	return VOV_ERROR;
    }
    return VOV_OK;
}



main(int argc, char* argv[])
{
    extern int optind;
    extern char* optarg;
    int exitStatus = 0;		// Do not save exit status.
    int dontrun = 0;		// If 1, do notexec transaction.
    
    int interactive = 0;	// 
    String stdinname;		// Null by default.
    String stdoutname;		// 
    String stderrname;		// 
    String executable;		// The name of the executable.
    String capsulecommand;	// Capsuled Command line.
    String actualcommand;	// Actual command line.
    String legalExitStatusString;
    
    char* option_root_name = "OPT";
    int   option_count = 1;
    
    vovRegisterMainArgs( argc, argv );
    
    String affinity = utcc_machine();		// 
    
    VovPlaceList io(20);	// List of all IO places.
    
    String host = utcc_getenv( "VOV_HOST_NAME" );
    String project = utcc_getenv( "VOV_PROJECT_NAME" );
    
    String traceonly = getenv( "VOV_TRACE_ONLY" );
    if ( traceonly == "TRUE" ) {
	// vovMessage( "VOV_TRACE_ONLY=TRUE: Transition will not be run" );
	dontrun = 1;
    }
    
    VOVcatchAllSignals();
    VovTrace trace( project, host, "a", VOV_TRACE_TYPE_CLIENT );
    
    int oldumask = umask( 0 );
    
    int option;
    while (( option = optGetOption( argc, argv ) ) != EOF) {
	switch (option) {
	case 'I':		// Interactive transaction, no redirect.
	    interactive = 1;
	    stdinname = "";
	    stdoutname = "";
	    stderrname = "";
	    break;
	case 'c':
	    actualcommand = optarg;
	    break;
	case 'x':
	    executable = findExecutable( optarg );
	    io.add( VOV_EXECUTABLE, path_canonicalize( executable ), VOV_INPUT_MASK );
	    break;
	case 'y':
	    executable =  optarg;
	    io.add( VOV_EXECUTABLE, path_canonicalize( executable ), VOV_OUTPUT_MASK );
	    break;
	case 'B':
	    io.add( VOV_BOOLEAN, path_canonicalize( optarg ), VOV_INPUT_MASK );
	    break;
	case 'P':
	    option_root_name = optarg;  // 
	    option_count = 1;		// Reset option count.
	    break;
	case 'p':
	    {
		String optionName = form("%s%d", option_root_name, option_count++ );
		String placeName = utcc_getwd() + "/" + optionName;
		VovPlace place( VOV_PROGRAM_OPTION, placeName );
		
		if ( trace.find( place ) == VOV_OK ) {
		    if ( place.getValue().string() != optarg ) {
			vovMessage( place.format() + " exists already, but it has a different value" );
			place.getValue() = optarg;
			VOV_ASSERT( trace.modify( place ) );
			vovMessage( "Modified to " + place.format() );
		    }
		}
		
		place.getValue() = optarg;
		// place.canonicalize();
		vovMessage( place.format() );
		
		VovPlace* placeP = new  VovPlace( place );
		io.add( *placeP, VOV_INPUT_MASK );
	    }
	    break;
	case 'i':
	    io.add( guessType( optarg ), path_canonicalize( optarg), VOV_INPUT_MASK );
	    break;
	case 'o':
	    io.add ( guessType( optarg ), path_canonicalize( optarg), VOV_OUTPUT_MASK );
	    break;
	case 'a':
	    io.add( guessType( optarg ), path_canonicalize( optarg), VOV_INPUT_MASK | VOV_OUTPUT_MASK );
	    break;
	case 'X':
	    exitStatus = 1;	// Record exit status.
	    break;
	case 'r':
	    stderrname = optarg;
	    break;
	case 'R':
	    stdoutname = optarg;
	    break;
	case 's':
	    stdinname = optarg;
	    break;
	case 'S':
	    legalExitStatusString = optarg;
	    break;
	case 'A':
	    affinity = optarg;
	    if ( affinity == "none" ) { // Special code for no affinity.
		affinity = 0;
	    }
	    break;
	case 'd':
	    VOVdebugCommunication( trace );
	    break;
	case 'n':
	    dontrun = 1;
	    break;
	case 'V':
	    cout << "VOV CAPSULE: " << VOV_VERSION << "\n";
	    break;
	default:
	    break;
	}
    }
    
    
    int new_argc = argc - optind;
    char** new_argv = &argv[optind];
    
    if (new_argc < 1 ) {
	optUsage();
    }
    
    String arglist;
    for (int j = 0; j < new_argc ; j++) {
	arglist += new_argv[j];
	arglist += " ";
    }
    
    
    if ( ! interactive ) {
	if ( actualcommand != 0 ) {
	    if ( stderrname == 0 && stdoutname == 0 ) {
		stdoutname = vov_stdoutRedirect( actualcommand );
		stderrname = stdoutname;
	    }
	}
    }
    capsulecommand = createCommand( arglist, stdinname, stdoutname, stderrname );
    if ( actualcommand == 0 ) actualcommand = capsulecommand;
    
    if ( stdinname  != 0 )  io.add( VOV_UNIX_FILE, path_canonicalize( stdinname ), VOV_INPUT_MASK );
    if ( stdoutname != 0 ) {
	io.add( VOV_UNIX_FILE, path_canonicalize( stdoutname ), VOV_OUTPUT_MASK );
	if ( stderrname != 0 && stderrname != stdoutname) {
	    io.add( VOV_UNIX_FILE, path_canonicalize( stderrname ), VOV_OUTPUT_MASK );
	}
    } else {
	if ( stderrname != 0) {
	    io.add( VOV_UNIX_FILE, path_canonicalize( stderrname ), VOV_OUTPUT_MASK );
	}
    }
    
    
    // VOVecho("Connected to server");
    
    VovTransition trans( actualcommand, capsulecommand );
    
    trans.setAffinity( affinity );			// Set.
    trans.setExitStatus( legalExitStatusString );
    
    
    VOV_ASSERT( trace.create( trans ) );
    VOV_ASSERT( trace.initiate( trans ) );
    
    VovUser user( utcc_getenv("USER") );
    VOV_ASSERT( trace.findOrCreate( user ) );
    VOV_ASSERT( trace.activeDesigner( user , trans ) );
    
    // First declare all inputs, then worry about the outputs.
    // This is necessary in the case that some places are both input and outputs.
    
    
    int errorflag = 0;
    int index;
    
    VovStatus s = trace.declareIOs( trans, io ); // Declare all IOs at once.
    if ( s != VOV_OK ) {			 // If there is a problem, check for it.
	for ( int i = 0 ; i < io.length() ; i++ ) {
	    if ( io.getStatus(i) == VOV_OK ) continue;
	    VovPlace& place = io.getPlace( i );
	    VOV_ASSERT( trace.getById( place ) );
	    // vovMessage( "Double checking on " + place.format() );
	    if ( io.getType(i) & VOV_INPUT_MASK ) {
		if ( VOVaddInput( trace, trans, place ) != VOV_OK ) {
		    errorflag = 1;
		}
	    }
	    if ( io.getType(i) & VOV_OUTPUT_MASK ) {
		if ( VOVaddOutput( trace, trans, place ) != VOV_OK ) {
		    errorflag = 1;
		}
	    }
	}
    } else {
	// vovMessage( "All IOs are OK" );
    }
    
    
    
    int ret = 0;			// The exit status of this capsule.
    
    if ( errorflag ) {
	vovError("Not executing because of error.");
	trace.destroy( trans );
	umask( oldumask ) ;
	exit( 255 );
    }
    
    if ( dontrun ) {
	vovMessage("NOT ACTUALLY RUNNING THE TRANSITION");
	VOV_ASSERT( trace.terminate( trans, VOV_ERROR, 0, 0 ) );
	VOV_ASSERT( trace.changeStatus( trans, VOV_PLAN_NODE, VOV_CHANGE_STATUS_SELF_REC ) );
    } else {
	shellSignals();
	String errmsg;
	util_unsetenv( "VOV_HOST_NAME" ); // prevent connection of encapsulated tool.
	if ( trans.fire( 0 , VOV_RETRACE_FAST, errmsg ) != VOV_OK ) {
	    vovFatalError( errmsg );
	}
	
	VovStatus ss = trans.waitCompletion( &ret, errmsg );
	
	{
	    // Output stdout and stderr.
	    if ( utcc_getenv( "VOV_NO_ECHO", 0 ) == 0 ) {
		if ( stderrname != 0  ) {
		    utcc_cat( stderrname );
		}
		if ( stdoutname != 0 && stderrname != stdoutname ) {
		    utcc_cat( stdoutname );
		}
	    }
	}
	// vovMessage( form("Transition completed: status %d", ret ) );
	
	if ( exitStatus ) {
	    VovPlace exitStatusPlace( VOV_EXIT_STATUS, form( "%d-%d", trans.id(), ret ) );
	    VOV_ASSERT( trace.create( exitStatusPlace ) );
	    String errmsg;
	    if ( trace.declareOutput( trans, exitStatusPlace , 1, 0, errmsg ) != VOV_OK ) {
		vovMessage( errmsg );
	    }
	}
	
	for ( index = 0; index < io.length() ; index++ ) {
	    VovPlace& o = io.getPlace( index );
	    if ( (io.getType(index) & VOV_OUTPUT_MASK) == 0 ) continue;	// Skip non inputs.
	    
	    int mode = o.getProtection();
	    if ( ( mode & 0666 ) != 0666 ) {
		// vovMessage( "Changing protections on " + o.format() );
		o.setProtection( 0666 | mode );
	    }
	}
	
	{
	    VovStatus ts = VOV_OK;	// Transition status code.
	    if ( ss == VOV_ERROR || ret == 127 ) {
		vovMessage( "The transaction has failed." );
		ts = VOV_ERROR;
	    }
	    if ( trace.terminate( trans, ts, ret, errmsg ) != VOV_OK ) {
		vovMessage( errmsg );
	    }
	}
	
	{
	    VovPlace cwd( VOV_UNIX_FILE, "." );
	    int mode = cwd.getProtection();
	    if ( ( mode & 0666 ) != 0666 ) {
		// vovMessage( "Changing protections on " + cwd.format() );
		cwd.setProtection( 0666 | mode );
	    }
	}
	umask( oldumask ) ;
    }
    
    // vovMessage( form( "Exiting %d", ret ) );
    exit( ret );
}







