#include "String.h"
#include "stream.h"

#include "port.h"
#include "vov.h"
#include "oct.h"
#include "voverror.h"
#include "errtrap.h"
#include "utcc.h"
#include "utility.h"

extern "C"
{
#include <sys/socket.h> 
#include <sys/time.h>
}

extern int vov_strhash( char* string, int modulus  );

static int unique_id = 0;
static char* executableName = "dummy";

#define CHANNEL_EOF_ERROR 77


void channelErrorHandler( char* pkgName, int code, char* message )
{
    if ( pkgName == "VOV CHANNEL" ) {
	if ( code == CHANNEL_EOF_ERROR ) {
	} else {
	    vovError( form("CHANNEL ERROR (code %d)\n\t%s", code, message));
	    ERR_IGNORE( VOV__end( -1 ) );
	}
	_exit( -1 );
    } 
    /* Pass error to next handler */
}

String readString( FILE* file )
{
    String s;
    int count = 0;
    int c;
    while (  ( c = getc( file ) ) != '\n' ) {
	if ( c == EOF ) {
	    errRaise( "VOV CHANNEL", CHANNEL_EOF_ERROR,
		     "Unexpected EOF. I assume tool exited without calling VOVend()." ); 
	}
	count++;
	s += c;
    }
    if ( count > 0 ) {
	return s;
    } else {
	cerr.form( "VOV CHANNEL: nothing to read" );
	return "";
    }
}

int readInt( FILE* file )
{
    String s = readString( file );
    return atoi( (char*)s );
}

inline void writeInt( FILE* file, int n )
{
    fprintf( file, " %d ", n );
    fflush( file );
}



void channel( FILE* readStream, FILE* writeStream )
{
    String s = readString( readStream );
    if ( s == "BEGIN" ) {
	//  
	// Allocate all args, because they may be needed by other
	// routines (e.g. dds_initialize())
	//
	int argc = readInt( readStream );
	char** argv = malloc( argc * sizeof(char*) );
	for ( int i = 0 ; i < argc ; i++ ) {
	    argv[i] = util_strsav( (char*)readString( readStream ) );
	}
	argv[argc] = 0;
	int rtn = VOV__begin( argc, argv );

	writeInt( writeStream, rtn );
	if ( rtn == oct_null_id ) {
	    vovMessage( "Channel keeps running to service I/O" );
	}

	{
	    String commandline = argv[0];
	    for ( int i = 1; i < argc ; i++ ) {
		commandline += form( " %s", argv[i] );
	    }
	    executableName = strrchr( argv[0], '/' );
	    if ( ! executableName ) {
		executableName = argv[0];
	    } else {
		executableName++; // Skip over the '/'
	    }
	    unique_id = vov_strhash( (char*)commandline, 0x00ffffff );
	}
    } else if ( s == "INPUT" ) {
	int type = readInt( readStream );
	String name =  readString( readStream );
	int rtn = VOV__input( type, (char*)name );
	writeInt( writeStream, rtn );
    } else if ( s == "OUTPUT" ) {
	int type = readInt( readStream );
	String name =  readString( readStream );
	int rtn = VOV__output( type, (char*)name );
	writeInt( writeStream, rtn );
    } else if ( s == "AFFINITY" ) {
	String aff = readString( readStream );
	VOV__affinity( (char*)aff );
	writeInt( writeStream, 0 );
    } else if ( s == "DELETE" ) {
	int type = readInt( readStream );
	String name = readString( readStream );
	int rtn = VOV__delete( type, (char*)name );
	writeInt( writeStream, rtn );
    } else  if ( s == "END" ) {
	VOV__end( readInt( readStream )); // this does not return.
    } else {
	cerr.form( "CHANNEL: Unparsable input \"%s\"\n", (char*)s );
	writeInt( writeStream, 0 );
    }
}



void tee( FILE* in, FILE* f1, FILE* f2 )
{
    int c;
    const int size = 1024;
    char buf[size];
    int n;

    n = read( fileno(in), buf, size );
    if ( n > 0 ) {
	for ( int i = 0; i< n ; i++ ) {
	    fputc( buf[i], f1 );
	    fputc( buf[i], f2 );
	}
	fflush( f1 );
	fflush( f2 );
    }
}


typedef enum { CH_STDOUT, CH_STDERR, CH_CHANNEL, CH_NONE } ChannelType;

struct channel_struct {
    char* name;
    char* mode;
    int  fd;
    int  mask;			// Should be changed to an array.
    FILE* fp;
    ChannelType type;
};

struct channel_struct chan[] = {
    { "stdout",   "r", -1 , 0, 0, CH_STDOUT },
    { "stderr",   "r", -1 , 0, 0, CH_STDERR },
    { "get"  ,    "r", -1 , 0, 0, CH_CHANNEL },
    { "respond" , "w", -1 , 0, 0, CH_NONE }
};




static FILE* saveOpenFile( char* name )
{
    char* fileName = form( ".%s.%s.%07d", name, executableName, unique_id );
    VOV__output( VOV_UNIX_FILE, fileName );
    return fopen( fileName, "w" );
}

main( int argc, char** argv )
{
    errProgramName( "vov_channel" );
    errCore( 1 );		// Dump core on error.
    int readMask = 0;
    
    
    int size = sizeof( chan ) / sizeof( struct channel_struct );
    if ( argc != size + 1 ) {
	cerr<<"Usage: vov_channel stdout stderr get respond\n";
	errRaise( "VOV CHANNEL", 1, "Should be called with %d arguments.", size );
    }
    
    int maxFd = 0;
    for ( int i = 0; i < size ; i++ ) {
	chan[i].fd = atoi( argv[ i + 1 ] );
	chan[i].mask = ( 1 << chan[i].fd );
	chan[i].fp = fdopen( chan[i].fd, chan[i].mode );
	
	if ( chan[i].fd > maxFd ) maxFd = chan[i].fd;
	if ( strcmp( chan[i].mode, "r" ) == 0 ) readMask |= chan[i].mask;
	// cerr.form( "CHANNEL: %s (%s) %d\n", chan[i].name, chan[i].mode, chan[i].fd );
    }
    
    
    while ( 1 ) {
	int selectMask = readMask;
	int zero = 0;
	int status = select( maxFd + 1, &selectMask, &zero, &zero, 0); // No timeout
	if ( status == -1 ) {
	    utcc_printerrno( "vov_channel" );
	    exit( 1 );
	}
	
	for ( int i = 0 ; i < size ; i++ ) {
	    if ( selectMask & chan[i].mask ) {
		switch ( chan[i].type ) {
		case CH_STDOUT:
		    {
			static FILE* saveOut = 0;
			if ( ! saveOut ) {
			    saveOut = saveOpenFile( chan[i].name );
			}
			tee( chan[i].fp, stdout, saveOut );
			break;
		    }
		case CH_STDERR:
		    {
			static FILE* saveErr = 0;
			if ( ! saveErr ) {
			    saveErr = saveOpenFile( chan[i].name );
			}
			tee( chan[i].fp, stderr, saveErr );
			break;
		    }
		case CH_CHANNEL:
		    errPushHandler( channelErrorHandler );
		    channel( chan[i].fp, chan[i+1].fp );
		    errPopHandler();
		    break;
		default:
		    vovFatalError( "CHANNEL: ?????" );
		}
	    }
	}
    }
    cerr << "VOV CHANNEL: Exiting for lack of input.\n";
    exit( 1 );
}



