/* C++ includes */
#include <stream.h>


/* C includes */
extern "C"
{
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h> 
#include <netinet/in.h>
#include <sys/time.h>

extern void setServerFd( int fd );

}


/* Local includes */
#include "bzocket.h"
#include "utcc.h"		// Utils  

#define INTEGER_DATA  0
#define STRING_DATA   2

int padded( int n )
{
    int lbits = n & 0x3;
    if ( lbits ) {
	n += 4 -lbits;
    }
    return n;
}

Bzocket::Bzocket(int srv = 1)
{
    on = 0;
    fd = 0;
    server = srv;
    debug = 0;
    errorString = "<no error>";
}

/* Do not copy the buffers. */
Bzocket::Bzocket( Bzocket& z )
{
    on = z.on;
    server = z.server;
    fd = z.fd;
    debug = z.debug;
    errorString = z.errorString;
}

Bzocket::~Bzocket()
{
}

void Bzocket::start(int i)
{
    debug = 0;			// Reset the debug flag for each new connection.
    fd = i;
    on = 1;

    /* The server must not get stuck */
    if (server) {
	setServerFd( fd );
    }
    inBuf.resize(DEFBUFSIZE);
    outBuf.resize(DEFBUFSIZE);
    outBuf.clear();
    if ( debug ){
	cerr.form("\tBZ:socket %d is turned on\n", fd);
    }
}

int Bzocket::ready()
    /*
    * Is there a packet ready to be read ?
    * RETURN: BUF_OK if yes, BUF_ERROR if no.
    */
{
    if ( ! on ) return BUF_ERROR;
    return inBuf.packetReady();
}


int Bzocket::read()
{
    if (inBuf.readSocket(fd) == BUF_ERROR) {
	if ( debug ) {
	    cerr << "\tBZ:failed reading from socket" << fd << "\n";
	    print();
	}
	return BUF_ERROR;
    }
    return BUF_OK;
}

int Bzocket::waitForPacket( int sec = 0, int usec = 0)
{
    struct timeval waittime = { sec, usec };

    fd_set mask;
    FD_ZERO( &mask );
    FD_SET( fd, &mask );
    int s = select( fd + 1, &mask, 0, 0 , &waittime );
    if ( s == -1 ) {
	utcc_error( "Waiting for packet." );
	return BUF_ERROR;
    } else if ( s == 0 ) {
	return BUF_TIMEOUT;
    } else { 
	return BUF_OK;
    }
}



/* If off, no problem, but if on it must be able to respond */
int Bzocket::check()
{
    if (! on) return BUF_OK;	

    struct timeval notime = {0, 0};
    fd_set mask;

    FD_ZERO(&mask);
    FD_SET( fd, &mask );

    if (select (fd + 1, &mask, (int *) NULL, (int *) NULL, &notime) < 0) {
	utcc_printerrno("BZ:check");
	if ( debug ) cerr.form("\tBZ:Failed check on %d\n",fd);
	return BUF_ERROR;
    }
    return BUF_OK ;
}

void Bzocket::close()
{
    ::close(fd);
    on = 0;
    inBuf.freeMemory();
    outBuf.freeMemory();
    if ( debug ) {
	cerr.form("\tBZ:closed %d\n",fd);
    }
}




void Bzocket::pack(int i)
{
    outBuf.put( INTEGER_DATA );
    outBuf.put( i  );
    if ( debug ) {
	cerr.form("\tBZ:pack integer %d\n",i);
    }
}


void Bzocket::pack( char* s)	/* null terminated string */
{
    int len = strlen(s) + 1 ;
    len = padded(len);
    outBuf.put( STRING_DATA) ;
    outBuf.put( len );
    outBuf.put( s , len );
    if ( debug ) {
	cerr.form("\tBZ:pack string %s (%d)\n",s,len);
    }
}

void Bzocket::pack( String& s)	
{
    int len = padded( s.length()+1 );
    outBuf.put( STRING_DATA) ;
    
    outBuf.put( len );
    outBuf.put( (char*)s , len );
    if ( debug ) {
	cerr.form("\tBZ:pack String %s (%d)\n",(char*)s, len );
    }
}

// Unpacking routines:
// 
int Bzocket::unpack( int* ip)
{
    int type = inBuf.getInt();
    if (type != INTEGER_DATA) {
	errorString = "Wrong data type: not an integer";
	inBuf.unget( sizeof(int) );
	return BUF_ERROR;
    }
    *ip = inBuf.getInt();
    if ( debug ) {
	cerr.form("\tBZ:unpack integer %d\n",*ip);
    }
    return BUF_OK;
}

int Bzocket::unpack( char** sp)
{
    int type = inBuf.getInt();
    if (type != STRING_DATA) {
	errorString = "Wrong data type: not a string";
	inBuf.unget( sizeof(int) );
	return BUF_ERROR;
    }
    int len = inBuf.getInt();
    if ((*sp = inBuf.get(len)) == 0) {
	errorString = "Cannot get length of string\n";
	inBuf.unget( sizeof(int) );
	return BUF_ERROR;
    }
    if ( debug ) {
	cerr.form("\tBZ:unpack string %s (%d)\n",*sp,len);
    }
    return BUF_OK;
}


/* Fragile routine. */
int Bzocket::unpackInt()
{
    int n;
    if ( unpack( &n ) == BUF_ERROR ) {
	utcc_fatalerror( errorString );
    }
    return n;
}

char* Bzocket::unpackStr()
{
    char* s;
    if ( unpack( &s ) == BUF_ERROR ) {
	utcc_fatalerror( errorString );
    } 
    return s;
}

String Bzocket::unpackString()
{
    String s = unpackStr();
    return s;
}


int Bzocket::discardPacket()
{
    int l;
    char* c = inBuf.getPacket(&l);
    if ( debug ) {
	cerr.form("\tBZ:discarding %d characters\n", l);
    }
    return BUF_OK;
}

/* Send packet in outBuf and clear outBuf. */
void Bzocket::flush()
{
    if ( outBuf.sendSocket(fd) == BUF_ERROR ) {
	utcc_printerrno("BZ:flush"); 		// Error, but continue anyway.
    }		
    outBuf.clear();			
    if ( debug ) cerr.form("\tBZ:flushed socket %d\n", fd);
    return;   // VOID.
}

void Bzocket::clear()
{
    outBuf.clear();
}

void Bzocket::print()
{
    cout.form("%s, %s\n", on ? "ON":"OFF", server ? "SERVER": "NOT SERVER");
    inBuf.print();
    outBuf.print();
}


