#include "chain.h"
#include "oh.h"
#include "voverror.h"

VovChainInfo::VovChainInfo(  )
    :
    ( OCT_NET ),
    first( "FIRST" ),
    current( "CURRENT" ),
    last( "LAST" )
{
}

VovChainInfo::~VovChainInfo(  )
{
}

VovStatus VovChainInfo::create( VovPlace& firstPlace, VovPlace& secondPlace )
{
    
    OH_ASSERT( octCreate( this, &first ) );
    OH_ASSERT( octCreate( this, &current ) );
    OH_ASSERT( octCreate( this, &last ) );
    OH_ASSERT( octAttach( this, &firstPlace ) );
    OH_ASSERT( octAttach( this, &secondPlace ) );

    first.add( firstPlace );
    last.add( secondPlace );
    current.add( secondPlace );

    secondPlace.current = 1;
    firstPlace.current = 0;
    OH_ASSERT( octModify( &firstPlace.current ) );
    OH_ASSERT( octModify( &secondPlace.current ) );

    return VOV_OK;
}


VovStatus VovChainInfo::get()
{
    OH_ASSERT( octGetOrCreate( this, &first ) ); // Backward compatibility.
    OH_ASSERT( octGetByName( this, &current ) ); // 
    OH_ASSERT( octGetOrCreate( this, &last ) );	 // Backward compatibility.
    return VOV_OK;
}

VovStatus VovChainInfo::del()
{
    OH_ASSERT( octDelete( &first ) );
    OH_ASSERT( octDelete( &current ) );
    OH_ASSERT( octDelete( &last ) );
    OH_ASSERT( octDelete( this ) );
    return VOV_OK;
}

// Destroy all places in chain.
VovStatus VovChainInfo::destroy()
{
    vovMessage( "Disassembling chain " + name() );
    VovContentsGenerator genPlace( this, VOV_PLACE_OCTMASK );
    VovPlace p1;
    while ( octGenerate( &genPlace, &p1 ) == OCT_OK ) {
	if ( p1.get() != VOV_OK ) {
	    vovError( form( "Generating faulty place from chain" ) );
	    continue;
	}
	p1.del();
    }
    return  del();			// Delete chain itself.
}

VovStatus VovChainInfo::setCurrent( VovPlace& newCurrentPlace )
{
    OH_ASSERT( octIsAttached( this, &newCurrentPlace ) );

    VovPlace curPlace;
    getCurrent( curPlace );
    curPlace.current = 0;
    OH_ASSERT( octModify( &curPlace.current ) );

    current.clear( VOV_PLACE_OCTMASK );
    current.add( newCurrentPlace );
    newCurrentPlace.current = 1;
    OH_ASSERT( octModify( &newCurrentPlace.current ) );

    return VOV_OK;
}

VovStatus VovChainInfo::addPlace( VovPlace& place )
{
    if ( octIsAttached( this, &place ) == OCT_OK ) {
	vovError("Chain: already added");
	return VOV_ERROR;
    }
    // vovMessage( "CHAIN: Added place in last position." );
    OH_ASSERT( octAttach( this, &place ) );

    last.clear( VOV_PLACE_OCTMASK );
    last.add( place );

    return VOV_OK;
}

VovStatus VovChainInfo::delPlace( VovPlace& place )
{
    VovPlace f,l,c;
    getFirst( f );
    getLast( l );
    getCurrent( c );

    if ( f.id() == place.id() ) {
	VOV_ASSERT( getNext( place, f ) );
	first.clear( VOV_PLACE_OCTMASK );
	first.add( f );
    }
    if ( l.id() == place.id() ) {
	VOV_ASSERT( getPrec( place, l ) );
	last.clear( VOV_PLACE_OCTMASK );
	last.add( l );
    }
    if ( c.id() == place.id() ) {
	if ( getPrec( place, c ) != VOV_OK ) { 		// If there is no preceding place,
	    VOV_ASSERT( getNext( place, c ) );		// get the next one.
	}						// 
	c.setCurrent();					// Set this place as the new current one.
    }
    OH_ASSERT( octDetach( this, &place ));
    return VOV_OK;
}

void VovChainInfo::getCurrent( VovPlace& p )
{
    OH_ASSERT( octGenFirstContent( &current, VOV_PLACE_OCTMASK, &p ));
    VOV_ASSERT( p.get() );
}

void VovChainInfo::getFirst( VovPlace& p )
{
    OH_ASSERT( octGenFirstContent( &first, VOV_PLACE_OCTMASK, &p ) );
    VOV_ASSERT( p.get() );
}

void VovChainInfo::getLast( VovPlace& p )
{
    OH_ASSERT( octGenFirstContent( &last, VOV_PLACE_OCTMASK, &p ) );
    VOV_ASSERT( p.get() );
}


VovStatus VovChainInfo::precedes( VovPlace& p1, VovPlace& p2 )
{
    if ( p1.id() == p2.id() ) return VOV_NOT_OK; // Places coincide.
    VovPlace p,q;
    p = p1;
    while ( getNext( p, q ) == VOV_OK ) {
	if ( q.id() == p2.id() ) {
	    return VOV_OK;
	} else {
	    p = q;
	}
    }
    return VOV_NOT_OK;		// p1 does not preced p2
    
}

VovStatus VovChainInfo::getPrec( VovPlace& place, VovPlace& prec )
{
    if ( &place == &prec ) {
	vovBug( "getPrec:Must give separate address" );
    }
    VovNode trans = 0;
    if ( octGenFirstContainer( &place, VOV_NODE_OCTMASK, &trans ) != OCT_OK ) {
	return VOV_ERROR;
    }
    VovContentsGenerator genPlace( this, VOV_PLACE_OCTMASK );
    while ( octGenerate( &genPlace, &prec ) == OCT_OK ) {
	VOV_ASSERT( prec.get() );
	if ( trans.isInput( prec ) == VOV_OK ) {
	    if ( place.id() == prec.id() ) {
		vovError("just a debugging message. getPrec screwed up" );
		return VOV_ERROR;
	    }
	    return VOV_OK;	// Found !!
	}
    }
    return VOV_ERROR;		// Not found
}

VovStatus VovChainInfo::getNext( VovPlace& place, VovPlace& next )
{
    if ( &place == &next ) {
	vovBug( "getNext:Must give separate address" );
    }
    VovContentsGenerator genPlace( this, VOV_PLACE_OCTMASK );
    while ( octGenerate( &genPlace, &next ) == OCT_OK ) {
	VOV_ASSERT( next.get() );
	VovNode trans= 0;	// Node that produces 'place'
	if ( octGenFirstContainer( &next, VOV_NODE_OCTMASK, &trans ) != OCT_OK ) {
	    continue;		// This cannot be the 'next' in chain.
	}
	if ( trans.isInput( place ) == VOV_OK ) {
	    if ( place.id() == next.id() ) {
		vovError("just a debugging message. getNext screwed up" );
		return VOV_ERROR;
	    }
	    return VOV_OK;	// Found !!
	}
    }
    return VOV_ERROR;
}

int VovChainInfo::position( VovPlace& place )
{
    VovPlace c1, c2;
    getFirst( c1 );
    int count = 1;
    while ( 1 ) {
	if ( c1.id() == place.id() ) {
	    vovMessage( form( "Selected place is %d in ", count ) + format() );
	    break;
	}
	count++;
	if ( getNext( c1, c2 ) != VOV_OK ) {
	    vovError( "Could not find position of place in chain" + format() ) ;
	    return 0;
	}
	c1 = c2;
    }
    return count;
}


String VovChainInfo::format()
{
    VovPlace f = 0;
    VovPlace c = 0;
    VovPlace l = 0;
    getFirst( f );
    getCurrent( c );
    getLast( l );
    int n = length();
    String s = form( "CHAIN %s: %d places. f%d c%d l%d", (char*)name(), n, f.id(), c.id(), l.id() ) ;
    return s;
}



int VovChainInfo::length()
{
    return ohCountContents( this, VOV_PLACE_OCTMASK );
}

VovStatus VovChainInfo::OK()
{
    // vovMessage( "Testing: " + format() );
    VovPlace c;			// Current.
    VovPlace f;			// First.
    VovPlace l;			// Last.
    VovPlace p1,p2;		// Another place used to run along the chain.

    String msg;
    VovStatus status = VOV_OK;

    getCurrent( c );
    getFirst( f );
    getLast( l );
    
    int len = length();
    if ( len > 1 ) {
	if ( precedes( f, l ) != VOV_OK ) {
	    msg = "first does not precede last.\n" ;
	    status = VOV_ERROR;
	}
    }

    // Run forward along chain.
    p1 = f;
    int count = 0;
    while ( getNext( p1, p2 ) == VOV_OK ) {
	if ( p1.id() == p2.id() ) {
	    status = VOV_ERROR;
	    msg += "getNext failed\n";
	    break;
	}
	p1 = p2;
	if ( count++ > len ) {
	    vovError( "Cycle in chain" ) ;
	    status = VOV_ERROR;
	    break;
	}
    }
    if ( p1.id() != l.id() ) {
	status = VOV_ERROR;
	msg += "Error in forward traverse\n";
    }

    p1 = l;
    count = 0;
    while ( getPrec( p1, p2 ) == VOV_OK ) {
	if ( p1.id() == p2.id() ) {
	    status = VOV_ERROR;
	    msg += "getPrec failed\n";
	    break;
	}
	p1 = p2;
	if ( count++ > len ) {
	    vovError( "backward cycle in chain" );
	    status = VOV_ERROR;
	    break;
	}
    }
    if ( p1.id() != f.id() ) {
	status = VOV_ERROR;
	msg += "Error in backward traverse\n";
    }

    if ( status == VOV_ERROR ) {
	vovError( msg + format() );
	return fix();
    }
    return VOV_OK;
}

VovStatus VovChainInfo::fix()
{
    vovMessage( "Fixing chain:" + format() );

    VovPlace c;			// Current.
    VovPlace f;			// First.
    VovPlace l;			// Last.
    VovPlace p1,p2;		// Another place used to run along the chain.

    int len = length();
    int tmp = 0;
    String msg;
    VovStatus status = VOV_OK;

    if ( octGenFirstContent( this, VOV_PLACE_OCTMASK, &p1 ) != OCT_OK ) {
	vovError( "No place on " + format() );
	del();
	return VOV_ERROR;
    }

    
    // Run forward along chain.
    while ( getNext( p1, p2 ) == VOV_OK ) {
	p1 = p2;
	VOV_ASSERT( p1.get() );
	if ( tmp++ > len ) {
	    vovError( "Severely corrupted chain wit forward loop" );
	    destroy();
	    return VOV_ERROR;
	}
    }
    last.clear( VOV_PLACE_OCTMASK );
    last.add( p1 );
    l = p1;
    VOV_ASSERT( l.get() );

    OH_ASSERT( octGenFirstContent( this, VOV_PLACE_OCTMASK, &p1 ) );
    tmp = 0;
    while ( getPrec( p1, p2 ) == VOV_OK ) {
	p1 = p2;
	VOV_ASSERT( p1.get() );
	if ( tmp++ > len ) {
	    vovError( "Severely corrupted chain wit backward loop" );
	    destroy();
	    return VOV_ERROR;
	}
    }
    first.clear( VOV_PLACE_OCTMASK );
    first.add( p1 );
    f = p1;
    VOV_ASSERT ( f.get() );
    p1 = f;
    VOV_ASSERT( p1.get() );
    int count = 1;				// 
    while ( getNext( p1, p2 ) == VOV_OK ) {	// 
	count++;				// 
	p1 = p2;				// 
	VOV_ASSERT( p1.get() );			// 
    }
    
    int dismantle = 0;
    if ( f.id() == l.id() ) {
	vovError( "First and last cannot be the same." );
	dismantle = 1;
    }
    if ( count != length() ) {
	vovError( form( "Severely corrupted chain: length mismatch: is %d but I count %d", count, length()) );
	dismantle = 1;
    }

    if ( dismantle ) {
	VOV_ASSERT(  destroy() );
	return VOV_ERROR;
    } else {
	// Reset chain so that first node is current.
	setCurrent( f );
	f.changeStatus( VOV_PLAN_NODE, VOV_CHANGE_STATUS_SELF | VOV_CHANGE_STATUS_RECURSIVE, msg );
	return VOV_OK;		// Chain has been repaired.
    }
}



#ifdef UNUSED
VovStatus VovChainInfo::split( VovPlace& place, VovTransition& trans, VovPlace& newplace )
{

}
VovStatus VovChainInfo::merge( VovPlace& place, VovTransition& trans, VovPlace& newplace )
{
    vovMessage( "chain::merge is not used" );
    return VOV_OK;
}
#endif
