#include "trace.h"
#include "place.h"
#include "transition.h"
#include "event.h"
#include "voverror.h"

static int eventSerialNumber = 0;


int vov_nextEventSerialNumber() { return eventSerialNumber; }

VovEvent::VovEvent()
{
    serialNumber = 0;
    date = 0;
    subjectmask = VOV_NULL_MASK;
    eventmask =  VE_NULL_EVENT;
    description = "NULL EVENT";
    actionMask = 0;
    id = oct_null_id;
    obsolete = 0;
}

VovEvent::VovEvent( int sm, int em, String& desc, int i, int am )
{
    serialNumber = eventSerialNumber++;
    date.now();
    description = desc;
    subjectmask = sm;
    eventmask = em;
    actionMask = am;
    id = i;
    obsolete = 0;
}

VovEvent::VovEvent( VovEvent& e )
{
    serialNumber = e.serialNumber;
    date = e.date;
    subjectmask = e.subjectmask;
    eventmask = e.eventmask;
    description = e.description;
    actionMask = e.actionMask;
    id = e.id;
    obsolete = e.obsolete;
}

VovEvent::~VovEvent()
{
    
}
void VovEvent::operator=( VovEvent& e )
{
    serialNumber = e.serialNumber;
    date = e.date;
    subjectmask = e.subjectmask;
    eventmask = e.eventmask;
    description = e.description;
    actionMask = e.actionMask;
    id = e.id;
    obsolete = e.obsolete;
}

void VovEvent::set( int sm, int em, String& desc, int i, int am )
{
    serialNumber = eventSerialNumber++;
    date.now();
    description = desc;
    subjectmask = sm;
    eventmask = em;
    actionMask = am;
    id = i;
    obsolete = 0;
}


String VovEvent::format()
{
    String s = form("EVENT %4d, %s", serialNumber, (char*)date.format());
    if ( obsolete ) {
	s += "\n" ;
    } else {
	s += "     ((((MOST RECENT))))\n" ;
    }
    
    String subject;
    String verb;

    switch ( subjectmask ) {
    case VOV_NULL_MASK:
	subject = "NOTHING";
	break;
    case VOV_ANNOTATION_MASK:
	subject = "ANNOTATION";
	break;
    case VOV_PLACE_MASK:
	subject = "PLACE";
	break;
    case VOV_TRANSITION_MASK:
	subject = "TRANSITION";
	break;
    case VOV_RETRACE_MASK:
	subject = "RETRACE";
	break;
    case VOV_USER_MASK:
	subject = "USER";
	break;
    case VOV_SLAVE_MASK:
	subject = "SLAVE";
	break;
    case VOV_GOAL_MASK:
	subject = "GOAL";
	break;
    case VOV_NODE_SET_MASK:
	subject = "SET";
	break;
    default:
	subject = "UNKNOWN SUBJECT";
	break;
    }

    switch ( eventmask ) {
    case VE_NULL_EVENT:
	verb = ": NOTHING HAPPENED";
	break;
    case VE_START:
	verb = "HAS STARTED";
	break;
    case VE_END:
	verb = "HAS ENDED";
	break;
    case VE_CONNECT:
	verb = "HAS CONNECTED";
	break;
    case VE_DISCONNECT:
	verb = "HAS DISCONNECTED";
	break;
    case VE_CHANGE:
	verb = "HAS CHANGED";
	break;
    case VE_FORGET:
	verb = "HAS BEEN FORGOTTEN";
	break;
    case VE_ERROR:
	verb = "HAS AN ERROR";
	break;
    default:
	verb = "????";
	break;
    }
    s += ( subject + " " + verb + "\n" );
    s += description;
    return s;
}

void VovEvent::pack( Bzocket& zoc )
{
    zoc.pack( serialNumber );
    zoc.pack( date.get() );
    zoc.pack( subjectmask );

    zoc.pack( eventmask );
    zoc.pack( actionMask );
    zoc.pack( description );

    zoc.pack( id );
    zoc.pack( obsolete );
}

void VovEvent::unpack( Bzocket& zoc )
{
    serialNumber = zoc.unpackInt();
    date.set( zoc.unpackInt() );
    subjectmask = zoc.unpackInt();

    eventmask = zoc.unpackInt();
    actionMask  = zoc.unpackInt();
    description = zoc.unpackString();

    id = zoc.unpackInt();
    obsolete = zoc.unpackInt();
}


VovEventQueue::VovEventQueue(  )
{
    eqtrace = 0;
    size = 0;
    array = 0;
    count = 0;
    writeIndex = 0;
    readIndex = 0;
}
VovEventQueue::VovEventQueue( VovTrace& t  )
{
    eqtrace = &t;
    size = 0;
    array = 0;
    count = 0;
    writeIndex = 0;
    readIndex = 0;
}

VovEventQueue::VovEventQueue( VovTrace& t, int s )
{
    size = 0;
    init( t, s );
}

VovEventQueue::VovEventQueue( VovEventQueue& q )
{
    eqtrace = q.eqtrace;
    size = q.size;
    array = q.array;
    count = q.count;
    // This is not right !!
    writeIndex = q.writeIndex;
    readIndex = q.readIndex;
}

VovEventQueue::~VovEventQueue( )
{
    size = 0;
    delete array;
}

void VovEventQueue::init( VovTrace& t, int s )
{
    eqtrace = &t;
    if ( size != 0 ) {
	vovFatalError( "cannot resize voveventqueue" );
    }
    size = s;
    array = new VovEvent*[ size ];
    count = 0;
    writeIndex = 0;
    readIndex = 0;
}

int VovEventQueue::length()
{
    return count;
}

VovStatus VovEventQueue::deq( VovEvent& ev )
{
    if ( count > 0 ) {
	ev = *array[ readIndex ];
	delete array[ readIndex ];

	count--;
	readIndex++;
	readIndex %= size;
	return VOV_OK;
    } else {
	vovError( "No events in queue" );
	return VOV_ERROR;
    }
}

void VovEventQueue::markObsolete()
{
    if ( count < 2 ) return;
    for ( int i = writeIndex - 1; i > readIndex ; i-- ) {
	int eId = array[i]->id;	// Event Id.
	if ( eId == 0 ) continue;
	for ( int j = i-1 ; j >= readIndex ; j-- ) {
	    int oeId = array[j]->id;
	    if ( eId != oeId ) continue;
	    if ( array[j]->obsolete ) break;
	    array[j]->obsolete = 1;
	}
    }
}

VovStatus VovEventQueue::get( VovEvent& ev, int index  )
{
    if ( count == 0 ) {
	vovError( "No events in queue" );
	return VOV_ERROR;
    }

    if ( index < 0 ) {
	vovError( "get event in queue: index cannot be negative" );
	return VOV_ERROR;
    }
    
    if ( index < count ) {
	// writeIndex-1 is the index where the latest event was written.
	int i = ( writeIndex - 1  - index + size ) % size;
	ev = *array[ i ];
	return VOV_OK;
    } else {
	vovError( "Not so many elements in queue." );
	return VOV_ERROR;
    }
}

VovEvent& VovEventQueue::get(  int index  )
{
    if ( index < 0 || index >= size  ) {
	vovFatalError( "index out of range" );
    }
    return *array[ index ];
}


VovStatus VovEventQueue::enq( int sm, int em, String& desc, int i, int am )
{
    VovEvent ev( sm, em, desc, i, am );
    return enq( ev );
}

VovStatus VovEventQueue::enq( VovEvent& event )
{
    VovEvent* ev = new VovEvent(event);
    if ( count == size ) {
	delete array[writeIndex]; 	// Event queue overflowing. Old event discarded.
	readIndex++;			// Same ops as for a deq.
	readIndex %= size;		// 
    } else {
	count++;
    }
    array[ writeIndex++ ] = ev;
    writeIndex %= size;
    return VOV_OK;
}


String VovEventQueue::format()
{
    String s = form("Event queue: size %d, %d events\n", size, count );
    int c = count;
    int i = readIndex;
    while ( c-- > 0 ) {
	s += form( "%3d :", i ) ;
	s += array[ i++ ]->format();
	s += "\n";
    }
    return s;
}


int VovEventQueue::newIndexForEventQueueEdit( int cur, int sm, int em, int dir )
    // Finds the index of an event satisfying the requirements dictated
    // by the masks sm and em. dir specifies the direction to search starting from index cur.
    // It can return an illegal index, if no appropriate event is found.
{
    int i = cur;
    do {
	i += dir;
	if ( i >= length() ) {
	    i = length();
	    break;
	} 
	if ( i < 0 ) {
	    i = -1;
	    break;
	}
    } while( !(array[i]->getSubjectMask() & sm && array[i]->getEventMask() & em) );
    return i;
}

int vov_test_event_queue()
{
    VovEventQueue queue( *(VovTrace*)0, 20 );
    
    cerr << queue.format();
    queue.enq( VOV_RETRACE_MASK, VE_START, "start 1", 0 ,  0 );
    queue.enq( VOV_RETRACE_MASK, VE_END, "end 1", 0 ,  0 );

    queue.enq( VOV_RETRACE_MASK, VE_START, "start 2", 0 ,  0 );
    queue.enq( VOV_RETRACE_MASK, VE_END, "end 2", 0 ,  0 );

    queue.enq( VOV_RETRACE_MASK, VE_START, "start 3", 0 ,  0 );
    queue.enq( VOV_RETRACE_MASK, VE_END, "end 3", 0 ,  0 );

    queue.enq( VOV_RETRACE_MASK, VE_START, "start 4", 0 ,  0 );
    queue.enq( VOV_RETRACE_MASK, VE_END, "end 4", 0 ,  0 );


    queue.enq( VOV_PLACE_MASK, VE_CHANGE, "some place", 0 , 0 );

    cerr << queue.format();

    while ( queue.length() > 0 ) {
	VovEvent ev;
	if ( queue.deq( ev ) != VOV_OK ) {
	    vovFatalError("Cannot happen" );
	}
	cerr << queue.format();
    }
    return VOV_OK;
}


VovEventFilter::VovEventFilter( int sm, int em , String& s )
{
    subjectmask = sm;
    eventmask = em;
    id = oct_null_id;
    rx = s;
}

VovEventFilter::VovEventFilter( int sm, int em , int match_id )
{
    subjectmask = sm;
    eventmask = em;
    id = match_id;
    rx = 0;
}

VovEventFilter::~VovEventFilter() 
{
}

VovStatus VovEventFilter::filter( VovEvent& event )
{
    // First check id of event. If no match, no sense checking more.
    if ( id != oct_null_id ) {
	if ( id != event.id ) return VOV_NOT_OK;
    }
    if (( subjectmask & event.subjectmask) && ( eventmask & event.eventmask ) ) {
	if ( rx == 0 || rx == SELECT_ALL ) {
	    return VOV_OK;
	}
	String desc = event.description;
	if ( desc.contains( rx ) ) {
	    return VOV_OK;	    // vovMessage( desc + " CONTAINS " + rx + "\n" );
	} 
    } 
    return VOV_NOT_OK;
}


