// Fast find routines for VOV.
// Find place by name using a symbol table.

#include "node.h"
#include "place.h"
#include "voverror.h"
#include "vovinternal.h"


extern "C"
{
#define ANSI_H     /* So ansi.h will not be included. */
#undef ARGS
#define ARGS(args)  ()
#include <st.h>
}

static st_table* getTable( VovPlaceType type )
{
    static st_table** allFindTables = 0; 			// One table for each place type.
    const int SIZE = VOV_LAST_PLACE_TYPE-VOV_FIRST_PLACE_TYPE-1;// 
    if ( ! allFindTables ) {					// Need to initialize?
	allFindTables = new st_table*[ SIZE ];			// Allocate SIZE pointers.
	for ( int i = 0; i < SIZE ; i++ ) allFindTables[i] = 0;	// Make sure they are cleared
    }								// 
    int ti = type - VOV_FIRST_PLACE_TYPE - 1 ; 			// Which table should we use ?
    if ( ti < 0 || ti >= SIZE ) {				// Is it a legal table?
	vovError( "Place type is totally bogus" ); 		// This should never happen.
	return 0;						// Not a legal table pointer.
    }								// 
    st_table* table = allFindTables[ ti ];			// Get appropriate table.
    if ( table == 0 ) {						// Does table need initialization.
	int size = 64;						// Large size.
	int density = 5;					// 
	double grow_factor = 2.0;				// Fast growth.
	int reorder_flag = 0;					// 
	table = st_init_table_with_params(strcmp, st_strhash,
					  size, density,
					  grow_factor, reorder_flag );
	allFindTables[ ti ] = table;
    }
    return table;
}

VovStatus vov_place_find_add( VovPlace& place )
{
    VovStatus rtn = VOV_OK;
    st_table* placeST = getTable( place.getPlaceType() ); // Get appropriate table.
    char* key = utcc_strsav( (char*)place.expandedName() );	  // 
    
    if ( st_lookup( placeST, key, 0 ) == 1 ) {
 	vovMessage( form("Duplicate key %s in ST.",key)); // Debugging message only.
    } else {
	st_add_direct( placeST, key, place.id() );
    }
    return rtn;
}

VovStatus vov_place_find_del( VovPlace& place )
{
    VovStatus rtn = VOV_ERROR;
    st_table* placeST = getTable( place.getPlaceType() );	// Get appropriate table.
    char* name = utcc_strsav( (char*)place.expandedName() );		// Copy place name.
    octId id;
    char* nametmp = name;					// No, id is not valid, must be deleted.
    if ( st_lookup( placeST, nametmp, &id ) == 1 ) {		// Is the place in the table?
	if ( id == place.id() ) {				// Yes, but is this the one in the table?
	    if ( st_delete( placeST, &nametmp, &id ) == 1 ) {	// 
		delete nametmp;					// Delete the old useless key.
		rtn =  VOV_OK;					// Deletion successful.
	    }
	} else {
	    vovMessage( form("DELETE: %s not the one in the ST",name) ); // Debugging.
	}
    } else {
	vovMessage( form("DELETE: %s not in ST",name) );
    }
    delete name;
    return rtn;
}


VovStatus vov_place_find( VovPlace& place )
{
    VovStatus rtn = VOV_NOT_FOUND;
    octId id;
    
    st_table* placeST = getTable( place.getPlaceType() );	// Get appropriate table.
    if ( placeST == 0 ) {					// Do we have to initialize this table ?
	vovError( "Place type is totally bogus" ); 		// This should never happen.
	return VOV_NOT_FOUND;					// 
    }
    char* name = (char*)place.expandedName();		        // Copy place name.
    if ( st_lookup( placeST, name, &id ) == 1 ) {		// Look in symbol table.
	VovPlace p2 = id;					// In table, get corresponding place.
	if ( p2.id() != id ) {					// Is this id still valid? 
								//   The place might have been erased.
	    vovError( "FIND: Invalid id in symbol table" );
	    VOV_ASSERT( vov_place_find_del( p2 ) );		// 
	} else {						// 
	    VovChainInfo chain;					// 
	    if ( p2.getChain( chain ) == VOV_OK ) {		// Chains of places are complicated.
		chain.getCurrent( p2 );				// Only current place can satify match.
	    }							// 
	    if ( place.match( p2 ) == VOV_OK ) {		// Make sure this is a real match.
		place.id() = p2.id();				// Avoid assignment to places (seems buggy).
		VOV_ASSERT(place.get());
		rtn = VOV_OK;					// Yes, place found.
	    }							// 
	}							// 
    }
    return rtn;
}

VovStatus vov_place_find_init( VovNodeSet& placeSet )
{
    VovStatus rtn = VOV_OK;
    VovPlace p2,p3;						// The place used to do the search.
    VovContentsGenerator genPlace(&placeSet,VOV_PLACE_OCTMASK); // 
    while ( octGenerate( &genPlace, &p2 ) == OCT_OK ) {		// 
	VOV_ASSERT( p2.get() );					// 
	p3.id() = p2.id();					// Save id().
	if ( vov_place_find( p2 ) == VOV_OK  ) {
	    if ( p2.isChain( ) != VOV_OK )  {
		vovMessage( "Already in table but not chain: " + p2.format() );
		p3.get();	// Recover current place.
		p3.del();	// Destroy current place (entry is removed)
		VOV_ASSERT(vov_place_find_add( p2 )); // Restore entry.
	    }
	} else {
	    VOV_ASSERT( vov_place_find_add( p2 ) );
	}
    }		
    return rtn;	
}

