#include "builtin.h"
#include "trace.h"
#include "annotation.h"
// #include "measure.h"
#include "user.h"
#include "placematrix.h"
#include "arc.h"
#include "voverror.h"
#include "VOVprotocol.h"
#include "comm.h"

extern void vov_setGeoColor( VovObject* , char* );
extern char* vov_nodeColorName( VovNodeStatus );

//
// Coordinate conversion operations.
//
//
const int TRACE_SPACING = 40;
const int X_OFFSET = 2000;
const int Y_OFFSET = 20000;


octPoint matrixToOct( octPoint point )
{
    int rowIsOdd = odd( point.y );
    point.y *= -TRACE_SPACING;
    point.y += Y_OFFSET;
    point.y *= 20;			// 20 units is a lambda.
	
    point.x *= TRACE_SPACING;
    if ( rowIsOdd ) point.x += TRACE_SPACING / 2 ;
    point.x += X_OFFSET;
    point.x *= 20;			// 20 units is a lambda.

    return point;			// 
}

octPoint octToMatrix( octPoint point )
{
    point.y /= 20;
    point.y -= Y_OFFSET;
    point.y /= -TRACE_SPACING;
    
    point.x /= 20;
    point.x -= X_OFFSET;
    if ( odd( point.y ) ) point.x -= TRACE_SPACING / 2;
    point.x /= TRACE_SPACING;
    
    return point;
}

//
// Simple routine to associate a node with 
// a placement matrix. Given that we are working with just one
// page, for now, only one matrix at a time can be handled.
//
VovPlacementMatrix* VovNode::getOrAddPlacementMatrix( VovPlacementMatrix* m )
{
    static VovPlacementMatrix* matrix = 0;
    if ( m == 0 ) {
	return matrix;
    } else  {
	if (matrix != 0 )  {
	    vovMessage("getOrAddPlacementmatrix can only handle 1 matrix" );
	} 
	matrix = m;
	return 0;
    }
}

//
// Find the 'best' place for a place.
// If the place is isolated, place it on the top  of the trace.
// If the place is not isolated, place it at the average of the
// position of its inputs and outputs.
//
octPoint VovNode::bestPlace()
{
    int count = 0;
    int sumX = 0;
    VovContainersGenerator genIn( this, VOV_NODE_OCTMASK );
    VovContentsGenerator genOut( this, VOV_NODE_OCTMASK );
    VovNode node = 0;
    while ( octGenerate( &genIn, &node ) == OCT_OK ) {
	count++;
	sumX += node.contents.instance.transform.translation.x;
    }
    while ( octGenerate( &genOut, &node ) == OCT_OK ) {
	count++;
	sumX += node.contents.instance.transform.translation.x;
    }
    
    
    VovPlacementMatrix* matrix = getOrAddPlacementMatrix(0);

    octPoint c = octToMatrix( contents.instance.transform.translation ); // Current place
    octPoint n = c;		// New place

    if ( count ) {
	octPoint ave;
	ave.x = sumX / count ;
	ave.y = 0;
	ave = octToMatrix( ave );
	n.x = ave.x;
	n.y = level.integer();	// Level remains the same.
	n.x = matrix->getClosestSpot( n.y, n.x, id() );
    } else {			// Node is isolated
	n.y = 0;		// Place it in row 0.
	n.x = 0;
	n.x = matrix->getClosestSpot( n.y, n.x, id() );
    }
    return  n;			// New position.
}

/*
** Since we need to place the instances in some way, do this in some
** simple way, just to get started.
** A negative level simply resets the routine
*/
void VovNode::doPlacement( )
{
    VovPlacementMatrix* matrix = getOrAddPlacementMatrix(0);
    int adjustArcsFlag = 0;		
    octPoint oldCoord = octToMatrix( contents.instance.transform.translation );


    // This test succeeds for example upon initialization.
    if ( (contents.instance.transform.translation.x == 0 && 
	  contents.instance.transform.translation.y == 0 ) ||
	matrix->verify( id(), oldCoord.y, oldCoord.x ) == VOV_ERROR ) {
	
	oldCoord.y = level.integer();
	oldCoord.x = matrix->getFirstSpot( oldCoord.y );
	
	if (matrix->add( id(), oldCoord.y, oldCoord.x ) == VOV_ERROR) {
	    vovError("Could not get a good spot for this node");
	} 
	move( matrixToOct( oldCoord ) );
    }
    
    if ( matrix->verify( id(), oldCoord.y, oldCoord.x ) == VOV_ERROR ) {
	vovError("Inconsistent placement matrix for:"+format() );
	vovError("doPlacement: This inconsistency should never happen" );
    }
    
    octPoint best = bestPlace();
    if ( best.x != oldCoord.x || best.y != oldCoord.y ) {
	if (matrix->del( id(), oldCoord.y, oldCoord.x ) == VOV_ERROR) {
	    vovError("doPlacement: error deleting from matrix");
	}
	if (matrix->add( id(), best.y, best.x ) == VOV_ERROR) {
	    vovError("doPlacement: error adding to matrix");
	}
	move(  matrixToOct( best ) );
    }
    // VOID.
}


#ifdef OBSOLETE
void VovGoal::doPlacement()
{
    int ty = Y_OFFSET + 1 * TRACE_SPACING;
    ty *= 20;			/* 20 units is a lambda */

    VovObject goalBag;
    if ( octGenFirstContainer( this, OCT_BAG_MASK, &goalBag ) != OCT_OK ) {
	vovMessage("cannot place this goal:" + format() );
	return;
    }
    int count = 0;
    VovContentsGenerator genGoal( &goalBag, VOV_GOAL_OCTMASK );
    VovGoal goal;
    while ( octGenerate( &genGoal, &goal) == OCT_OK ) {
	if ( goal.id() == id() ) {
	    break;
	} else {
	    count++;
	}
    }
    
    int tx =  count * TRACE_SPACING;
    tx += X_OFFSET;
    tx *= 20;			/* 20 units is a lambda */

    contents.instance.transform.translation.x = tx;
    contents.instance.transform.translation.y = ty;
    OH_ASSERT( octModify( this ) );

    /* VOID */
}
#endif

void VovUser::doPlacement()
{
    int ty = Y_OFFSET + 2 * TRACE_SPACING;
    ty *= 20;			/* 20 units is a lambda */

    VovObject userBag;
    if ( octGenFirstContainer( this, OCT_BAG_MASK, &userBag ) != OCT_OK ) {
	vovMessage("cannot place this user:" + format() );
	return;
    }
    int count = 0;
    VovContentsGenerator genUser( &userBag, VOV_USER_OCTMASK );
    VovUser user("<dummy>");
    while ( octGenerate( &genUser, &user) == OCT_OK ) {
	if ( user.id() == id() ) {
	    break;
	} else {
	    count++;
	}
    }
    
    int tx =  count * TRACE_SPACING;
    tx += X_OFFSET;
    tx *= 20;			/* 20 units is a lambda */

    contents.instance.transform.translation.x = tx;
    contents.instance.transform.translation.y = ty;
    OH_ASSERT( octModify( this ) );

    /* VOID */
}


void VovTrace::doPlacement()
{
    Bzocket& zoc = getBzocket();
    if ( isClient() ) {
	zoc.pack( VOV_PlaceTrace );
	zoc.flush();
    } else {
	if ( isServer( VOV_SERVER_ENTRY ) ) {
	}
	VovPage& page = workPage();
    
	VovNode node = 0;

	//
	// Relevel the trace.
	// Step 1: give all nodes level -1
	VovContentsGenerator genNodes( &page.nodes, VOV_NODE_OCTMASK );
	while ( octGenerate( &genNodes, &node ) == OCT_OK ) {
	    node.get();
	    node.setLevel( -1 );
	}

	//
	// Step 2: relevel everything again, starting from the sources.
	//
	VovContentsGenerator genPlaces( &page.places, VOV_NODE_OCTMASK );
	while ( octGenerate( &genPlaces, &node ) == OCT_OK ) {
	    node.get();
	    if ( node.hasInputs() == VOV_NOT_OK ) {
		node.levellize( 1 );	// Now place it in the first legal level.
	    }
	    node.doPlacement();
	}
	VovContentsGenerator genTransitions( &page.transitions, VOV_NODE_OCTMASK );
	while ( octGenerate( &genTransitions, &node ) == OCT_OK ) {
	    node.get();
	    node.doPlacement();
	}
	
	// A few passes of pairwise interchanges.
	VovPlacementMatrix* matrix = node.getOrAddPlacementMatrix( 0 );
	long startDate = getDate();
	for (int i = 0 ; i < 16; i++ ) { 	//  Do 16 rows at a time if there is time.
	    matrix->pairwiseInterchanges();
	    long  et = getDate() - startDate;
	    // vovMessage( form( "Elapsed Time %d", et ) );
	    if ( et > 8 ) {
		vovMessage( form("pairwiseInterchanges take too long: breaking out of loop after %d rows",i+1 ) );
		break;
	    }
	}

	if ( isServer( VOV_SERVER_EXIT ) ) {
	    VS_ok( zoc );
	}
    }
    if ( isClient() ) {
	VC_wait_for_ok_reply( zoc );
    }
    
}









