#include "place.h"
#include "transition.h"
#include "trace.h"
#include "path.h"
#include "voverror.h"
#include "travfunc.h"

extern String vov_normalizeString( const String& line );
extern void vovUnpackDefaults( VovObject* f, char* m, char* s);
extern octStatus vovUnpackFacetName( VovObject *facet,  char* name);

extern int vov_gsub( VovProp& prop, const String& os, const  String& ns );

extern int unix_info( char*, int*, int*, Date*);
extern int unix_cp( char*, char* );
extern void unix_chmod( char*, int );
extern int facet_info( char*, int*, int*, Date*);
extern int facet_cp( char*, char* );
extern void facet_chmod( char*, int );


int chain_debug = 0;

VovPlace::VovPlace()
    :
    name( "NAME", OCT_STRING ),
    placeType( "TYPE", OCT_INTEGER ), 
    host( "HOST", OCT_STRING ),
    value( "VALUE", OCT_STRING ),
    timeStamp( "TIMESTAMP", OCT_INTEGER ),
    current( "CURRENT", OCT_INTEGER ),
    goalBag("GOALS"),
    measureBag("MEASURES")
{
}


VovPlace::VovPlace( VovPlaceType t, const String& nm )
    : 
    ( getCellName( t ), VOV_PLACE ),
    name( "NAME", OCT_STRING ),
    placeType( "TYPE", OCT_INTEGER ), 
    host( "HOST", OCT_STRING ),
    value( "VALUE", OCT_STRING ),
    timeStamp( "TIMESTAMP", OCT_INTEGER ),
    current( "CURRENT", OCT_INTEGER ),

    goalBag("GOALS"),
    measureBag("MEASURES")
{
    placeType = t;
    
    switch ( t ) {
    case VOV_PROGRAM_OPTION:
    case VOV_BOOLEAN:
    case VOV_EXIT_STATUS:
    case VOV_MEASURE:
	timeStamp = getDate();	// These places have the date of creation.
	break;
    default:
	timeStamp = 0;
	break;
    } 
    current = 1;
    host = "";
    value = "";
    setName( nm );
}

VovPlace::~VovPlace()
{
}


void VovPlace::setName( String& nm )
{
    switch ( getPlaceType() ) {
    case VOV_EXIT_STATUS:
    case VOV_MEASURE:
    case VOV_PROGRAM_OPTION:
	name = nm;
	break;
    default:
	name = nm;
    }
    // Give a short name to the instance representing the node.
    
    if ( nm.contains( "/" ) ) {
	String shortName = nm.after( "/", -1 );
	VovNode::setName( shortName );
    } else {
	VovNode::setName( name.string() );
    }
    canonicalize();
}

#ifndef VOVCLIENT
VovPlace::VovPlace( octId i )
    :
    ( i ),
    name( "NAME", OCT_STRING ),
    placeType( "TYPE", OCT_INTEGER ), 
    host( "HOST", OCT_STRING ),
    value( "VALUE", OCT_STRING ),
    timeStamp( "TIMESTAMP", OCT_INTEGER ),
    current( "CURRENT", OCT_INTEGER ),
    goalBag("GOALS"),
    measureBag("MEASURES")

{
    if ( id() ) {
	get();
    }
}
#endif

#ifndef VOVCLIENT
VovPlace::VovPlace( VovNode& node )
    :
    ( node.id() ),
    name( "NAME", OCT_STRING ),
    placeType( "TYPE", OCT_INTEGER ), 
    host( "HOST", OCT_STRING ),
    value( "VALUE", OCT_STRING ),
    timeStamp( "TIMESTAMP", OCT_INTEGER ),
    current( "CURRENT", OCT_INTEGER ),
    goalBag("GOALS"),
    measureBag("MEASURES")
{
    if ( objectId ) {
	get();
    }
}
#endif

VovPlace::VovPlace(VovPlace& p)
    :
    ( p ),
    name( "NAME", OCT_STRING ),
    placeType( "TYPE", OCT_INTEGER ), 
    host( "HOST", OCT_STRING ),
    value( "VALUE", OCT_STRING ),
    timeStamp( "TIMESTAMP", OCT_INTEGER ),
    current( "CURRENT", OCT_INTEGER ),
    goalBag("GOALS"),
    measureBag("MEASURES")
{
    placeType = p.placeType;
    name = p.name;
    timeStamp = p.timeStamp;
    current = p.current;
    host = p.host;
    value = p.value;
}

void VovPlace::operator=( VovPlace& p)
{
    VovNode::operator=( p );

    name = p.name;
    placeType = p.placeType;
    host = p.host;
    timeStamp = p.timeStamp;
    current = p.current;
    value = p.value;
}


#ifndef VOVCLIENT
VovStatus VovPlace::create(  )
{
    VovNode::create( );
    OH_ASSERT( octCreate( this, &name ) );
    OH_ASSERT( octCreate( this, &placeType ) );
    OH_ASSERT( octCreate( this, &host ) );
    OH_ASSERT( octCreate( this, &value ) );
    OH_ASSERT( octCreate( this, &timeStamp ) );
    OH_ASSERT( octCreate( this, &current ) );

    int existflag;
    Date timestamp ;
    int mode;
    info( &existflag, &mode, &timestamp, 1);
    
    if (existflag) {
	setStatus( VOV_TRACE_NODE );
    } else {
	setStatus( VOV_PLAN_NODE );
    }
    vov_place_find_add( *this );
    return VOV_OK;
}
#endif



#ifndef VOVCLIENT
VovStatus VovPlace::get(  )
{
    VovStatus rtn = VOV_ERROR;
    if ( VovNode::get( ) == VOV_OK ) {
	if ( octGetByName( this, &placeType ) == OCT_OK ) {
	    OH_ASSERT( octGetByName( this, &name ) );
	    OH_ASSERT( octGetByName( this, &host ) );
	    OH_ASSERT( octGetOrCreate( this, &value ) ); 	// Create also, for backward compatibility.
	    OH_ASSERT( octGetByName( this, &timeStamp ) );
	    OH_ASSERT( octGetByName( this, &current ) );
	    rtn = VOV_OK;
	}
    }
    if ( rtn == VOV_ERROR ) {
	vovError( form( "Problems getting %s", ohFormatName( this ) ) );
	id() = 0;
    }
    return rtn;
}
#endif

#ifndef VOVCLIENT
VovStatus VovPlace::modify(  )
{
    VovPlace oldPlace = id();
    vov_place_find_del( oldPlace );

    VovNode::modify( );
    OH_ASSERT( octModify( &name ) );
    OH_ASSERT( octModify( &placeType ) );
    OH_ASSERT( octModify( &host ) );
    OH_ASSERT( octModify( &timeStamp ) );
    OH_ASSERT( octModify( &current ) );
    OH_ASSERT( octModify( &value ) );

    vov_place_find_add( *this );
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovPlace::del(  )
{
    vov_place_find_del( *this );
    OH_ASSERT( octDelete( &name ) );
    OH_ASSERT( octDelete( &placeType ) );
    OH_ASSERT( octDelete( &host ) );
    OH_ASSERT( octDelete( &value ) );
    OH_ASSERT( octDelete( &timeStamp ) );
    OH_ASSERT( octDelete( &current ) );
    VovNode::del( );
    return VOV_OK;
}
#endif






// CHAINS OF PLACES
// These routines deal with the case in which tools run 'in place'
// i.e. tools which modify their inputs.
// These tools can only be executed in sequence.
// 
#ifndef VOVCLIENT
VovStatus VovPlace::isChain( )
{
    VovObject chain(VOV_CHAIN_OCTMASK);		
    if ( octGenFirstContainer( this, VOV_CHAIN_OCTMASK , &chain ) == OCT_OK ) {
	return VOV_OK;		
    } 
    return VOV_ERROR;
}
#endif

#ifndef VOVCLIENT
VovStatus VovPlace::getChain( VovChainInfo& chain )
{
    if ( octGenFirstContainer( this, VOV_CHAIN_OCTMASK , &chain ) == OCT_OK ) {
	VOV_ASSERT( chain.get() );
	return VOV_OK;		
    } else {
	return VOV_ERROR;
    }
}
#endif


#ifndef VOVCLIENT
VovStatus VovPlace::duplicate( VovTransition& tran, VovPlace& newPlace, int direction = VOV_FORWARD )
{
    vovMessage( "Chain split: " + format()  );
    
    VovChainInfo chain;
    if ( getChain( chain ) == VOV_OK ) {
	chain.addPlace( newPlace );
	chain.setCurrent( newPlace );
    } else {
	VovObject facet;						// 
	octGetFacet( this, &facet );					// 
	String chainname = getCommonName();				// Use a short name for the chain.
	vovMessage( "chain name is " + chainname );
	OH_ASSERT( ohCreateNet( &facet, &chain, (char*)chainname ) );	// 
	VOV_ASSERT( chain.create( *this, newPlace ) );			// 
	vovMessage( "Created: " + chain.format() );
    }
    //
    // Move the connections around: all outputs of this place become outputs of teh newplace.
    // 
    if ( direction == VOV_FORWARD ) {
	VovContentsGenerator genTrans( this, VOV_TRANSITION_OCTMASK );
	VovTransition t;
	while ( octGenerate( &genTrans, &t ) == OCT_OK ) {
	    if ( t.id() != tran.id() ) {
		t.get();
		VOV_ASSERT( disconnectOutput( t ) );
		VOV_ASSERT( newPlace.connectOutput( t ) );
	    }
	}
	VOV_ASSERT( tran.connectInput( *this ) );
	VOV_ASSERT( tran.connectOutput( newPlace ) );
    } else {
	vovMessage( "Duplicate backwards" );
	VovContainersGenerator genTrans( this, VOV_TRANSITION_OCTMASK );
	VovTransition t;
	while ( octGenerate( &genTrans, &t ) == OCT_OK ) {
	    if ( t.id() != tran.id() ) {
		t.get();
		VOV_ASSERT( disconnectInput( t ) );
		VOV_ASSERT( newPlace.connectInput( t ) );
	    }
	}
	VOV_ASSERT( tran.connectInput( newPlace ) );
	VOV_ASSERT( tran.connectOutput( *this ) );
    }
    VOV_ASSERT(  chain.OK() );	// Just for debugging.
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovPlace::merge( VovTransition& tran, VovPlace& place, int direction = VOV_FORWARD )
{
    if ( direction == VOV_BACKWARD ) {
	VovStatus s = place.merge( tran, *this, VOV_FORWARD );
	if ( s == VOV_OK ) {
	    *this = place;
	}
	return s;
    }

    VovChainInfo chain;
    if ( getChain( chain ) != VOV_OK ) {
	vovError( "Cannot merge places not in chain" );
	return VOV_ERROR;
    }

    vovMessage( "Chain merge: " + chain.name() );
    
    if ( chain.OK() != VOV_OK ) {
	vovError( "Paranoid check for chain consistency failed" );
	return VOV_ERROR;
    }

    VovPlace p1;
    if ( chain.getNext( *this, p1 ) != VOV_OK ) {
	vovError( "Cannot get next place in chain" );
	return VOV_ERROR;
    }

    if ( p1.id() != place.id() ) {
	vovError( "Cannot merge two places which are not next to each other." );
	return VOV_ERROR;
    }
    VOV_ASSERT( chain.delPlace( place ) );
    VOV_ASSERT( tran.disconnectInput( *this ) );
    VOV_ASSERT( tran.disconnectOutput( place ) );

    VovContentsGenerator genTrans( &place, VOV_TRANSITION_OCTMASK );
    VovTransition t;
    while ( octGenerate( &genTrans, &t ) == OCT_OK ) {
	VOV_ASSERT( t.get() );
	VOV_ASSERT( place.disconnectOutput( t ) );
	VOV_ASSERT( connectOutput( t ) );
    }
    if ( chain.length() == 1 ) {
	VOV_ASSERT( chain.del() ); 	// Eliminate chain.
	VOV_ASSERT( setCurrent() );	// Make sure this place becomes current.
    } else {				// 
	VOV_ASSERT( chain.OK() );	// This is just for debugging. Should be removed later.
    }
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovPlace::setCurrent( )
{
    if ( current.integer() ) {
	return VOV_OK;				// The place is already current.
    }
    
    VovChainInfo chain;
    if ( getChain( chain ) == VOV_OK ) {
	VOV_ASSERT( chain.setCurrent( *this ) );
    } else {
	current.modify( 1 );
    }
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovPlace::unsetCurrent( )
{
    VovChainInfo chain;
    if ( getChain( chain ) != VOV_OK ) {
	return VOV_OK;		// Place has no chain.
    } 

    VovPlace cp;
    chain.getCurrent( cp );

    if ( cp.id() != id() && chain.precedes( cp, *this ) == VOV_OK ) {
	return VOV_OK;
    } 
    
    if ( chain.getPrec( *this, cp ) == VOV_OK ) {
	chain.setCurrent( cp );
    }

    return VOV_OK;
}
#endif


#ifndef VOVCLIENT
static void correctInChainTest( VovPlace& place )
{
    VovChainInfo chain;						// 
    if ( place.getChain( chain ) == VOV_OK ) {			// Not a chain, nothing to do.
	VovPlace c;						// Again, just for debugging
	chain.getCurrent( c ) ;					// Get the current place in chain.
	if ( c.id() != place.id() ) {				// Check ids
	    vovError( "correctInChainTest() non current:" + place.format() ); // Error, 
	    place.id() = c.id();					  // Fix error
	    VOV_ASSERT( place.get() );				  // 
	}							  // 
    }								  // 
}

VovStatus VovPlace::correctInputInChain( VovTrace& trace, VovTransition& tran )
{
    correctInChainTest( *this ); 		// PARANOID check.
    
    int getPrecedentFlag = 0;
    if ( tran.isOutput( *this ) == VOV_OK ) {
	return VOV_OK;				//  If place is already an output, then ignore input declarations.
    } 
    if ( getStatus() == VOV_RETRACING_NODE ) {
	getPrecedentFlag = 1;
    }
    
    
    VovChainInfo chain;					  	  // 
    if ( getPrecedentFlag ) {					  // 
	VovPlace tmpPlace = id();				  // Then, we need to duplicate the place.
	if ( getChain( chain) == VOV_OK  ) {			  // If it is a chain, get prec.
	    if ( chain.getPrec( tmpPlace, *this ) == VOV_OK ) {	  // The place already has a precedent in chain
		chain.position( *this );
		return VOV_OK;					  // Then we are done.
	    }							  // 
	}							  // 
	VOV_ASSERT( trace.create( tmpPlace ) );			  // Create the precedent.
	VOV_ASSERT( duplicate( tran, tmpPlace, VOV_BACKWARD ) );  // Duplicate place in chain.
	tmpPlace.setStatus( VOV_PLAN_NODE );			  // New place is not valid.
	operator = ( tmpPlace );				  // 
    } else {							  // Place is not already an output and it is not retracing.
	if ( getChain( chain ) == VOV_OK ) {			  // 
	    octId placeid = id();				  // Save id to check if different from last.
	    chain.getLast( *this );				  // Get last place in chain.
	    if ( chain_debug ) {
		if ( id() != placeid ) {				  // 
		    vovMessage("Selecting last place in chain");	  // Only for debugging.
		}			
	    }				  // 
	    chain.position( *this );
	}							  // 
    }								  // 
    
    return VOV_OK;						  // 
}

VovStatus VovPlace::correctOutputInChain( VovTrace& trace, VovTransition& tran )
{
    correctInChainTest( *this ); 				  // PARANOID.
    VovChainInfo chain;						  // 
    if ( getChain( chain ) == VOV_OK ) {			  // 
	VovPlace p1,p2;						  // The place is already a chain.
	chain.getFirst( p1 );
	while (1) {
	    if ( tran.isInput( p1 ) == VOV_OK ) {
		if ( chain.getNext( p1, p2 ) != VOV_OK ) { 	// If there is no next, make it.
		    p2.id() = id();				// 
		    p2.get();					// Duplicate the node.
		    VOV_ASSERT( trace.create( p2 ) );		// 
		    VOV_ASSERT(p1.duplicate(tran,p2,VOV_FORWARD));	// The place is both an input and an output.
		}
		operator = ( p2 );
		chain.position( *this );
		return VOV_OK;
	    } 
	    if ( chain.getNext( p1, p2 ) == VOV_OK ) {
		p1 = p2;
	    } else {
		// It is a chain, but no place in chain is an input.
		if ( chain_debug ) {
		    vovMessage( chain.format() );
		    vovMessage( "Output is chain, but chain is not an input, doing nothing." );
		}
		chain.position( *this );
		return VOV_OK;
	    }
	}
    } else {							// Not a chain.
	if ( tran.isInput( *this ) == VOV_OK ) {		// Check for transition running 'in place'.
	    VovPlace newplace = id();				// Duplicate the node.
	    VOV_ASSERT( trace.create( newplace ) );		// 
	    VOV_ASSERT(duplicate(tran,newplace,VOV_FORWARD));	// The place is both an input and an output.
	    operator = ( newplace );
	    VOV_ASSERT( getChain( chain ) );
	    chain.position( *this );
	}
    }
    return VOV_OK;
}
#endif

String VovPlace::format( )
{
    String s = getName();
    if ( getPlaceType() == VOV_MEASURE || getPlaceType() == VOV_PROGRAM_OPTION ) {
	s += " " + value.format();
    }
    if ( ! current.integer() ) {
	s += "(CH-)";	// Not current.
    }
    return s ;
}


#ifndef VOVCLIENT
VovStatus VovPlace::match(VovPlace& p)
    // Returns VOV_OK if there is a match, VOV_NOT_OK otherwise.
{
    if ( ! p.current.integer() ) 			// 
      return VOV_NOT_OK;				// No match with non current place.
    if (getPlaceType() != p.getPlaceType() )		// Different types ?
      return VOV_NOT_OK;				//  No match.
    if ( getName() != p.getName() )			// Different names ?
      return VOV_NOT_OK;				//  No match.
    return VOV_OK;					// Then, this is a match.
}
#endif





VovStatus VovPlace::copy( VovPlace& place )
{
    switch ( getPlaceType() ) {
    case VOV_EXECUTABLE:
	vovError( "Cannot copy executables. (does not seem to make sense.)" );
	break;
    case VOV_UNIX_FILE:
	if ( unix_cp( (char*)place.getName(), (char*)getName() ) < 0 ) {
	    // vovError( "Copy file failed" );
	    return VOV_ERROR;
	}
	break;
    case VOV_OCT_FACET:
	if ( facet_cp( (char*)place.getName(), (char*)getName() ) < 0 ) {
	    // vovError("Copy facet failed" );
	    return VOV_ERROR;
	}
	break;
    case VOV_PROGRAM_OPTION:
    case VOV_BOOLEAN:
    case VOV_EXIT_STATUS:
    case VOV_MEASURE:
	break;
    default:
	vovFatalError("Invalid place type in vovPlace::copy");
	break;
    }
    return VOV_OK;
}

#ifndef VOVCLIENT
VovStatus VovPlace::mvlib( String& oldpath, String& newpath )
    // Return: VOV_OK if changed, VOV_NOT_OK if no change.
{
    int changes = 0;
    changes += vov_gsub( name, oldpath, newpath );
    if ( changes > 0 ) {
	setName( name.string() );
	VOV_ASSERT( modify() );
    }

    return  (changes) ? VOV_OK : VOV_NOT_OK;
}
#endif


#ifndef VOVCLIENT
VovStatus vov_place_exists(VovPlace& p, void* d)
{
    if ( d ) ;			/* avoid complaint from the compiler */
    return p.exists();
}
#endif


#ifndef VOVCLIENT
VovStatus VovPlace::exists()
{
    int existflag;
    int mode;
    Date dummyTime(0);		// discarded info.

    info( &existflag, &mode, &dummyTime, 0);
    return existflag ? VOV_OK : VOV_NOT_OK;
}
#endif

#ifndef VOVCLIENT
int VovPlace::getProtection()
{
    int existflag;
    int mode;
    Date dummyTime(0);		// discarded info.

    info( &existflag, &mode, &dummyTime, 0);
    return existflag ? mode : -1;
}
#endif

void VovPlace::setProtection( int p )
{
    String fullpath ;		// Used with ASCII and facets.

    switch ( getPlaceType() ) {
    case VOV_EXECUTABLE:
    case VOV_UNIX_FILE:
	fullpath = getName();
	unix_chmod( (char*)fullpath, p );
	break;
    case VOV_OCT_FACET:
	fullpath = getName();
	facet_chmod( (char*)fullpath, p );
	break;
    case VOV_PROGRAM_OPTION:
    case VOV_BOOLEAN:
    case VOV_EXIT_STATUS:
    case VOV_MEASURE:
	break;
    default:
	vovFatalError("setProtection");
	break;
    }
}


Date VovPlace::getTimeStamp()
{
    Date date = timeStamp.integer();
    date.set( host.string() );
    return date;
}


VovStatus VovPlace::changeTransition( VovTransition& tran, int flag  )
    // Return: VOV_CHANGE, VOV_NO_CHANGE, VOV_ERROR.
{
    VovStatus rtn = VOV_NO_CHANGE;
    if ( getPlaceType() == VOV_PROGRAM_OPTION ) {
	String cmd = " " + tran.getCommandLine() + " ";
	String n = " ${" + getCommonName() + "} ";
	String v = " " + value.string() + " ";
	int subs = 0;		// Number of substitutions.
	vovMessage( cmd );
	vovMessage( form("`%s' `%s'", (char*)n, (char*)v ) );
	if ( flag == 0 ) {
	    subs = cmd.gsub( v, n );
	} else {
	    subs = cmd.gsub( n, v );
	}
	if ( subs ) {
	    vovMessage( "Changed cmd to: " + cmd );
	    tran.setCommandLine( cmd );
	    rtn = VOV_CHANGE;
	} else {
	    rtn = VOV_ERROR;	// Option does not exist.
	}
    }
    return rtn;
}

#ifndef VOVCLIENT
VovStatus  VovPlace::info( int* existflag, int* mode, Date* cur_timestamp, int modifyflag )
{
    if ( ! current.integer() ) {
	VovChainInfo chain;
	if ( getChain( chain ) == VOV_OK )  {
	    VovPlace cp = 0;	// Current place in chain.
	    chain.getCurrent( cp );
	    //
	    // Return the current information, as if the place had not changed.
	    // However, update the information with the info of the current place
	    // in the chain.
	    //
	    *cur_timestamp = getTimeStamp();
	    
	    Date dummyDate;		// 

	    VovStatus ret = cp.info( existflag, mode, &dummyDate, modifyflag );
	    if ( ret == VOV_OK ) {
		timeStamp.modify( dummyDate.get() );
		host.modify( dummyDate.getHost() );
	    }
	    return ret;
	} else {
	    // No current place, no chain.
	    return VOV_ERROR;
	}
    }

    String fullpath ;		// Used with ASCII and facets.

    switch ( getPlaceType() ) {
    case VOV_EXECUTABLE:
    case VOV_UNIX_FILE:
	fullpath = getName();
	(void)unix_info( (char*)fullpath, existflag, mode, cur_timestamp );
	break;
    case VOV_OCT_FACET:
	fullpath = getName();
	(void)facet_info( (char*)fullpath, existflag, mode, cur_timestamp );
	break;
    case VOV_PROGRAM_OPTION:
    case VOV_BOOLEAN:
    case VOV_EXIT_STATUS:
    case VOV_MEASURE:
	*existflag = 1;
	*cur_timestamp = getTimeStamp();
	*mode = 0777;
	break;
    default:
	vovFatalError("Invalid place type in vovPlace::info");
	break;
    }

    if ( *existflag && modifyflag ) {
	if (*cur_timestamp < getTimeStamp() ) {
	    vovMessage( getCommonName() + ":Timestamp got smaller:" 
		       + getTimeStamp().format() + "->" + cur_timestamp->format() );
	}
	// if ( *cur_timestamp > getTimeStamp() ) {
	// vovMessage( getCommonName() + ":\n"+ 
	//    getTimeStamp().format() +" -> "+ cur_timestamp->format() );
	//   }
	timeStamp.modify( cur_timestamp->get() );
	host.modify( cur_timestamp->getHost() );
    }

    if ( ! existflag ) {
	return VOV_ERROR;
    } else {
	return VOV_OK;
    }
}
#endif



String VovPlace::expandedName( )
{
    return name.string();
#ifdef OLD
    switch ( getPlaceType() ) {
    case VOV_EXECUTABLE:
    case VOV_UNIX_FILE:
    case VOV_OCT_FACET:
	 return path_expand( name.string() );
    default:
	return name.string();
    }
#endif
}

String  VovPlace::getCommonName( )
{
    switch ( getPlaceType() ) {
    case VOV_UNIX_FILE:
    case VOV_EXECUTABLE:
    case VOV_OCT_FACET:
    case VOV_PROGRAM_OPTION:
	return path_lastComponent( getName() );
    case VOV_BOOLEAN:
    case VOV_EXIT_STATUS:
    case VOV_MEASURE:
	return getName();
    default:
	vovError("Invalid place type for " + format() );
	return "$$INVALID_PLACE_TYPE$$";
    }
}

String vov_completeOctName( const String& in )
{
    // Use the oh package to this this simple manipulation.
    VovObject facet( OCT_FACET ); 
    vovUnpackDefaults( &facet, "", "cell:view:contents" );
    OH_ASSERT( vovUnpackFacetName( &facet, (char*)in ) );

    String out = form( "%s:%s:%s",
		      facet.contents.facet.cell,
		      facet.contents.facet.view,
		      facet.contents.facet.facet );
    return out;
}


void VovPlace::canonicalize( )
{
    switch ( getPlaceType() ) {
    case VOV_OCT_FACET:
	{
	    String currentName = getName();
	    String canonicalName =  vov_completeOctName( currentName );
	    String cell = canonicalName.before(":");
	    String viewfacet = canonicalName.after(":");
	    canonicalName = path_canonicalize( cell ) + ":" + viewfacet;
	    if ( currentName != canonicalName) {
		name = utcc_strsav_and_track( (char*)canonicalName );
	    }
	}
	break;
    case VOV_UNIX_FILE:
    case VOV_EXECUTABLE:
	{
	    String currentName = getName();
	    String canonicalName = path_canonicalize( currentName );
	    if ( currentName != canonicalName) {
		name = utcc_strsav_and_track( (char*)canonicalName );
	    }
	}
	break;
    case VOV_PROGRAM_OPTION:
    case VOV_BOOLEAN:
    case VOV_EXIT_STATUS:
    case VOV_MEASURE:
    default:
	break;
    }
}

#ifndef VOVCLIENT
VovStatus VovPlace::freeResources( )
{
    extern int unix_delete( char* );
    extern int facet_delete( char* );
    VovStatus rtn = VOV_OK;

    String name = getName();
    switch ( getPlaceType() ) {
    case VOV_EXECUTABLE:
    case VOV_PROGRAM_OPTION:
    case VOV_EXIT_STATUS:
    case VOV_BOOLEAN:
    case VOV_MEASURE:
	// No resources associated with these places.
	break;
    case VOV_UNIX_FILE:
	rtn =  unix_delete( (char*)getName() ) ? VOV_OK : VOV_ERROR;
	break;
    case VOV_OCT_FACET:
	rtn =  facet_delete( (char*)getName() ) ? VOV_OK : VOV_ERROR;
	break;
    default:
	vovError("Invalid place type");
	break;
    }
    if ( rtn == VOV_OK ) {
	// Is this the right status to use ? I am not sure.
	setStatus( VOV_NONEXISTING_NODE );
    }
    return rtn;
}
#endif



#ifdef OBSOLETE
VovStatus VovPlace::addGoal( VovGoal& goal )
{
    OH_ASSERT(octGetOrCreate( this, &goalBag ));
    OH_ASSERT(octAttachOnce( &goalBag, &goal ));
    return VOV_OK;
}
#endif

#ifdef OBSOLETE
VovStatus VovPlace::scheduleGoals()
{
    if ( octGetByName( this, &goalBag ) != OCT_OK ) {
	return VOV_OK;		// No goal defined for this place.
    }
    VovContentsGenerator genGoal( &goalBag, OCT_INSTANCE_MASK );
    VovGoal goal;
    while ( octGenerate( &genGoal, &goal ) == OCT_OK ) {
	VOV_ASSERT( goal.get() );
	VOV_ASSERT( goal.schedule( *this ) ); 
    }
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovPlace::updateMeasure( VovPlace& measure )
{
    if ( measure.getPlaceType() != VOV_MEASURE ) {
	vovError( "Measure is not of type VOV_MEASURE" );
	return VOV_ERROR;
    }
    OH_ASSERT( octGetOrCreate( this, &measureBag ) );
    measureBag.add( measure );
    vovMessage( format() + "\n\t" + measure.format() );
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
String VovPlace::formatMeasures()
{
    String s;
    if (octGetByName( this, &measureBag ) == OCT_OK ) {
	VovContentsGenerator genMeasure( &measureBag, VOV_PLACE_OCTMASK );
	VovPlace measure;
	while ( octGenerate( &genMeasure, &measure ) == OCT_OK ) {
	    VOV_ASSERT( measure.get() );
	    if ( measure.getStatus() == VOV_TRACE_NODE ) {
		s += "\t" + measure.format() + "\n";
	    }
	}
    }
    return s;
}
#endif




