#include "builtin.h"
#include "String.h"
#include "placematrix.h"
#include "vovinternal.h"
#include "node.h"
#include "voverror.h"

extern octPoint octToMatrix( octPoint );
extern octPoint matrixToOct( octPoint );
VovPlacementMatrix::VovPlacementMatrix( int rows, int cols )
{
    size_cols = -1;
    size_rows = -1;
    matrix = 0;
    resize( rows, cols );
    OK();
}

VovPlacementMatrix::~VovPlacementMatrix( )
{
}
void VovPlacementMatrix::resize( int rows, int cols )
{
    int tmp_size_cols = size_cols >? 16;				// Make it at least 16.
    int tmp_size_rows = size_rows >? 16;				// Same.
    while ( tmp_size_rows < rows ) tmp_size_rows <<= 1;			// 
    while ( tmp_size_cols < cols ) tmp_size_cols <<= 1;			// 
    if ( tmp_size_cols == size_cols && tmp_size_rows == size_rows ) return;	// Do not resize;
    

    // vovMessage(form( "resizing matrix: %dx%d to %dx%d", size_rows, size_cols, tmp_size_rows, tmp_size_cols ));
    int** tmp_matrix = new int*[tmp_size_rows];
    for ( int row=0; row < tmp_size_rows; row++ ) {
	tmp_matrix[row] = new int[tmp_size_cols];
    }
    int* tmp_row_first   = new int[tmp_size_rows];
    int* tmp_row_last    = new int[tmp_size_rows];
    int* tmp_row_entries = new int[tmp_size_rows];

    // Clear all the entries now. Proceed by rows.
    for ( int r = 0 ; r < tmp_size_rows ; r++ ) {
	if ( r < size_rows ) {
	    tmp_row_first[r]   = row_first[r];
	    tmp_row_last[r]    = row_last[r];
	    tmp_row_entries[r] = row_entries[r];
	} else {
	    tmp_row_first[r]   = 0;
	    tmp_row_last[r]    = 0;
	    tmp_row_entries[r] = 0;
	}
	for ( int c = 0 ; c < tmp_size_cols ; c++ ) {
	    if ( c < size_cols &&  r < size_rows ) {
		tmp_matrix[r][c] = matrix[r][c];
	    } else  {
		tmp_matrix[r][c] = 0;
	    }
	}
    }
    if ( matrix ) {
	for ( int r = 0; r < size_rows; r++ ) {
	    delete matrix[r];
	}
	delete matrix;
	delete row_first;
	delete row_last;
	delete row_entries;
    } 
    matrix      = tmp_matrix;
    row_first   = tmp_row_first;
    row_last    = tmp_row_last;
    row_entries = tmp_row_entries;
    size_cols   = tmp_size_cols;
    size_rows   = tmp_size_rows;
}

VovStatus VovPlacementMatrix::checkBoundaries( int row, int  col )
{
    if ( col < 0 || col >= size_cols || row < 0 || row >= size_rows ) {
	vovMessage( form("Out of boundary: %d %d.", row, col ) );
	return VOV_ERROR;
    }
    return VOV_OK;
}

int VovPlacementMatrix::getFirstSpot( int row )
{
    if ( checkBoundaries( row, 0 ) != VOV_OK ) return 0; // Not something I expect; in any case, inside the array.
    for ( int c = 0 ; c < size_cols ; c++ ) {		 // 
	if ( matrix[row][c] == 0 ) return c;		 // 
    }							 // 
    resize( size_rows, size_cols * 2 );			 // No spot found, row is full. Resize matrix.
    return  row_last[row] + 1;				 // 
}

int VovPlacementMatrix::getClosestSpot( int row, int target, int id )
{
    if ( checkBoundaries( row, target ) != VOV_OK ) {
	return 0;
    } 
    if ( matrix[row][target] == 0 || matrix[row][target] == id ) {
	return target;
    }
    int left, right;
    for ( left = target ; left >= 0 ; left-- ) { 			// Search to left of target.
	if ( matrix[row][left] == 0 || matrix[row][left] == id ) break;	// 
    }									// 
    if ( left == -1 ) left = -99999;					// If not found, move way outside the array.
    for ( right = target ; right < size_cols ; right++ ) {		// Search to the right.
	if ( matrix[row][right] == 0 || matrix[row][right] == id  )	// 
	  break;							// 
    }									// 
    if ( right >= size_cols ) right = size_cols;			// Keep just outside the array, to the right.
    return ( target - left <= right - target ) ? left : right;		// Return spot closets to target.
}

VovStatus VovPlacementMatrix::add( int id, int row, int col )
{
    if ( checkBoundaries( row, col ) == VOV_ERROR ) {
	resize( row + 1, col + 1 );
	if ( checkBoundaries( row, col ) == VOV_ERROR ) {
	    vovError("Resize failed");
	    return VOV_ERROR;
	}
    }
    if ( matrix[row][col] != 0  ) {
	vovError( form( "Matrix[%d][%d] = %d ", col, row, matrix[row][col] ) );
	return VOV_ERROR;	// Spot already taken.
    }
    matrix[row][col] = id;
    if ( row_entries[row]++ == 0 ) {
	row_first[row] = row_last[row] = col;
    } else {
	row_first[row] = row_first[row] <?  col ; 	// MIN
	row_last[row]  = row_last[row]  >?  col ;	// MAX
    }
    return VOV_OK;
}

VovStatus VovPlacementMatrix::del( int id, int row, int col )
{
    if ( checkBoundaries( row, col ) == VOV_ERROR ) return VOV_ERROR;
    if ( matrix[row][col] != id ) {
	vovMessage("error in vovplacementmatrix::del()");
    } 
    matrix[row][col] = 0;
    row_entries[row]--;
	
    if ( row_entries[row] == 0 ) {
	row_last[row] = row_first[row] = 0;
	return VOV_OK;
    }

    if ( row_first[row] == col ) {
	for ( int c = col+1 ; c <= row_last[row]; c++ ) {
	    if ( matrix[row][c]) {
		row_first[row] = c;
		break;
	    }
	}
    }
    if ( row_last[row] == col ) {
	for ( int c = col-1 ; c >= row_first[row]; c-- ) {
	    if ( matrix[row][c]) {
		row_last[row] = c;
		break;
	    }
	}
    }
    return VOV_OK;
}

VovStatus VovPlacementMatrix::verify( int id, int row, int col )
{
    if ( checkBoundaries( row, col ) == VOV_ERROR ) return VOV_ERROR;
    return ( matrix[row][col] == id ) ? VOV_OK : VOV_ERROR;
}


int nextHigherPowerOfTwo( int n )
{
    int N = 1;
    while ( N < n ) N <<= 1;
    // cout.form( "nextHigherPowerOfTwo: %d ->> %d\n", n, N );
    return N;
}

VovStatus VovPlacementMatrix::init( VovNodeSet& nodes )
{
    // Compute initial matrix size.
    {
	VovNode node;
	node.getOrAddPlacementMatrix( this );
	VovContentsGenerator genNode( &nodes, VOV_NODE_OCTMASK );
	const int LARGEINT = ( 1 << 31 ) - 1;
	octPoint ur = { -LARGEINT, -LARGEINT };
	octPoint ll = { LARGEINT, LARGEINT };
	int count = 0;
	while ( octGenerate( &genNode, &node ) == OCT_OK ) {
	    ur.x = ur.x >? node.contents.instance.transform.translation.x;
	    ur.y = ur.y >? node.contents.instance.transform.translation.y;
	    ll.x = ll.x <? node.contents.instance.transform.translation.x;
	    ll.y = ll.y <? node.contents.instance.transform.translation.y;
	    count++;
	}
	// vovMessage( form("%d nodes for matrix initialization.", count ) );
	if ( count ) {
	    extern octPoint octToMatrix( octPoint );
	    ur = octToMatrix( ur ); 	// 
	    ll = octToMatrix( ll );	// 
	    int approx_cols = ur.x - ll.x;	// 
	    int approx_rows = ll.y - ur.y; 	// Because Y axis is inverted in matrix coordinates.
	    // Convert to next higher power of two.
	    approx_cols = nextHigherPowerOfTwo( approx_cols );
	    approx_rows = nextHigherPowerOfTwo( approx_rows );
	    resize( approx_rows + 1, approx_cols + 1 );
	} 
    } 
    // Place nodes in the right spot.
    {
	VovContentsGenerator genNode( &nodes, VOV_NODE_OCTMASK );
	VovNode node;
	while ( octGenerate( &genNode, &node ) == OCT_OK ) {
	    octPoint coord = octToMatrix( node.contents.instance.transform.translation );
	    int row = coord.y;
	    int col = coord.x;
	    if ( add( node.id(), row, col ) == VOV_ERROR ) {
		vovMessage( "Placement conflict: must find another spot." );
		VOV_ASSERT( node.get() );
		node.doPlacement();
	    }
	}
    }
    OK();
    return VOV_OK;
}


VovStatus VovPlacementMatrix::OK()
{
    // vovMessage(form("MATRIX: %d x %d", size_cols , size_rows ));
    for ( int row = 0 ; row < size_rows ; row++) {
	int f = -1;
	int l = -1;
	int count = 0;
	for ( int i = 0 ; i < size_cols ; i++ ) {
	    if ( matrix[row][i] ) {
		count++;
		if ( f == -1 ) f = i;
		l = i;
	    }
	}
	if ( row_entries[row] != count ) {
	    vovError( form( "Wrong row count: %d -> %d", row_entries[row], count ) );
	    row_entries[row] = count;
	}
	if ( row_entries[row] ) {
	    if ( row_first[row] != f ) {
		vovError( form( "Wrong row first: %d -> %d", row_first[row], f ) );
		row_first[row] = f;
	    }
	    if ( row_last[row] != l ) {
		vovError( form( "Wrong row last: %d -> %d", row_last[row], l ) );
		row_last[row] = l;
	    }
	    // cerr.form("Row %3d: entries: %3d first: %3d last %3d\n", row, row_entries[row], row_first[row], row_last[row]);
	}
    }
    return VOV_OK;
}

// This routine could take a tremendous amount of time.
// It improves only one row at a time.
int VovPlacementMatrix::pairwiseInterchanges(  )

{
    static int row = 0;
    int interchanges = 0;
    // OK();
    // vovMessage( "Pairwise interchanges:start" );
    VovNode n1,n2,n; 					// Allocate outside loop (might save time).
    while ( row_entries[row] < 2 ) {			// Find a row which has at least two nodes.
	row++;						// 
	if ( row >= size_rows ) {			// Wrap around.
	    row = 0;					// 
	    return 0;					// Exit right away when wrapping around.
	}
    }
    
    // Find two adjacent nodes in this row.
    for ( int col = row_first[row]; col < row_last[row] ; col++ ) {
	int col2 = col + 1;
	if ( matrix[row][col] == 0 || matrix[row][col2] == 0 ) continue;
	n1.id() = matrix[row][col];
	n2.id() = matrix[row][col2];
	OH_ASSERT( n1.getById( VOV_NODE_OCTMASK ) );
	OH_ASSERT( n2.getById( VOV_NODE_OCTMASK ) );
	int cost_old = 0;
	int cost_new = 0;
	VovContainersGenerator genIn1( &n1, VOV_NODE_OCTMASK ); // These generators might have to be
	VovContainersGenerator genIn2( &n2, VOV_NODE_OCTMASK ); //  allocated outside the loops
	VovContentsGenerator genOut1( &n1, VOV_NODE_OCTMASK );
	VovContentsGenerator genOut2( &n2, VOV_NODE_OCTMASK );
	
	int x1 = n1.contents.instance.transform.translation.x;
	int x2 = n2.contents.instance.transform.translation.x;
	
	while (octGenerate( &genIn1, &n ) == OCT_OK ) {
	    int x = n.contents.instance.transform.translation.x;
	    cost_old += abs( x - x1 );
	    cost_new += abs( x - x2 );
	}
	while (octGenerate( &genIn2, &n ) == OCT_OK ) {
	    int x = n.contents.instance.transform.translation.x;
	    cost_old += abs( x - x2 );
	    cost_new += abs( x - x1 );
	}
	while (octGenerate( &genOut1, &n ) == OCT_OK ) {
	    int x = n.contents.instance.transform.translation.x;
	    cost_old += abs( x - x1 );
	    cost_new += abs( x - x2 );
	}
	while (octGenerate( &genOut2, &n ) == OCT_OK ) {
	    int x = n.contents.instance.transform.translation.x;
	    cost_old += abs( x - x2 );
	    cost_new += abs( x - x1 );
	}
	if ( cost_old > cost_new ) {
	    //vovMessage( form("Cost: old %d new %d", cost_old, cost_new ) );
	    VOV_ASSERT( n1.get() );
	    VOV_ASSERT( n2.get() );
	    octPoint p1 = octToMatrix( n1.contents.instance.transform.translation );
	    octPoint p2 = octToMatrix( n2.contents.instance.transform.translation );
	    del( n1.id(), p1.y , p1.x );
	    del( n2.id(), p2.y , p2.x );
	    add( n1.id(), p2.y , p2.x );
	    add( n2.id(), p1.y , p1.x );
	    n1.move( matrixToOct( p2 ) );
	    n2.move( matrixToOct( p1 ) );
	    interchanges++;
	}
    }
    // vovMessage( form( "%d interchanges in row %d.", interchanges, row ) );
    row++;			// Prepare index for next row.
    return interchanges;
}

