#include "trace.h"
#include "travfunc.h"
#include "user.h"
#include "VOVprotocol.h"
#include "place.h"
#include "transition.h"
#include "annotation.h"
#include "retrace.h"
#include "path.h" 
#include "slaveset.h"

#include "voverror.h"
#include "comm.h"

extern "C"
{
/* #include "timer.h"   */
}

extern VovStatus vov_cycle( VovNode&, VovNode&);
extern void vovEditSet( VovTrace& trace, VovNodeSet& set );



VovTrace::VovTrace()
{
    serverEntryCount = 0;		//
    projectName = 0;			// 
    hostName = 0;			// 
    readonly = 0;			// 
    type = VOV_TRACE_TYPE_DISABLED;	// Default.
    zocketPointer = 0;			// 
    numberOfPages = 0;			// 
    pageArray = 0;
}

VovTrace::VovTrace( String& project, String& host, String& mode, int t )
{
    serverEntryCount = 1;	

    VovStatus s = open( project, host, mode, t );
    if ( s != VOV_OK ) {
	type = VOV_TRACE_TYPE_DISABLED;	// Disable the trace. Do not do anything.
	if ( isClient() ) {
	    vovError(form("Cannot connect with VOV server: %s@%s", (char*)project, (char*)host));
	} else {
	    vovError("Cannot open trace.\n\tIs this the correct working directory?");
	}
    }
}

VovTrace::VovTrace( VovTrace& t )
{
    vovMessage( "VovTrace(VovTrace&): not expected to be called" );
    projectName = t.projectName; 	// 
    hostName = t.hostName;		// 
    readonly = t.readonly;		// 
    type = t.type;
    if ( isClient() ) {
	zocketPointer = t.zocketPointer;
    } else {
	numberOfPages = t.numberOfPages;
	pageArray = t.pageArray;
    }
}


#ifndef VOVCLIENT
VovTrace::~VovTrace()
{
    if ( pageArray ) {
	delete pageArray;
    }
}
#endif

Bzocket& VovTrace::getBzocket()
{
    if ( zocketPointer == 0 ) {
	zocketPointer = new Bzocket(0);
    }
    return *zocketPointer;
}

// Good for RPC applications
#ifndef VOVCLIENT
VovStatus VovTrace::open( VovObject& facet )
{
    String facetName = facet.contents.facet.cell;

    projectName = facetName.before( "\\.page" );
    hostName = "<local>";
    readonly = 0;
    type = VOV_TRACE_TYPE_LOCAL;
    zocketPointer = 0;
    
    numberOfPages = 1;				// Page management.
    pageArray = new VovPage[numberOfPages];
    if ( pageArray[0].open( facet ) != VOV_OK ) {
	return VOV_ERROR;
    }

    VOV_ASSERT( init() );
    return VOV_OK;
}
#endif

VovStatus VovTrace::open( String& project, String& host, String& mode, int t)
{
    projectName = project;     		// 
    hostName = host;			// 
    readonly = ( mode == "r" );		// 
    type = t;				// 
    zocketPointer = new Bzocket( 0 );
    
    if ( isClient() ) {
	// vovMessage( "Opening connection to server  " + projectName );
	if ( VC_start_up( hostName, projectName, *zocketPointer ) != VOV_OK ) {
	    vovError(form("Cannot connect with VOV server: %s@%s", (char*)project, (char*)host));
	    type = VOV_TRACE_TYPE_DISABLED;	// Disable the trace. Do not do anything.
	    return VOV_ERROR;
	}
    } else {
#ifndef VOVCLIENT
	numberOfPages = 1;				// Page management.
	pageArray = new VovPage[numberOfPages];
	for (int i = 0; i < numberOfPages ; i++ ) {
	    String facetname = project + ".page." + dec(i);
	    if ( pageArray[i].open( facetname, mode ) != VOV_OK ) {
		vovError( form( "Cannot open page %d of trace.", i ) );
		return VOV_ERROR;
	    }
	}    
	
	VOV_ASSERT( init() );
#endif
    }
    return VOV_OK;
}


#ifndef VOVCLIENT
VovStatus VovTrace::init()
{
    // Initialize event queues.
    workQ.init( *this, 400 );	
    unreportedQ.init( *this, 400 );
	    
    // vovMessage( "// Initialize the table for fast finding of places." );
    extern VovStatus vov_place_find_init( VovNodeSet& );
    if ( vov_place_find_init( workPage().places ) != VOV_OK ) {
	vovError("Could not initialize tables for fast place finding" );
	return VOV_ERROR;
    }
	    
    // vovMessage( "// Initialize search ids in all nodes.");
    VovContentsGenerator genNode( &workPage().nodes, VOV_NODE_OCTMASK );
    VovNode node;
    while ( octGenerate( &genNode, &node ) == OCT_OK ) {
	VOV_ASSERT( node.get() );
	node.markVisited();
    }
    return VOV_OK;
}
#endif

VovStatus VovTrace::close()
{
    if ( isClient() ) return VOV_OK;

#ifndef VOVCLIENT
    projectName += "<closed>";
    for (int i = 0; i < numberOfPages ; i++ ) {
	if ( pageArray[i].close() != VOV_OK ) {
	    return VOV_ERROR;
	}
    }    
    delete pageArray;
    pageArray = 0;
#endif
    return VOV_OK;
}


VovPage& VovTrace::workPage()
{
    return pageArray[0];
}


VovStatus VovTrace::startProject()
{
    VovProp propHost("HOST_NAME", OCT_STRING );
    VovProp propDate("START_DATE", OCT_INTEGER );
    VovBag  bag( "PROJECT_STARTS" );

    Date currentDate;
    propHost = hostName;
    propDate = currentDate.get();
    VovPage& page = workPage();
    OH_ASSERT( octGetOrCreate( &page, &bag ) );
    OH_ASSERT( octCreate( &bag, &propHost ) );
    OH_ASSERT( octCreate( &bag, &propDate ) );
    return VOV_OK;
}

Date VovTrace::getProjectStartDate()
{
    VovPage& page = workPage();
    VovProp propDate("START_DATE", OCT_INTEGER );
    VovBag  bag( "PROJECT_STARTS" );

    if ( octGetByName( &page, &bag ) != OCT_OK ) {
	// This is possible only in the case of old traces.
	vovError( "Missing info on project start. Creating info." );
	startProject();
	OH_ASSERT( octGetByName( &page, &bag ) );
    }
    OH_ASSERT( octGetByName( &bag, &propDate ) );
    
    Date date = propDate.integer();
    return date;
}



VovStatus VovTrace::create( VovSuperObject& obj )
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    String errmsg = "<no error>";
    
    if ( isClient() ) {
	zoc.pack( VOV_Create );
	obj.pack( zoc );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    obj.unpack( zoc );
	}
	
	
	switch ( obj.type ) {
	case VOV_NODE:
	    errmsg = "cannot create a node. Must be either transition or place";
	    rtn = VOV_ERROR;
	    break;
	case VOV_TRANSITION:
	    { 
		VovPage& page = workPage();
		VovTransition& tran = *obj.contents.transition;
		// Is this transition the retracing of another one?
		// Must do this check before the new transition is created.
		// Create immediately after the check.
		
		VovTransition retracingTran = tran; // Retracing transition
		int retracingFlag = ( find(retracingTran) == VOV_OK );
		
		
		if ( retracingFlag ) {
		    switch (  retracingTran.getStatus() ) {
		    case VOV_RETRACING_NODE:
			// Do nothing.
			break;
		    case VOV_TRACING_NODE:
			rtn = VOV_ERROR; // Error, transition already running.
			errmsg = "Identical transition already running.";
			break;
		    default:
			vovMessage( "Manual retracing:" + retracingTran.getCommandLine() );
			VOV_ASSERT( retracingTran.initiate( VOV_RETRACING_NODE ) );
			retracingTran.setPid( 0 ); // That is the marker for a manual retracing.
		    }
		}
		
		if ( rtn == VOV_OK ) {
		    OH_ASSERT( octCreate( &page, &tran ));
		    OH_ASSERT( octAttach( &page.nodes, &tran ));
		    OH_ASSERT( octAttach( &page.transitions, &tran ) );
		    VOV_ASSERT( tran.create() );
		    
		    if ( retracingFlag ) { // RETRACING!! found a match with retracingTran.
			VOV_ASSERT( tran.retracingInvariants( *this, retracingTran ) );
		    } else {
			getQueue().enq(VOV_TRANSITION_MASK,VE_START,tran.getCommandLine(),tran.id(),0);
		    }
		}
	    }
	    break;
	case VOV_PLACE:
	    {
		VovPage& page = workPage();
		VovPlace& place = *obj.contents.place;
		
		OH_ASSERT( octCreate( &page, &place ) );
		OH_ASSERT( octAttach( &page.nodes, &place ) );
		OH_ASSERT( octAttach( &page.places, &place ) );
		VOV_ASSERT( place.create() );
		
		int  existflag;
		Date timestamp;
		int mode;
		page.placesToCheck.add( place ); 		// A new place to check.
		place.info( &existflag, &mode, &timestamp, 1 );	// 
		checkPlace( place, &existflag, &timestamp );	// Immediately check the place.
	    }
	    break;
#ifdef OBSOLETE
	case VOV_GOAL:
	    {
		VovPage& page = workPage();
		VovGoal& goal =  *obj.contents.goal;
		OH_ASSERT( octCreate( &page, &goal ));
		OH_ASSERT( octAttach( &page.goals, &goal ));
		VOV_ASSERT( goal.create() );
		getQueue().enq( VOV_GOAL_MASK, VE_START, goal.format(), goal.id() , 0 );
	    }
	    break;
#endif
	case VOV_NODE_SET:
	    {
		VovPage& page = workPage();
		VovNodeSet& set = *obj.contents.set;
		OH_ASSERT( octCreate( &page.sets, &set ));
		VOV_ASSERT( set.create() );
		int l = set.select( page.nodes );
		String msg = set.format() + "\nLength: " + dec( l );
		getQueue().enq( VOV_NODE_SET_MASK, VE_START, msg, set.id(), 0 );
	    }
	    break;
	case VOV_USER:
	    {
		VovPage& page = workPage();
		VovUser& user = *obj.contents.user;
		OH_ASSERT( octCreate( &page, &user ) );
		OH_ASSERT( octAttach( &page.users, &user ) );  
		VOV_ASSERT( user.create() );
	    }
	    break;
	case VOV_RETRACE:
	    {
		VovPage& page = workPage();
		VovRetrace& retrace = *obj.contents.retrace;
		OH_ASSERT( octCreate( &page.retracings, &retrace ) );
		VOV_ASSERT( retrace.create() );
	    }
	default:
	    break;
	}
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    if ( rtn != VOV_OK ) {
		zoc.pack( errmsg );
	    }
	    obj.pack( zoc );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
	if ( rtn != VOV_OK ) {
	    errmsg = zoc.unpackString();
	    vovError( errmsg );
	}
	obj.unpack( zoc );
    }
    
    return rtn;
}




VovStatus VovTrace::getById( VovSuperObject& obj )
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    String errmsg;
    
    if ( isClient() ) {
	zoc.pack( VOV_GetById );
	obj.pack( zoc );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    obj.unpack( zoc );
	}
	
	if ( octGetById( obj.contents.object ) != OCT_OK ) {
	    // vovMessage( form("GetById: Not a valid id: %d", obj.getId() ) );
	    rtn = VOV_ERROR;
	}  else {
	    rtn = obj.get();
	    if ( rtn == VOV_ERROR ) {
		errmsg = form( "getById: id %d is correct but type is wrong.",obj.getId() );
	    }
	}
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack ( rtn );
	    if ( rtn == VOV_OK ) {
		obj.pack( zoc );
	    } else {
		zoc.pack( errmsg );
	    }
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
	if ( rtn == VOV_OK ) {
	    obj.unpack( zoc );
	} else {
	    errmsg = zoc.unpackString();
	}
    }
    if ( errmsg != 0 ) vovError( errmsg );
    return rtn;
}



VovStatus VovTrace::modify( VovSuperObject& obj) 
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    String errmsg;
    
    if ( isClient() ) {
	zoc.pack( VOV_Modify );
	obj.pack( zoc );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    obj.unpack( zoc );
	}
	
	// Do at least some checking on the type of the object we are about to modify.
	// Should do many more checks, actually.
	if ( octGetById( obj.contents.object ) != OCT_OK ) {
	    errmsg = "Modify:Corrupt id";
	    rtn = VOV_ERROR;
	} else {
	    switch ( obj.type ) {
	    case VOV_TRANSITION:
		{
		    VovTransition& tran = *obj.contents.transition;
		    VovNodeStatus s = tran.getStatus();
		    if ( s == VOV_RETRACING_NODE || s == VOV_TRACING_NODE ) {
			// Cannot change status of retracing transition.
			rtn = VOV_ERROR;
		    } else {
			rtn =  tran.modify();
			if ( rtn == VOV_OK ) {
			    rtn = changeStatus( tran, VOV_PLAN_NODE, VOV_CHANGE_STATUS_SELF_REC );
			}
		    }
		}
		break;
	    case VOV_PLACE:
		{
		    VovPlace& place = *obj.contents.place;
		    rtn = place.modify();
		    if ( rtn == VOV_OK ) {
			rtn = changeStatus( place, VOV_PLAN_NODE, VOV_CHANGE_STATUS_RECURSIVE );
		    }
		}
		break;
	    case VOV_NODE_SET:
		rtn =  obj.contents.set->modify();
		break;
#ifdef OBSOLETE
	    case VOV_GOAL:
		rtn = obj.contents.goal->modify();
		break;
#endif
	    case VOV_USER:
		errmsg = "cannot modify a user";
		rtn = VOV_ERROR;
		break;
	    case VOV_ANNOTATION:
		rtn = obj.contents.note->modify();
		break;
	    default:
		errmsg = "modifyById NYI";
		rtn = VOV_ERROR;
	    }
	    
	    if ( rtn == VOV_OK ) {
		getQueue().enq((1<<obj.type),VE_CHANGE,obj.format()+"\nhas been modified.", obj.getId(), 0 );
	    } else {
		getQueue().enq((1<<obj.type),VE_ERROR,obj.format()+"\ncould not be modified.", obj.getId(), 0 );
	    }
	    
	}
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack ( rtn );
	    if ( rtn == VOV_OK ) {
		obj.pack( zoc );
	    } else {
		zoc.pack( errmsg );
	    }
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
	if ( rtn == VOV_OK ) {
	    obj.unpack( zoc );
	} else {
	    errmsg = zoc.unpackString();
	}
    }
    if ( rtn != VOV_OK ) vovError( errmsg );
    return rtn;
    // End of ::modify()
}

static void vi_helpforlist( const String& name, const String& rx, String& list, int& count )
{
    // vovMessage( "VHFL: name=" + name + " rx=" + rx );
    if ( name.matches( rx )  || name.contains( rx ) ) {
	list += form( " %s \n", (char*)name );
	count++;
    }
}

int VovTrace::list( int objtype, String& rx, String& objlist )
{
    Bzocket& zoc = getBzocket();
    int count = 0;
    objlist = "";
    
    if ( isClient() ) {
	zoc.pack( VOV_List );
	zoc.pack( objtype );
	zoc.pack( rx );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    objtype = zoc.unpackInt();
	    rx = zoc.unpackString();
	}
	
	String name;		// 
	VovPage& page = workPage();
	switch ( objtype ) {
	case VOV_TRANSITION:
	    {
		VovContentsGenerator genTran(&page.transitions,VOV_TRANSITION_OCTMASK);
		VovTransition t2 = 0;
		while ( octGenerate( &genTran, &t2 ) == OCT_OK ) {
		    VOV_ASSERT( t2.get() );
		    name = t2.getCommandLine();
		    vi_helpforlist( name, rx, objlist, count );
		}
	    }
	    break;
	case VOV_PLACE:
	    {
		VovContentsGenerator genPlace(&page.places,VOV_PLACE_OCTMASK);
		VovPlace p2 = 0;
		while ( octGenerate( &genPlace, &p2 ) == OCT_OK ) {
		    VOV_ASSERT( p2.get() );
		    name = p2.getName();
		    vi_helpforlist( name, rx, objlist, count );
		}
	    }
	    break;
	case VOV_NODE_SET:
	    {
		VovContentsGenerator genSet( &page.sets, VOV_NODE_SET_OCTMASK );
		VovNodeSet set;
		while ( octGenerate( &genSet, &set ) == OCT_OK ) {
		    VOV_ASSERT( set.get() );
		    name = set.getName();
		    vi_helpforlist( name, rx, objlist, count );
		} 
	    }
	    break;
#ifdef OBSOLETE
	case VOV_GOAL:
	    {
		VovContentsGenerator genGoal( &page.goals, VOV_GOAL_OCTMASK );
		VovGoal goal;
		while ( octGenerate( &genGoal, &goal ) == OCT_OK ) {
		    VOV_ASSERT( goal.get() );
		    name = goal.getName();
		    vi_helpforlist( name, rx, objlist, count );
		}		    
	    }
	    break;
#endif
	case VOV_USER:
	    vovError("list user NYI");
	    break;
	case VOV_NODE:
	    vovError("list node NYI");
	    break;
	default:
	    vovFatalError("Illegal type in list.");
	}
	
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack ( count );
	    if ( count ) zoc.pack( objlist );
	    zoc.flush();
	}
#endif
    }

    if ( isClient() ) {
	VC_get_reply( zoc );
	count = zoc.unpackInt();
	if ( count ) objlist = zoc.unpackString();
    }
    return count;
    // End of ::list()
}

VovStatus VovTrace::find( VovSuperObject& obj )
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_NOT_FOUND;
    
    if ( isClient() ) {
	zoc.pack( VOV_Find );
	obj.pack( zoc );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    obj.unpack( zoc );
	}
	
	switch ( obj.type ) {
	case VOV_TRANSITION:
	    {
		//
		// Just do a linear search.
		// In the future, this routine will have to be made more efficient.
		//
		VovPage& page = workPage();
		VovTransition& trans = *obj.contents.transition;
		VovContentsGenerator genTran(&page.transitions,VOV_TRANSITION_OCTMASK);
		VovTransition t2 = 0;
		while ( octGenerate( &genTran, &t2 ) == OCT_OK ) {
		    VOV_ASSERT( t2.get() );
		    if ( trans.match( t2 ) == VOV_OK ) {
			trans = t2;
			rtn = VOV_OK;
			break;
		    }
		}
	    }
	    break;
	case VOV_PLACE:
	    VovPlace& place = *obj.contents.place;
	    rtn = vov_place_find( place );
	    if ( rtn == VOV_OK ) {
		int exists=0;
		Date ts;
		checkPlace( place, &exists, &ts );
	    }
	    break;
	case VOV_NODE_SET:
	    {
		VovPage& page = workPage();
		VovNodeSet& set = *obj.contents.set;
		if ( octGetByName( &page.sets, &set ) == OCT_OK ) {
		    VOV_ASSERT( set.get() );
		    rtn = VOV_OK;
		} 
	    }
	    break;
#ifdef OBSOLETE
	case VOV_GOAL:
	    {
		VovPage& page = workPage();
		VovGoal& goal = *obj.contents.goal;
		if ( octGetByName( &page.goals, &goal ) == OCT_OK ) {
		    VOV_ASSERT( goal.get() );
		    rtn = VOV_OK;
		}
	    }
	    break;
#endif
	case VOV_USER:
	    {
		VovPage& page = workPage();
		VovUser& user = *obj.contents.user;
		if ( octGetByName( &page.users, &user ) == OCT_OK ) {
		    VOV_ASSERT( user.get() );
		    rtn = VOV_OK;
		}
	    }
	    break;
	case VOV_NODE:
	    vovError("find node NYI");
	    break;
	default:
	    vovFatalError("Illegal type in find.");
	}
	
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack ( rtn );
	    if ( rtn == VOV_OK ) {
		obj.pack( zoc );
	    } 
	    zoc.flush();
	}
#endif
    }

    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
	if ( rtn == VOV_OK ) {
	    obj.unpack( zoc );
	} 
    }
    return rtn;
    // End of ::find()
}


VovStatus VovTrace::findOrCreate( VovSuperObject& obj )
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    if ( find( obj ) == VOV_NOT_FOUND ) {
	return create( obj );
    }
    return VOV_OK;
}

VovStatus VovTrace::destroy( VovSuperObject& obj) 
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    String errmsg;
    
    if ( isClient() ) {
	zoc.pack( VOV_Delete );
	obj.pack( zoc );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    obj.unpack( zoc );
	}
	
	if ( getById( obj ) != VOV_OK ) {
	    errmsg = "Destroy:Corrupt id";
	    rtn = VOV_ERROR;
	} else {
	    
	    switch ( obj.type ) {
	    case VOV_NODE:
		{
		    VovNode& node = *obj.contents.node;
		    if ( node.nodeType() == VOV_PLACE ) {
			VovPlace p = node.id();
			rtn = destroy( p );
		    } else {
			VovTransition t = node.id();
			rtn = destroy( t );
		    }
		}
		break;
	    case VOV_TRANSITION:
		{
		    // Worry about chains first.
		    VovTransition& t = *obj.contents.transition;
		    VovChainInfo chain;
		    VovContainersGenerator genInputs( &t, VOV_PLACE_OCTMASK );
		    VovPlace p;
		    while ( octGenerate( &genInputs, &p ) == OCT_OK ) {
			VOV_ASSERT( p.get() );
			if ( p.getChain( chain ) == VOV_OK ) {
			    VovPlace p2;
			    if ( chain.getNext( p, p2 ) == VOV_OK && t.isOutput( p2 ) == VOV_OK ) {
				VOV_ASSERT( p.merge( t, p2, VOV_FORWARD ) );
			    }
			}
		    }
		    t.disconnect( VOV_IOS_MASK );
		    rtn = obj.contents.transition->del();
		}
		break;
	    case VOV_PLACE:
		{
		    VovChainInfo chain;
		    VovPlace& place = *obj.contents.place;
		    if ( place.getChain( chain ) == VOV_OK ) {
			rtn = chain.destroy();
		    } else {
			rtn = place.del();
		    }
		}
		break;
	    case VOV_NODE_SET:
		rtn =  obj.contents.set->del();
		break;
#ifdef OBSOLETE
	    case VOV_GOAL:
		rtn = obj.contents.goal->del();
		break;
#endif
	    case VOV_USER:
		errmsg = "Cannot delete a `user'";
		rtn = VOV_ERROR;
		break;
	    case VOV_ANNOTATION:
		rtn = obj.contents.note->del();
		break;
	    case VOV_RETRACE:
		rtn = obj.contents.retrace->del();
		break;
	    default:
		vovError("destroy NYI");
	    }
	    
	    if ( rtn == VOV_ERROR ) {
		errmsg = "Cannot forget";
	    } else {
		if ( obj.type != VOV_NODE ) { // Just a special case.
		    getQueue().enq( (1<<obj.type),VE_FORGET,obj.format()+"\nForgotten.",0,0);
		}
	    }
	}
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack ( rtn );
	    if ( rtn != VOV_OK ) {
		zoc.pack( errmsg );
	    }
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
	if ( rtn != VOV_OK ) {
	    errmsg = zoc.unpackString();
	}
    }
    if ( rtn != VOV_OK ) vovError( errmsg );
    return rtn;
    // End of ::destroy()
}


// Locking of objects.
VovStatus VovTrace::lock( VovObject& obj, String& desc, int force, octId& lockId, String& msg )
{
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
 
   if ( isClient() ) {
	zoc.pack( VOV_Lock );
	zoc.pack( obj.id() );
	zoc.pack( desc );
	zoc.pack( force );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer( VOV_SERVER_ENTRY ) ) {
	    obj.id() = zoc.unpackInt();
	    desc = zoc.unpackString();
	    force = zoc.unpackInt();
	}
	if ( obj.getById() != OCT_OK ) {
	    msg = "Illegal object id.";
	    rtn = VOV_ERROR;
	} else {
	    if ( force ) {
		journalLog( "Forcing a lock:" + obj.format() + "\n" + desc );
	    }
	    rtn = obj.lock( desc, force, lockId,  msg );
	    if ( rtn != VOV_OK ) {
		accumulateStatistics( String("CONFLICT_LOCKS"), 1 );
	    }
	}
	if ( isServer( VOV_SERVER_EXIT ) ) {
	    zoc.pack( rtn );
	    zoc.pack( lockId );
	    zoc.pack( msg );
	    zoc.flush();
	}
#endif	
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
	lockId = zoc.unpackInt();
	msg = zoc.unpackString();
    }
    return rtn;
}

VovStatus VovTrace::unlock( VovObject& obj, octId lockId )
{
    Bzocket& zoc = getBzocket();
    
    VovStatus rtn = VOV_OK;
    if ( isClient() ) {
	zoc.pack( VOV_Unlock );
	zoc.pack( obj.id() );
	zoc.pack( lockId );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer( VOV_SERVER_ENTRY ) ) {
	    obj.id() = zoc.unpackInt();
	    lockId = zoc.unpackInt();
	}
	if ( obj.getById() != OCT_OK ) {
	    rtn = VOV_ERROR;	   //  msg = "Illegal object id.";
	} else {
	    rtn = obj.unlock( lockId );
	}
	if ( isServer( VOV_SERVER_EXIT ) ) {
	    zoc.pack( rtn );
	    zoc.flush();
	}
#endif	
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
    }
    return rtn;
}




int VovTrace::testConnectivity( VovNode& node )
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    
    int mask = 0;
    
    if ( isClient() ) {
	zoc.pack( VOV_TestConnectivity );
	zoc.pack( node.id() );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    node.id() = zoc.unpackInt();
	    if ( node.get() != VOV_OK ) {
		vovError( "Bogus id" );
	    }
	}
	
	if ( node.hasInputs() == VOV_OK ) mask |= VOV_INPUT_MASK;
	if ( node.hasOutputs() == VOV_OK ) mask |= VOV_OUTPUT_MASK;
	
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( mask );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	mask = zoc.unpackInt();
    }
    
    return mask;
}



// Set operations 
#define VOV_CONTENTS_GENERATOR   1
#define VOV_CONTAINERS_GENERATOR 2

VovStatus VovTrace::initGenContents( VovObject& obj, octObjectMask mask, VovGenerator& gen )
{
    return initGenerator( VOV_CONTENTS_GENERATOR, obj, mask, gen );
}

VovStatus VovTrace::initGenContainers( VovObject& obj, octObjectMask mask, VovGenerator& gen )
{
    return initGenerator( VOV_CONTAINERS_GENERATOR, obj, mask, gen );
}

VovStatus VovTrace::initGenerator( int type, VovObject& obj, octObjectMask mask, VovGenerator& gen )
{
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_ERROR;

    if ( isClient() ) {
	zoc.pack( VOV_InitGenerator );
	zoc.pack( type );
	zoc.pack( obj.id() );
	zoc.pack( mask );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    type = zoc.unpackInt();
	    obj.id() = zoc.unpackInt();
	    mask = zoc.unpackInt();
	}

	if ( type == VOV_CONTENTS_GENERATOR ) {
	    rtn = octInitGenContents( &obj, mask, &gen );
	} else {
	    rtn = octInitGenContainers( &obj, mask, &gen );
	}

	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    gen.pack( zoc );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
	gen.unpack( zoc );
    }
    return ( rtn == OCT_OK ) ? VOV_OK : VOV_ERROR;
    return rtn;
}

VovStatus VovTrace::generate( VovGenerator& gen, VovObject& obj )
{
    Bzocket& zoc = getBzocket();
    octStatus rtn;

    if ( isClient() ) {
	zoc.pack( VOV_Generate );
	gen.pack( zoc );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    gen.unpack( zoc );
	}

	rtn = octGenerate( &gen, &obj );

	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    obj.pack( zoc );
	    gen.pack( zoc );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
	obj.unpack( zoc );
	gen.unpack( zoc );
    }
    return ( rtn == OCT_OK ) ? VOV_OK : VOV_ERROR;
}

VovStatus VovTrace::freeGenerator( VovGenerator& gen )
{
    Bzocket& zoc = getBzocket();
    octStatus rtn;

    if ( isClient() ) {
	zoc.pack( VOV_FreeGenerator );
	gen.pack( zoc );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    gen.unpack( zoc );
	}

	rtn = octFreeGenerator( &gen );

	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
    }
    return ( rtn == OCT_OK ) ? VOV_OK : VOV_ERROR;
}




VovStatus VovTrace::OK( String&  msg )
{
    VovStatus rtn = VOV_OK;
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    
    if ( isClient() ) {
	zoc.pack( VOV_Ok );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    // Nothing to get.
	}
	
	VovPage& page = workPage();
	
	msg += "Making sure each place has at most one input.\n";

	VovContentsGenerator genPlace( &page.places, VOV_PLACE_OCTMASK );
	VovPlace place;
	octId storeIdForSpeedTest = oct_null_id;
	while ( octGenerate( &genPlace, &place ) == OCT_OK ) {
	    VOV_ASSERT( place.get() );
	    storeIdForSpeedTest = place.id();
	    // Skip places which are tracing or retracing.
	    if ( place.getStatus() == VOV_RETRACING_NODE || place.getStatus() == VOV_TRACING_NODE ) continue;
	    int count = place.numberOfInputs();
	    if ( count > 1 ) {
		msg +=  place.format() + "\n\thas more than one input.\n";
		VovContainersGenerator genTrans( &place, VOV_TRANSITION_OCTMASK );
		VovTransition trans = 0;
		while ( count-- > 1 ) {
		    OH_ASSERT( octGenerate( &genTrans, &trans ) );
		    VOV_ASSERT( trans.get() );
		    msg += "Destroying spurious input " + trans.format() + "\n";
		    VOV_ASSERT( destroy( trans ) );
		}
	    }
	}
	
	if ( storeIdForSpeedTest != oct_null_id ) {
	    int iter = 1000;
	    // timer tt = timerCreate();
	    // timerStart( tt );
	    // for ( int i = 0 ; i < iter; i++ ) {
	    // VovPlace place( storeIdForSpeedTest );
	    // }
	    // timerStop( tt );
	    // double interval = timerElapsedTime( tt );
	    // vovMessage(form("%d place.get() in %.3fs = %.6fs/place",iter,interval,interval/iter));
	 }
	msg += "Making sure each transition has at least one output and one input.\n";
	
	VovContentsGenerator genTransition( &page.transitions, VOV_TRANSITION_OCTMASK );
	VovTransition transition;
	while ( octGenerate( &genTransition, &transition ) == OCT_OK ) {
	    VOV_ASSERT( transition.get() );
	    VovNodeStatus s = transition.getStatus();
	    if ( s == VOV_RETRACING_NODE || s == VOV_TRACING_NODE ) continue;
	    
	    if ( transition.hasOutputs()==VOV_NOT_OK || transition.hasInputs() == VOV_NOT_OK ) {
		msg += ( "Destroying transition with no I/Os\n\t" + transition.getCommandLine() + "\n");
		VOV_ASSERT( destroy( transition ) );
	    }
	}
	
	{
	    msg += "Making sure the status of all nodes is consistent.\n" ;

	    VovContentsGenerator genTransition( &page.transitions, VOV_TRANSITION_OCTMASK );
	    VovTransition transition;
	    while ( octGenerate( &genTransition, &transition ) == OCT_OK ) {
		VOV_ASSERT( transition.get() );
		VovNodeStatus status = transition.getStatus();
		if ( status == VOV_RETRACING_NODE || status == VOV_TRACING_NODE || status == VOV_DEAD_NODE ) continue;
		
		VovContentsGenerator genOutputs( &transition, VOV_NODE_OCTMASK );
		VovNode node;
		while ( octGenerate( &genOutputs, &node ) == OCT_OK ) {
		    VOV_ASSERT( node.get() );
		    VovNodeStatus s = node.getStatus();
		    if ( s == VOV_DEAD_NODE ) continue;
		    if ( s != transition.getStatus() ) {
			String tmpMsg;
			transition.changeStatus( VOV_TRACE_NODE, VOV_CHANGE_STATUS_SELF, tmpMsg );
			transition.changeStatus( VOV_PLAN_NODE, VOV_CHANGE_STATUS_SELF_REC, tmpMsg );
			node.changeStatus( VOV_PLAN_NODE, VOV_CHANGE_STATUS_SELF_REC, tmpMsg );
			if ( msg.length() < 10000 ) {
			    msg += ( "Inconsistent state of outputs of\n" + transition.format() + "\n" + tmpMsg + "\n" );
			} 
		    }
		}
	    }
	}
	
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    zoc.pack( msg );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
	msg = zoc.unpackString();
    }
    return rtn;
}



void VovTrace::journalLog( const String& msg )
{
#ifdef JOURNAL_IS_USELESS
    static int msg_count = 0;			// Count messages sent to journal.
    static int previousdate = 0;		// Date of latest entry.
    Date date;					// Get current date.
    int delta = ( previousdate ) ? date.get() - previousdate : 0 ;
    previousdate = date.get();			// 
    String jn = "journal." + projectName;	// Journal name.
    ostream journal( (char*)jn , io_appendonly, a_use); 
    if ( journal.good() ) {	// Could be bad. e.g. due to NFS timeout.
	journal.form("\n\t####%4d: %s: [%d]\n",
		     msg_count++, (char*)date.format(), delta );	
	journal.form( msg );			// Print the message.
	journal.form( "\n" );			// Space the messages.
	journal.close();				// Always close journal.
    }
    // vovMessage( msg );				// Print message on tty too.
#endif
}



/*
**
*/
#ifndef VOVCLIENT
VovStatus VT_dumpTransitions( VovNode& node , void* data )
{
    ostream& streamdump = *(ostream*)data;
    VovTransition t( node );
    streamdump << t.format() << "\n\n";
    return VOV_OK;
}
#endif

VovStatus VovTrace::saveOntoDisk( int asciiDump = 0 , String& asciiFile = 0 )
{
    VovStatus rtn = VOV_OK;
    Bzocket& zoc = getBzocket();
    
    if (isClient() ) {
	zoc.pack( VOV_SaveOntoDisk );
	zoc.pack ( asciiDump );
	if ( asciiDump ) zoc.pack( asciiFile );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    asciiDump = zoc.unpackInt();
	    if ( asciiDump ) asciiFile = zoc.unpackString();
	}
	
	if ( asciiDump ) {
	    String dumpname = asciiFile;
	    if ( dumpname == 0 ) {
		dumpname = "/usr/tmp/vovdump." + projectName ;
	    }
	    vovMessage( "Dumping the trace in " + dumpname  );
	    ostream dump( (char*)dumpname, io_writeonly, a_create); 
	    VovPage& page = workPage();
	    page.transitions.cover( VT_dumpTransitions, (void*)&dump );
	} else {
	    vovMessage( form("Saving trace: %d pages...",numberOfPages) );
	    for (int i = 0; i < numberOfPages; i++) {
		rtn = pageArray[i].saveOntoDisk( );
		if ( rtn == VOV_ERROR ) {
		    getQueue().enq( VOV_NULL_MASK, VE_ERROR, "\nMAJOR ERROR\n\tSaving of trace has failed.\n", 0 , 0 );
		    break;
		}
	    }
	}
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( VOV_SaveOntoDisk );
	    zoc.pack( rtn );
	    zoc.flush();
	}
#endif
    }	
    if ( isClient() ) {
	VC_wait_for_code( zoc, VOV_SaveOntoDisk );
	rtn = zoc.unpackInt();
    }
    return rtn;
}




String VovTrace::userFormat( int extended = 0 ) 
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    String user_string ;
    String all_string ;

    String  username = utcc_getenv( "USER" ) ;

    if ( isClient() ) {
	zoc.pack( VOV_UserFormat );
	zoc.pack( utcc_getenv( "USER" ) );
	zoc.pack( extended );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    username = zoc.unpackString();
	    extended = zoc.unpackInt();
	} 

	VovPage& page = workPage();
    
	VovContentsGenerator genUser( &page.users, VOV_USER_OCTMASK );
	VovUser user  = 0;
	while ( octGenerate( &genUser, &user ) == OCT_OK ) {
	    VOV_ASSERT( user.get() );
	    if ( user.getName() == username ) {
		user_string += user.format( 1 ) + "\n" ; 	// Always extended format for user.
	    } else {						// 
		all_string += user.format( extended ) + "\n" ;	// Possibly extended format for all others.
	    }
	}
	all_string = user_string + all_string;
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( all_string );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	all_string = zoc.unpackString();
    }
    return all_string;
}

SlaveSet& serverSlaveSet( SlaveSet* ptr )
{
#ifndef VOVCLIENT
    static SlaveSet* slaveSetPtr = 0;
    if ( ptr ) {
	slaveSetPtr = ptr;
    } 
    return *slaveSetPtr;
#endif
}

String serverFormat()
{
#ifndef VOVCLIENT
    extern char  edata;
    int size = (int)sbrk(0) - (int)&edata;
    extern int updateServerDelay( int, VovCode );
    int delay = updateServerDelay( 0, 0 );
    String s = form("SERVER: size:%8d delay:%4d.\n", size, delay );
    return s;
#endif
}


// FUNCTION updateServerDelay()
// The server slows down for lack of activity.
// Almost all activities wake up the server and make it
// run full speed, except those activities that are
// periodically initiated by slaves.
int updateServerDelay( int timeout,  VovCode code )
{
    const  int minDelay = 2;		// 
    const  int maxDelay = 7200;		// That is 2 hr.
    const  int maxTimeOut = 120;	// 2 minutes
    static int timeOutCounter = 0; 	// 
    static int delay = minDelay * 50;	// Start up slow.

    if ( timeout ) {
	timeOutCounter += delay;
	if ( timeOutCounter > maxTimeOut ) {
	    timeOutCounter = 0;
	    delay *= 2;
	    if (delay > maxDelay) delay = maxDelay;
	}
    } else {
	switch ( code ) {		// 
	case (VovCode)0:
	case VOV_StartUp:		// Startup procedure.
	case VOV_SynchClocks:		// Startup procedure.
	case VOV_Debug:			// 
	case VOV_EchoLog:		// 
	case VOV_SlaveReady:		// Coming from slave.
	case VOV_GetInfo:		// 
	case VOV_GetById:
	case VOV_GetIntermediateNodes:
	case VOV_Find:
	case VOV_Edit:
	case VOV_CheckPlaces:
	case VOV_GetAnnotations:
	case VOV_GetHistory:
	case VOV_Format:
	case VOV_UserFormat:
	case VOV_GetScript:
	case VOV_GetHostLoad:
	    ;				// Do not do anything.
	    break;			// 
	default:			// Wake up.
	    delay = minDelay;		// 
	    timeOutCounter = 0;		// 
	    break;			// 
	}				// 
    }
    return delay;			// 
}

String VovTrace::getInfo( int mask, int detailed )
{
    String s;
    Bzocket& zoc = getBzocket();
    
    if ( isClient() ) {
	zoc.pack( VOV_GetInfo );
	zoc.pack( mask );
	zoc.pack( detailed );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    mask = zoc.unpackInt();
	    detailed = zoc.unpackInt();
	}

	if ( mask & VOV_INFO_TRACE ) {
	    s += format();
	}
	if ( mask & VOV_INFO_USER  ) {
	    s  += userFormat( detailed );
	}
	if ( mask & VOV_INFO_SLAVE ) {
	    extern SlaveSet& serverSlaveSet( SlaveSet* );
	    s += serverSlaveSet(0).format();
	}
	if ( mask & VOV_INFO_SERVER) {
	    extern String serverFormat();
	    s += serverFormat( );
	} 
	if ( mask & VOV_INFO_CWD ) {
	    s += ( "WORKING DIRECTORY: " + path_canonicalize( utcc_getwd() ) + "\n") ;
	}
	if ( mask & VOV_INFO_DATE  ) {
	    Date dummyDate;
	    s += dummyDate.mapFormat();
	}
	if ( mask & VOV_INFO_STATS ) {
	    s += formatStatistics();
	}

	if ( isServer( VOV_SERVER_EXIT ) ) {
	    zoc.pack( s );
	    zoc.flush();
	}
#endif
    } 
    
    if ( isClient() ) {
	VC_get_reply( zoc );
	s = zoc.unpackString();
    }
    return s;
}


String VovTrace::format()
{
    Bzocket& zoc = getBzocket();
    
    String s;

    if ( isClient() ) {
	zoc.pack( VOV_Format );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    // nothing to retrieve.
	}
	for (int i = 0 ; i <  numberOfPages; i++ ) {
	    s += pageArray[i].format();
	    s += "\n";
	}
	
	VovPage& page = workPage();
	VovContentsGenerator gen( &page.retracings, VOV_RETRACE_OCTMASK );
	VovRetrace retrace;
	while ( octGenerate( &gen, &retrace ) == OCT_OK ) {
	    VOV_ASSERT( retrace.get() );
	    s += retrace.format() + "\n" ;
	}
	s += form( "%s retracing.\n", retraceFastMode == VOV_RETRACE_FAST ? "FAST" : "SLOW" );

	s += form( "%d unreported events.\n", unreportedQ.length() );

	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( s );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	s = zoc.unpackString();
    }

    return s;
}



