/*
traversal.cc

Routines for traversal of a trace.

*** Place to place shortest path

*** Place to place longest path
    to find minimum time to completion 

*** Depth-first   search from node (place or transition)
*** Breadth-first search from node (place or transition)
    To invalidate children of place

*** backward from place 
    to read history
*/
#include "trace.h"
#include "transition.h"
#include "place.h"
#include "VOVprotocol.h"
#include "event.h"
#include "voverror.h"
#include "invalidate.h"
#include "comm.h"

extern "C"
{
#include <signal.h>
}


static int vovNodeSearchId_gl = 0;
int VovNode::newVisit()
{
    return vovNodeSearchId_gl++;
}

void VovNode::markVisited()
{
    searchId.modify( vovNodeSearchId_gl );
}

VovStatus VovNode::alreadyVisited()
{
    return ( searchId.integer() == vovNodeSearchId_gl ) ? VOV_OK : VOV_NOT_OK ;
}


//
// The VovStatusFunc functions return the following values:
// VOV_OK   : all ok, continue search.
// VOV_STOP : stop search from this node, but continue on other nodes, and search return VOV_OK.
// VOV_ERROR: stop search now, return VOV_ERROR;
//

//
// This is the main search routine for the trace.
// It is actually a depth first search.
// 
VovStatus 
VovNode::bfs( int recDir, 		// Recurs. flg & direction:forwrd or backwrd
	     VovStatusFunc before,   	// called just after entering.
	     VovStatusFunc visit_in,	// called on all the inputs.
	     VovStatusFunc visit_out,	// called on all the outputs.
	     VovStatusFunc after,	// called just before returning.
	     void* data
	     )
{
    //
    // Determine if this is the beginning of a search.
    // If so, choose an id for this search. 
    // The id is then used to avoid multiple visits to the same node.
    // Then set the RECURSION bit.
    //
    const int RECURSION_MASK = 0x10;
    if ( ! (recDir & RECURSION_MASK) ) {
	newVisit();				// Begin a new traversal of nodes.
	recDir |= RECURSION_MASK;		// Set the recursion bit.
    } else {					// This is a recursive call.
	if ( alreadyVisited() == VOV_OK ) {
	    return VOV_OK;			// Yes, return.
	} 
    }
    
    // vovMessage( form("visiting node %d", id() ) );
    
    markVisited();

    int direction = ( recDir & ~RECURSION_MASK );
    
    if ( before != 0) {
	VovStatus rtn = (*before)( *this , data );
	if ( rtn != VOV_OK ) {
	    return ( rtn == VOV_STOP ) ? VOV_OK : rtn ;
	}
    }
    
    if ( visit_in != 0) {
	VovContainersGenerator genInst( this, OCT_INSTANCE_MASK );
	VovNode node = 0;
	while ( octGenerate( &genInst, &node ) == OCT_OK ) {
	    node.get();
	    VovStatus rtn = (*visit_in)( node , data );
	    if ( rtn != VOV_OK ) {
		return ( rtn == VOV_STOP ) ? VOV_OK : rtn ;
	    }
	}
    }
    
    if ( visit_out != 0) {
	VovContentsGenerator genInst( this, OCT_INSTANCE_MASK );
	VovNode node = 0;
	while ( octGenerate( &genInst, &node ) == OCT_OK ) {
	    node.get();
	    VovStatus rtn =  (*visit_out)( node , data );
	    if ( rtn != VOV_OK ) {
		return ( rtn == VOV_STOP ) ? VOV_OK : rtn ;
	    }
	}
    }
    
    if (direction == VOV_FORWARD ) {
	VovContentsGenerator genInst( this, OCT_INSTANCE_MASK );
	VovNode node = 0;
	while ( octGenerate( &genInst, &node ) == OCT_OK ) {
	    node.get();
	    VovStatus rtn=node.bfs(recDir,before,visit_in,visit_out,after,data);
	    if ( rtn != VOV_OK) {
		return rtn;
	    }
	}
	
    } else {
	VovContainersGenerator genInst( this, OCT_INSTANCE_MASK );
	VovNode node = 0;
	while ( octGenerate( &genInst, &node ) == OCT_OK ) {
	    node.get();
	    VovStatus rtn=node.bfs(recDir,before,visit_in,visit_out,after,data);
	    if ( rtn !=VOV_OK)  {
		return rtn;
	    }
	    
	}
    }
    
    if ( after != 0) {
	VovStatus rtn = (*after)( *this , data );
	if ( rtn != VOV_OK ) {
	    return ( rtn == VOV_STOP ) ? VOV_OK : rtn ;
	}
    }
    
    return VOV_OK;
}



/*
** Cycle detection routines.
*/

typedef struct cycle_struct {
    octId endId;
    int nodesVisited;
} VovCycleData;

VovStatus VT_cycleVisitOutput( VovNode& node, void* d)
{
    VovCycleData& data = *(VovCycleData*)d;
    data.nodesVisited++;
    if ( node.id() == data.endId ) {
	return VOV_ERROR;
    }
    return VOV_OK;
}

VovStatus vov_cycle( VovNode& start, VovNode& end )
{
    static int maxCycleWork = 0; // To record the max work done by a cycle detection routine.

    VovCycleData data;
    data.endId = end.id();
    data.nodesVisited = 0;
    if (start.bfs(VOV_FORWARD,0,0,VT_cycleVisitOutput,0,(void*)&data)==VOV_ERROR) {
	vovError( "Cycle detected from" + start.format() + " to " + end.format() );
	return VOV_ERROR;
    }
    if ( data.nodesVisited > maxCycleWork )   {
	maxCycleWork = data.nodesVisited;
	vovMessage(form("CYCLE detection: nodes= %d (MAX=%d)",data.nodesVisited,maxCycleWork));
    }
    return VOV_OK;
}

VovStatus VT_invalidate( VovNode& node, void* data)
{
    VovStatus rtn = VOV_OK;
    VovInvalidateInfo& info = *(VovInvalidateInfo*)data;
    switch ( node.getStatus() ) {
    case VOV_TRACING_NODE:
    case VOV_RETRACING_NODE:
	if ( node.nodeType() == VOV_TRANSITION ) {
	    // VovTransition trans = node.id();
	    // vovMessage( "Killing " + trans.getCommandLine() );
	    // trans.kill( SIGKILL );
	    // trans.terminate();
	    // info.msg += (trans.getCommandLine() + " killed.\n" );
	    info.transitions++;
	} else {
	    info.places++;
	    // node.setStatus( info.status );
	}
	break;
    case VOV_TRACE_NODE:
	if ( node.nodeType() == VOV_PLACE ) {
	    VovPlace place = node.id();
	    place.unsetCurrent();
	    info.places++;
	} else {
	    VovTransition tran( node );
	    info.transitions++;
	    info.durations += tran.duration();
	}
	node.setStatus( info.status );
	break;
    case VOV_DEAD_NODE:
	if ( info.revive ) {
	    // vovMessage( "Reviving dead node" );
	    node.setStatus( info.status );
	} else {
	    // vovMessage( "Not reviving a dead node" );
	    rtn = VOV_STOP;
	}
	break;
    case VOV_PLAN_NODE:
    case VOV_NONEXISTING_NODE:
	node.setStatus( info.status );
	break;
    default:
	vovError( "VT_invalidate: funny status" );
	rtn = VOV_STOP;
	break;
    }
    return rtn;
}


VovStatus VovTrace::changeStatus( VovNode& node, VovNodeStatus s, int flags )
{
    Bzocket& zoc = getBzocket();
    
    if ( isClient() ) {
	zoc.pack( VOV_ChangeStatus );
	node.pack( zoc );
	zoc.pack( s );
	zoc.pack( flags );
	zoc.flush( );
    } else {
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    node.unpack( zoc );
	    s = zoc.unpackInt();
	    flags = zoc.unpackInt();
	    if ( node.get() != VOV_OK ) {
		vovFatalError( "Wrong node" );
	    }
	}
	String msg;
	String cs_msg = "\n";		// Returned by changeStatus.
	node.changeStatus( s, flags, cs_msg );
	if ( node.nodeType() == VOV_PLACE ) {
	    VovPlace place(node);
	    msg = place.format() + "\nis now " +  node.getStatusString() ;
	} else {
	    VovTransition tran(node);
	    msg = tran.getCommandLine() + "\nis now " +  node.getStatusString() ;
	}
	getQueue().enq( (1 << node.nodeType() ), VE_CHANGE, msg + cs_msg, node.id(), 0 );
	if ( isServer(VOV_SERVER_EXIT) ) {
	    node.pack( zoc );
	    zoc.flush();
	}
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	node.unpack( zoc );
    } 
    return VOV_OK;
}

VovStatus VT_checkNodeStatus( VovNode& n, void* s )
{
    VovNodeStatus  ns = *(VovNodeStatus*)s;
    return  ( n.getStatus() == ns ) ?  VOV_OK : VOV_ERROR;
}

VovStatus VT_setNodeStatus( VovNode& n, void* s )
{
    n.setStatus( *(VovNodeStatus*)s );
    return  VOV_OK ;
}

VovStatus VovNode::forallOutputs( VovStatusFunc func, void* data )
{
    VovContentsGenerator genOut( this, VOV_NODE_OCTMASK );
    VovNode node = 0;
    while (octGenerate( &genOut, &node ) == OCT_OK ) {
	node.get();
	if ((*func)( node , data) == VOV_ERROR ) {
	    return VOV_ERROR;
	}
    }
    return VOV_OK;
}

VovStatus VovNode::forallInputs( VovStatusFunc func, void* data )
{
    VovContainersGenerator genIn( this, VOV_NODE_OCTMASK );
    VovNode node = 0;
    while (octGenerate( &genIn, &node ) == OCT_OK ) {
	node.get();
	if ((*func)( node , data) == VOV_ERROR ) {
	    return VOV_ERROR;
	}
    }
    return VOV_OK;
}

VovStatus VovNode::foreachOutput( VovStatusFunc func , void* data)
{
    VovStatus ret = VOV_OK;
    VovContentsGenerator genOut( this, VOV_NODE_OCTMASK );
    VovNode node = 0;
    octId firstId = 0;		// To avoid infinite loops.
    while (octGenerate( &genOut, &node ) == OCT_OK ) {
	if ( node.id() == firstId )	    // Has this already been processed?
	    break;			    // Yes, exit the loop.
	node.get();			    // Else process this node.
	if ((*func)( node, data ) == VOV_ERROR ) {
	    ret =  VOV_ERROR;
	}
	if ( ! firstId ) firstId = node.id();
    }
    return ret;
}



// Routines to gather nodes which lie between two sets of nodes.

class VovINS {
public:
    VovNodeSet* sources;
    VovNodeSet* sinks;
    VovSelectRule* rule;
    VovNodeSet* trans;
    VovINS( VovNodeSet& so, VovNodeSet& si, VovSelectRule& r, VovNodeSet& tr )
      { sources = &so; sinks = &si ; rule = &r; trans = &tr; }
};

VovStatus VT_selectIN( VovNode& node, void* ins )
{
    VovINS& i = *(VovINS*)ins;
    if ( i.rule->select( node ) ) {
	// vovMessage("VT_IN: Add node" );
	i.trans->add( node );
    }
    if ( i.sources->contains(node) ) {
	// vovMessage( "VT_IN: stopping" );
	return VOV_STOP;
    }
    return VOV_OK;
}


VovStatus VovTrace::getIntermediateNodes(VovNodeSet& sources, 
					 VovNodeSet& sinks,
					 VovSelectRule& rule,
					 VovNodeSet& trans )
{
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    String errmsg;

    if ( isClient() ) {
	zoc.pack( VOV_GetIntermediateNodes );
	sources.pack( zoc );
	sinks.pack( zoc );
	rule.pack( zoc );
	trans.pack( zoc );
	zoc.flush();
    } else {
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    sources.unpack( zoc );
	    sinks.unpack( zoc );
	    rule.unpack( zoc );
	    trans.unpack( zoc );
	}
	
	VovINS ins( sources, sinks, rule, trans );
	trans.clear( VOV_NODE_OCTMASK ) ;
	VovContentsGenerator genNode( &sinks, VOV_NODE_OCTMASK );
	VovNode node;
	while ( octGenerate( &genNode, &node ) == OCT_OK ) {
	    node.get();
	    rtn = node.bfs(VOV_BACKWARD,VT_selectIN,0,0,0,&ins);
	    if ( rtn == VOV_ERROR ) {
		errmsg = "in generating script";
		break;
	    }
	}
	
	if ( isServer(VOV_SERVER_EXIT) ) {
	    trans.pack( zoc );
	    zoc.pack( rtn );
	    zoc.pack( errmsg );
	    zoc.flush();
	}
    }
    if (isClient() ) {
	VC_get_reply( zoc );
	trans.unpack( zoc );
	rtn = zoc.unpackInt();
	errmsg = zoc.unpackString();
    }
    return rtn;
}
