#include "port_internal.h"

#include <errno.h>

#include "utility.h"		// From octtools.

#include "utcc.h"		// Vov.
#include "path.h"
#include "voverror.h"
#include <errtrap.h>



// This is also in sys/param.h, but including it causes problems.
#define MAXPATHLEN 1024

/* C includes */
extern "C"
{
#include <sys/types.h>
#include <sys/stat.h>
int stat();
int lstat();
typedef struct stat stat_struct;
}

void utcc_fatalerror( String&  s )
{
    errRaise("utcc",-1,"Fatal Error: %s", (char*)s );
}



void utcc_error( String& s )
{
    cerr << "UTCC: Error: " << s << "\n";
}

void utcc_message( String& s )
{
    cerr << "UTCC: Message: " << s << "\n"; 
    fflush(stderr);
}

void utcc_printerrno( String& where )
{
    char* errdesc = (errno > 0 && errno < sys_nerr) ? sys_errlist[errno] : "?";
    cerr << "UTCC error: " <<  errdesc;
    if ( where != 0 ) {
	cerr << " at " << where ;
    }
    cerr << "\n";
    fflush (stderr);
}



String utcc_getenv( String& name, int loud = 1 )
{
    String ret = getenv( (char*)name );
    if (ret == 0 && loud ) {
	utcc_error( "Missing environment variable " + name );
    }
    return ret;
}


String utcc_gethostname()
{
    static char* hostname = 0;
    if ( ! hostname ) {
	char buffer[64];				// 
	gethostname(buffer, 64); 			// Get the complete hostname.
	char* period = index( buffer, '.' );		// Find the period in the name: e.g. canova.Berkeley.EDU
	if ( period ) *period = '\0';			// If there is a period, truncate string there.
	hostname = utcc_strsav( (char*)buffer );	// Cash name.
    }
    return (String)hostname;
}

String utcc_machine()
    // Make an egregious effort to find the type of machine
    // we are running on. Normally, this is determined using the
    // MACHINE environment variable, but on some systems we need
    // to call some strange programs. 
{
    String s = getenv( "MACHINE" );
    if ( s == 0 ) {
	if ( util_path_search( "machine" ) != NIL(char) ) {
	    s  = utcc_popen( "machine" ); // This works on mips and vaxes
	} else if ( util_path_search( "arch" ) != NIL(char) ) {
	    s  = utcc_popen( "arch" ); // This works on suns
	}
	if ( s.contains( "\n" ) ) {
	    s = s.before( "\n" );
	}
    }
    return s;
}

String utcc_popen( const String& cmd )
{
    FILE* fp = popen( (char*)cmd , "r" ); 		// Open a pipe, for reading only.
    char buf[256];					// Buffer to read form the pipe.
    String ret;						// Returned value;
    int n;						// Characters read by fread.
    while ( (n = fread(buf,sizeof(char),255,fp))>0) {	// Read what you can from the pipe into buffer.
	buf[n] = '\0';					// 
	ret += buf;					// 
    }							// 
    if (pclose( fp )==-1) utcc_printerrno("popen");	// 
    return ret;						// Return the string.
}

String utcc_which( const String& prog )
{
    stat_struct buf;
    if ( ::stat( (char*)prog, &buf ) == 0 ) return prog;

    String path = utcc_getenv( "PATH" );
    
    while ( path != 0 ) {
	String first = path.before( ":" );
	String try = first + "/" + prog;
	vovMessage( "Trying: " + try );
	if ( ::stat( (char*)try, &buf ) == 0 ) return try;
	path = path.after( ":" );
    }
    return "";			// Not found.
}

String utcc_getwd()
{
    static char* cwd_static = 0;
    if ( ! cwd_static )  {
#ifdef SPRITE
	// utcc_message( "Using popen to find pwd" );
	String tmp = utcc_popen( "pwd" );
	String s = tmp;
	if  ( tmp.contains( "\n" ) ) {
	    s = tmp.before( "\n" );
	}
	cwd_static = utcc_strsav( (char*)s );
#else
	cwd_static = new char[MAXPATHLEN];
	if ( getwd( cwd_static ) == 0) {
	    utcc_printerrno("Getting cwd");
	    utcc_fatalerror("Getting cwd");
	}
#endif 
    }
    return  (String)cwd_static ;
}




void utcc_cat( const String& path )
{
    // cout.form( "--- Start of File: %-20s -----\n", (char*)path );
    istream  source( util_file_expand((char*)path), io_readonly, a_use );
    char* s;
    while ( source.good() ) {
	source.gets( &s ,'\n');
	if ( source.eof() ) break;
	cout << s << "\n";
	delete s;
    }
    // cout.form( "--- End of File: %-20s -----\n", (char*)path );
}

