/* C++ includes */
#include <stream.h>
#include <builtin.h>

/* C includes */
extern "C"
{
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h> 
#include <netinet/in.h>
#include <errno.h>
}

/* Local includes */
#include "VOVprotocol.h"
#include "vovinternal.h"
#include "bzocket.h"
#include "date.h"
#include "voverror.h"
#include "utcc.h"
#include "comm.h"

#define VOV_ERROR_CANNOT_CONNECT 33


extern "C"
{
#include "errtrap.h"

unsigned long inet_addr( char* );
#ifdef vax 
u_short htons(  );
#endif
int socket( );
}
extern int connect( int, void* , int );
extern int net_portNumber( String& );

void VC_synchClocks( Bzocket& zoc )
{
    const int MAX_UNCERTAINTY = 2; // Should not take more than two seconds.
    int max_tries = 3;		   // 
    int uncertainty;		   // In the determination of the delta.
    Date start;			   // ... of synch algo.
    Date end;			   // ... of synch algo.
    
    while ( 1 ) {
	start.now();
	zoc.pack( VOV_SynchClocks );
	zoc.pack( 1 );		// step 1
	zoc.flush();
	
	if (VC_get_reply(zoc) == BUF_ERROR) {
	    vovError("No reply ");
	    return;
	}
	if ( VC_wait_for_code( zoc, VOV_SynchClocks ) != VOV_OK)  {
	    vovError( "Cannot sync clocks" );
	    return;
	}
	int serverclock = zoc.unpackInt();
	int delta = time(0) - serverclock;
	end.now();
	
	uncertainty = end.get() - start.get();
	if ( uncertainty > MAX_UNCERTAINTY ) {
	    vovMessage( form("Failed clock synchronization due to slow network.\n\tTrying again (%d).",max_tries)); 
	    if ( max_tries-- == 0 ) {
		vovMessage( "Could not synchronize clocks." );
		break;
	    }
	} else {
	    zoc.pack( VOV_SynchClocks );
	    zoc.pack( 2 );		// step 2
	    zoc.pack( delta );
	    zoc.pack( uncertainty );
	    zoc.flush();
	    break;
	}
    }
    return;
}


/*
 * RETURN the file descriptor of the new socket
 * or 0 if any error occurs
 */
int VC_connect_to_server( String& hostName, int internetPortNumber)
{
    struct sockaddr_in inaddr;		// INET socket address.
    
    //
    // Get info on the specified host.
    // 
    if ((inaddr.sin_addr.s_addr = inet_addr((char*)hostName)) == -1) {
	struct hostent *host_ptr;
	if ((host_ptr = gethostbyname((char*)hostName)) == NULL) {
	    errRaise("VOV",VOV_ERROR_CANNOT_CONNECT,"Invalid host name: %s ",(char*)hostName );
	}
	// Check the address type for an internet host.
	if (host_ptr->h_addrtype != AF_INET) {
	    errRaise("VOV",VOV_ERROR_CANNOT_CONNECT,"%s is not an internet host", (char*)hostName);
	}
	
	//
	// Set up the socket data.
	//
	bzero((char*) &inaddr, sizeof(struct sockaddr_in)); /* clear that area of memory */
	inaddr.sin_family = host_ptr->h_addrtype;
	bcopy((char *)host_ptr->h_addr, (char *)&inaddr.sin_addr, sizeof(inaddr.sin_addr));
    } else {
	inaddr.sin_family = AF_INET;
    }

    struct sockaddr* addr = (struct sockaddr *)&inaddr;	// Address to connect to.
    int addrlen = sizeof (struct sockaddr_in);		// Actual length of address.
    inaddr.sin_port = htons( internetPortNumber );	// Convert values between host and network byte order.
    
    
    // Open the network socket.
    int fd = 0;
    if ((fd = socket(addr->sa_family, SOCK_STREAM, 0)) < 0) {
	errRaise("VOV", VOV_ERROR_CANNOT_CONNECT, "socket() failed: %s\n\tport number %d", 
		 sys_errlist[errno], internetPortNumber );
    }
    
    //
    // Open the connection to the specified VOV server.
    // Connect can fail if there is no active server, which is a common failure.
    //
    if ( connect(fd, addr, addrlen) == -1) {
	errRaise("VOV", VOV_ERROR_CANNOT_CONNECT, "connect() failed: %s\n\tport number %d", 
		 sys_errlist[errno], internetPortNumber);
    }
    return fd ;			// Return the file descriptor of the connected socket.
}


VovStatus VC_start_up( String& host, String& project, Bzocket& zoc)
{
    int sock_number = net_portNumber( project );
    int fd = VC_connect_to_server( host, sock_number);
    if ( fd == 0 ) {
	return VOV_ERROR;    
    }
    
    zoc.start(fd);			// Start the bzocket.
    //
    // Start up protocol.
    //
    zoc.pack( VOV_StartUp );		// Second packet: Check protocol version and project names
    zoc.pack( VOV_PROTOCOL_VERSION );	// 
    zoc.pack( utcc_getenv( "USER" ) );	// 
    zoc.pack( utcc_gethostname() );	// 
    zoc.pack( getpid() );
    zoc.flush();			// Flush everything.


    
    if ( VC_wait_for_code( zoc, VOV_StartUp ) != VOV_OK ) {
	errRaise( "VOV", VOV_ERROR_CANNOT_CONNECT, "Could not start up connection with server.");
    }
    
    //
    // Check that the project name is OK.
    // 
    String proj =  zoc.unpackString();
    if ( proj != project ) {
	String m = "Connected to the wrong server: server's project is " + proj + " not " + project;
	errRaise( "VOV", VOV_ERROR_CANNOT_CONNECT , (char*)m );
    }
    
    // Do not always synch clocks.
    // VC_synchClocks( zoc );
    
    return VOV_OK;
}





/*
*  The library routines that require a communication
* with the server need a to know the bzocket to use for
* the communication. This routine returns the bzocket,
* and, if it is the case, it opens a new connection.
*/
Bzocket& VC_get_bzocket_core( String& hostName, String& projectName )
{
    static int connected = 0;
    Bzocket* new_zoc = 0;
    if (! connected ) {
	new_zoc = new Bzocket(0);
	if (VC_start_up( hostName, projectName, *new_zoc) == VOV_OK) {
	    connected = 1;
	}
    }
    return  *new_zoc;
}

Bzocket& VC_get_bzocket( String& hostName, String& projectName )
{
    static int connected = 0;
    static Bzocket* zoc_ptr = 0;
    if (! connected ) {
	String h = hostName;
	String p = projectName;
	if ( h == 0 ) {
	    h = utcc_getenv("VOV_HOST_NAME");
	    p = utcc_getenv("VOV_PROJECT_NAME");
	} 
	zoc_ptr = &(VC_get_bzocket_core( h, p ));
	if ( zoc_ptr == 0 ) {
	    vovError("Cannot connect to server " + h + "(project " + p + ")" );
	} else {
	    connected = 1;
	}
    }
    if ( !zoc_ptr->active() ) {
	vovFatalError("Connection no longer active.");
    }
    return  *zoc_ptr;
}



