#include "trace.h"
#include "transition.h"
#include "clients.h"
#include "slaveset.h"
#include "priority.h"		// SortedArray defined here
#include "dispatch.h"

#include "voverror.h"


static SortedArray* scheduledTransactions_gl = 0 ;

void VS_stop_retracing()
{
    if ( scheduledTransactions_gl ) {
	scheduledTransactions_gl->clear();
    }
}





/*
* If no service is requested, the server performs a number of routine tasks
* aimed at checking correctness and consistency of the design. These tasks
* are:
* 	normal checks on the trace;
* 	proper memory management;
* 	check status of all connected clients;
*	verification of goals
*
* NOTE: All these tasks must be light weight, and return after a 
*       very short time (less than a second).
*       More time consuming tasks are dispatched to slaves.
*
*/
void VS_dispatch_job(VovTrace& trace,SlaveSet& slaveset,int mode) 
{
    if ( slaveset.availableSlaves() == 0 ) return;     // If there are no slaves, do not even bother.

    int ARRAY_SIZE = 40;	// Max number of transactions to work with.
    //
    // Allocate array for the best (longest) transactions.
    // 
    if ( scheduledTransactions_gl == 0 ) {
	scheduledTransactions_gl = new SortedArray( ARRAY_SIZE, 0 ); 
    }

    //
    // Find a few transactions which are ready to fire
    // and dispatch them to some slave.
    // Dispatch the longest transactions first. 
    // Use the Sorted Array to find the longest transaction.
    //
    if ( mode == VOV_DISPATCH_MODE_SLOW && scheduledTransactions_gl->length() <= (slaveset.availableSlaves()/2) ) {
	VovBag readySet("TRANSITIONS_READY_TO_FIRE");   // Transactions ready to be fired.
	if ( trace.getAllReadyTransitions( readySet, ARRAY_SIZE ) != VOV_OK ) {
	    return;
	}

	if ( readySet.length( VOV_TRANSITION_OCTMASK ) <= 0 ) {
	    vovError("I expected some transitions in ready set");
	    return;
	}
	// vovMessage( "Looked for and found transitions ready to fire");
	
	scheduledTransactions_gl->clear();
	VovContentsGenerator genTran( &readySet, VOV_TRANSITION_OCTMASK );
	VovTransition t =  0;
	while ( octGenerate( &genTran, &t ) == OCT_OK ) {
	    VOV_ASSERT( t.get() );
	    if (t.getStatus() == VOV_PLAN_NODE ) {
		double expectedDuration = t.duration();
		scheduledTransactions_gl->add( t.id(), expectedDuration );
	    } else if ( t.getStatus() != VOV_RETRACING_NODE ) {
		vovError("Hey, not ready, nor firing!!" + t.format() );
	    } else {
		vovMessage( "Transition is already retracing" );
	    }
	}
    }
    
    //
    // Dispatch as much stuff as you can.
    //
    while (1) {							// 
	octId trans_id = scheduledTransactions_gl->next();	// 
	if ( trans_id == 0 ) break;				// No more transitions availble.
	VovTransition tran = trans_id;				// Get the transition.
	VovStatus s = vov_dispatchJob(trace,tran,slaveset);	// Dispatch the transition to best slave.
	if ( s != VOV_OK ) {					// Dispatching failed.
	    if ( s == VOV_NOT_OK ) {				// No more slaves available.
		scheduledTransactions_gl->undo_next();		// Push transition back.
		break;						// Exit while loop.
	    }							// 
	}							// 
    }								// 
}

//
// Tasks to perform continuously.
//
void VS_routine_tasks_always( VovTrace& trace, SlaveSet& slaveset )
{
    VS_dispatch_job( trace, slaveset, VOV_DISPATCH_MODE_FAST );
}

//
// Routine tasks to perform when the server goes on timeout.
//
void VS_routine_tasks_timeout( VovTrace& trace, SlaveSet& slaveset )
{
    int maxPlacesToCheck = 20;
    trace.checkPlaces( maxPlacesToCheck ); 
    // trace.manageGoals();
    
    VS_dispatch_job( trace, slaveset, VOV_DISPATCH_MODE_SLOW);

    int free_allocated_strings = 1;
    if ( free_allocated_strings ) {
	utcc_strfreeall();
    }

}

//
// Routine tasks to perform when the server slows down and there is
// little activity in the design.
//
void VS_routine_tasks_idle( VovTrace& trace, SlaveSet& slaveset )
{
    String msg;
    trace.OK( msg );
    trace.doPlacement();
    trace.saveOntoDisk();	// Save trace onto disk.
}



