/* Initialization and socket routines */


/* C++ includes */
#include <stream.h>
#include <builtin.h>

/* C includes */
extern "C"
{
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h> 
#include <signal.h>
#include <netinet/in.h>
int setsockopt( int, int, int, void*, int);
int shutdown( int, int );
#ifdef vax
u_short htons(  );
u_long htonl( );
#endif
}

/* Local includes */
#include "clients.h"
#include "intmask.h"
#include "utcc.h"
#include "dialog.h"
#include "voverror.h"

/* 
 * Global variables for this file.
 */
Client*	clientArray_gl = 0;			// Array of all clients.
Intmask selectMask_gl;				// Mask of open connections.
RoundRobinSelector RRhaveMask_gl;		// The mask of all sockets that have packets ready.
int     request_gl = 0;				// The main reception socket for the server.


/*
 * Just a debugging routine.
 */
int checkRequestSocket( int fd )
{
    Intmask mask( fd );		// Get a large enough mask (generally just 1 int ).
    mask.add( fd );		// Set bit
    struct timeval notime = {0, 0}; 	// No wait.
    if (select( mask.last()+1, mask.mask(), 0, 0, &notime) < 0) {
	vovFatalError( "socket is stale.");
    }
    return 1;			// Socket is OK.
}


/*
 * Close down a client, freeing its resources.
 */
void Close_down(int i)
{
    selectMask_gl.del(i);	// 
    if (RRhaveMask_gl.test(i) ) {	// 
	RRhaveMask_gl.del(i);	// 
    }				// 
    RRhaveMask_gl.OK();		// Just for debugging.
    clientArray_gl[i].close();	// 
}

/*
 * Check the status of each client, and close our side of any goners.
 */
void Check_connections()
{
    for (int i = selectMask_gl.first(); i <= selectMask_gl.last(); i++) {
	if (clientArray_gl[i].check() == BUF_ERROR) {
	    Close_down(i);
	}
    }
}

int VS_getdtablesize()
{
    static int n = 0;
    if ( n == 0 ) n = getdtablesize();
    return n;
}


void VS_InitReceive()
{
    int numdesc = VS_getdtablesize();
    
    selectMask_gl.resize( numdesc );
    RRhaveMask_gl.resize( numdesc );
    if ( clientArray_gl == 0 ) {
	clientArray_gl = new Client[numdesc];
    }
    selectMask_gl.clear();
    selectMask_gl.add( request_gl );
    checkRequestSocket( request_gl );
    RRhaveMask_gl.clear();
}

/* 
 * Called from Dispatcher when there are no complete requests to process, or
 * when multiple sockets have requests, or when there is input to be read.
 */
Client* VS_Receive ( int maxwait)
{
    int keepServingCurrentClient = 0;

    if (RRhaveMask_gl.length()) {          			// If the current client still has requests, keep serving it.
	int client = RRhaveMask_gl.index();			// Get index to current client.
	if ( clientArray_gl[client].ready() == BUF_OK ) {	// Make sure client is in fact ready.
	    if ( keepServingCurrentClient ) {
		return &clientArray_gl[client];			// Yes, return client.
	    }
	} else {						// 
	    RRhaveMask_gl.del( client );			// Client is no longer ready.
	}							// 
    }								// 
    while (1) {
	if (RRhaveMask_gl.length()) { 					//If we already have data available,
	    break;							// don't try to read more.
	}								// 
	Intmask mask = selectMask_gl; 					// Prepare selection mask.
	struct timeval waittime = {maxwait, 0};				// Timeout is controlled by function argument.
	int i = select( mask.last()+1, mask.mask(), 0,0, &waittime); 	// Do the selection.
	if (i == -1) {							// Serious error?
	    utcc_printerrno("net_Receive (select)");			// Print error, but try to continue.
	    continue;							// 
	}								// 
	if (i == 0) {							// Time out.
	    Check_connections();					// Check status of all connected clients.
	    return 0;							// Return no legal client.
	}								// 
	mask.fix();							// select() may have changed mask, but not in a consistent way.
	// 
	// If request socket, deal with the new request.
	// 
	if (mask.test(request_gl)) {
	    mask.del(request_gl); 					// 
	    int nfd = accept( request_gl, 0, 0);			// accept() allocates a new file descriptor.
	    if ( nfd == -1 ) {						// 
		utcc_printerrno("in accept");				// accept() returned -1, an error code.
	    } else {							// 
		if ( nfd >= VS_getdtablesize() ) {			// Can we deal with the new file descriptor?
		    utcc_message("Connection refused");			// No, just close it down.
		    close( nfd );					// 
		} else {						// 
		    clientArray_gl[nfd].start(nfd);				// Start the corresponding client.
		    selectMask_gl.add(nfd);				// Add to selection mask.
		}							// 
	    } 
	}
	// 
	// Read from all selected sockets.
	//
	while ( mask.length() ) { 				// While there are sockets to read.
	    int i = mask.first();				// Get the first file descriptor.
	    mask.del(i);					// 
	    if (clientArray_gl[i].read() == BUF_ERROR) {	// Read from socket.
		Close_down(i);					// Connection has dropped,
								// but we must filter event up to the dispatch loop.
		RRhaveMask_gl.add(i);				// This is how we sent it up, 
								// assuming that the client has something to do.
		continue;					// 
	    }							// 
	    if (! clientArray_gl[i].active()) {			// Is client still active ?
		selectMask_gl.del(i);				// No, eliminate it from selection mask.
		continue;					// 
	    }
	    if (clientArray_gl[i].ready() == BUF_OK ) {
		RRhaveMask_gl.add(i);
	    }
	}
    }
    
    if (RRhaveMask_gl.length() > 1) { 			// If there is more than one client, handle multiple requests
	RRhaveMask_gl.next();				// on a round-robin basis.Get the next client.
    }							// 
    int client = RRhaveMask_gl.index();			// 
    return &clientArray_gl[client];				// Return a pointer to the client.
}


void net_Initialize ( int tcpport )
{
    int request = 0;
    if ((request = socket (AF_INET, SOCK_STREAM, 0)) <= 0) {
	utcc_printerrno ("Creating TCP socket");
    } 
    request_gl = request;

    struct sockaddr_in mysock;
    bzero((char*)&mysock,sizeof(struct sockaddr_in));

    mysock.sin_family = AF_INET;
    mysock.sin_port = htons ( tcpport );
    mysock.sin_addr.s_addr = htonl(INADDR_ANY);

    int retry = 6;

    vovMessage( form("Socket  address %d (net=%d)", tcpport, htons(tcpport) ) );
    while (bind (request, (struct sockaddr *) &mysock, sizeof (mysock))) {
	utcc_printerrno( form("Binding TCP socket: retrying %d", retry) );
	if (retry-- <= 0) {
	    // Allow reuse of local address. Useful during debugging of vov.
	    int sockOptionOn = 1;
	    if ( ttyUserDecision(form("Do you really want to reuse the address %d?", tcpport ),"Yes","Abort")==1){
		if ( setsockopt( request, SOL_SOCKET, SO_REUSEADDR, &sockOptionOn, sizeof( sockOptionOn ) ) ) {
		    utcc_printerrno("Setting SOCKET: REUSEADDR");
		}
		retry = 6;
	    } else {
		vovFatalError( form( "Cannot bind socket to port %d.",tcpport ) );
	    }
	}
	(void)sleep( (unsigned)5 );
    }


    // Set socket options:
    // 
    int sockOptionOn = 1; 	// Used to turn on some options.
#ifdef sequent
    if ( setsockopt( request, SOL_SOCKET, SO_DONTLINGER, &sockOptionOn, sizeof(int) ) ) {
	utcc_printerrno("Setting TCP DONTLINGER");
    }
#else 
    static int linger[2]={0,0};
    if ( setsockopt( request, SOL_SOCKET, SO_LINGER, linger, sizeof(linger) ) ) {
	utcc_printerrno("Setting TCP DONTLINGER");
    }
#endif
    // Allow reuse of local address. Useful during debugging of vov.
    if ( setsockopt( request, SOL_SOCKET, SO_REUSEADDR, &sockOptionOn, sizeof(int) ) ) {
	utcc_printerrno("Setting SOCKET: REUSEADDR");
    }
    

    if (listen (request, 5)) {
	utcc_printerrno ("TCP Listening");
    }


    
    selectMask_gl.add(request_gl);

    checkRequestSocket( request_gl );
}

void net_Close()
{
    utcc_message( "Closing connections" );
    if ( shutdown( request_gl, 0 ) == -1 ) {
	utcc_printerrno( "Shutting down server socket" );
    }
}



/* Force connections to close on SIGHUP from init */

void net_HangUp ()
{
    for (int i = selectMask_gl.first(); i <= selectMask_gl.last(); i++) {
	if (clientArray_gl[i].active()) {
	    clientArray_gl[i].close();
	}
    }
    RRhaveMask_gl.clear();
}


