/* C++ includes */
#include "port_internal.h"

/* C includes */
extern "C"
{
#include "errtrap.h"
#include "utility.h"
}

/* Local includes */
#include "trace.h"
#include "place.h"
#include "retrace.h"
#include "vovlib.h"
#include "voverror.h"
#include "path.h"
#include "ddsplus.h"
#include "dialog.h"

extern String vov_signature();
extern String vov_normalizeString( const String& line );
extern void vovEditSet( VovTrace& , VovNodeSet& );
extern void vovEditNodeStatus( VovTrace& , VovNode& );
extern VovPlaceType guessType( String& name );


extern VovStatus SHELL_synchNotify( VovTrace&, int  );


static VovTrace* GlobalTracePointer = 0;

void  vis_placetrace( ddsHandle item )
{
    VovTrace& trace = *GlobalTracePointer;
    static int warned = 0;
    if ( ! warned ) {
	String ret  =  vovConfirm( "CPU INTENSIVE OPERATION", "\
Improving the placement of the nodes in\n\
the trace might require several seconds,\n\
depending on the size of the trace.\n\
\n\
This warning is given only once.\n\n\
Please confirm to continue.\n\
", "Confirm", "I changed my mind" ) ;
	warned = 1;
	if ( ret == "I changed my mind" )  return ;
    }
    trace.doPlacement();
}

void vis_monitorSlaves( ddsHandle item )
{
    if ( util_csystem( "vov_sh -m < /dev/null >& /dev/null &" ) == 127 ) {
	vovError("Failed to start monitor. I wonder why." );
    }
}



typedef struct _retracedata {
    int speed;
    VovTrace* tracePtr;
}  VovLocalRetraceData;

void vov_speed_ccb( ddsHandle item, int offset )
{
    VovDialogList list( item );
    VovLocalRetraceData& data = *(VovLocalRetraceData*)list.getData();
    
    data.speed = (int)list.getWhichOne();
}

void vov_retrace_all_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalRetraceData& data = *(VovLocalRetraceData*)ctrl.getData(); 		// 
    VovTrace& trace = *data.tracePtr;				// 
    String retraceTitle = "All for " + utcc_getenv( "USER" );	// 
    VovRetrace retrace( retraceTitle );				// 
    VOV_ASSERT( trace.create( retrace ) );			// 
    String theText;						// 
    trace.evolve( retrace, data.speed, theText );		// No need to check return code here.
    vovShortMessage( "RETRACE" , theText );			// OK or error, just print the message.
}

void vov_stopRetrace_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalRetraceData& data = *(VovLocalRetraceData*)ctrl.getData(); 		// 
    VovTrace& trace = *data.tracePtr;				// 
    String theText;						// 
    trace.stopRetracing( theText );				// 
    vovShortMessage( "STOPPED RETRACING", theText );		// 
}

void vov_stopTransition_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalRetraceData& data = *(VovLocalRetraceData*)ctrl.getData(); 		// 
    VovTrace& trace = *data.tracePtr;				// 
    trace.stopAllTransitions();					// 
    vovShortMessage( "CONFIRM", "All transitions have been stopped.\n(I hope it worked, this is not fully implemented.)" );
}

void vis_retrace( ddsHandle item )
{
    VovTrace& trace = *GlobalTracePointer;

    static VovLocalRetraceData* dataPtr = 0;
    if ( dataPtr == 0 ) {
	dataPtr = new VovLocalRetraceData;
    }
    VovLocalRetraceData& data = *dataPtr;

    data.tracePtr = &trace;
    data.speed = VOV_RETRACE_SLOW;
    VovDialog top_dialog( "Global retracing dialog", DDS_HORIZONTAL, 1);
    
    VovDialogShowText titleDialog( top_dialog, "", "Retracing speed", 1, 40 );
    VovDialogList speedDialog( top_dialog, DDS_TRUE );

    speedDialog.addItem( "Quick (risky)", VOV_RETRACE_FAST, data.speed == VOV_RETRACE_FAST );
    speedDialog.addItem( "Normal", VOV_RETRACE_SLOW, data.speed == VOV_RETRACE_SLOW );
    speedDialog.endOfList( vov_speed_ccb, (ddsData)&data );

    top_dialog.addControl( "Retrace", 		vov_retrace_all_ccb, (ddsData)&data );
    top_dialog.addControl( "Stop retracing", 	vov_stopRetrace_ccb, (ddsData)&data );
    top_dialog.addControl( "Stop transitions",	vov_stopTransition_ccb, (ddsData)&data );

    
    top_dialog.addHelp( "\
This menu allows you to perform global retracing on the trace.\n\
These operations affect the whole trace, not just your own part of the\n\
trace.\n\
\n\
Use `Normal' retracing for safety. In a few special cases, that is if you\n\
know that no transition is going to change its dependencies, you can use\n\
the `Quick' retracing mode.\n\
\n\
`Retrace': Start retracing of the whole trace.\n\
`Stop retracing': Stop the retracing; no more transitions will be\n\
\tdispatched, but the transitions already dispatched will be\n\
\tcompleted.\n\
`Stop transitions': Stop all transitions currently running.\n\
\tUse for emergencies only.\n\
" );
    top_dialog.postUnmoded();
	
}



void  vis_getscript( ddsHandle item )
{
    VovTrace& trace = *GlobalTracePointer;
    String script;
    VOV_ASSERT( trace.prepareExecutionScript( script ) );
    String wd =  path_canonicalize( utcc_getwd() );
    script.gsub( wd, String(".") );
    vovLongMessage( "EXECUTION SCRIPT", script, 20, 90 );
}

	    
void vis_browser( ddsHandle item )
{
    VovTrace& trace = *GlobalTracePointer;
    extern void vovBrowser( VovTrace& );
    vovBrowser( trace );
}
	    

typedef struct {
    ddsHandle oldpath;
    ddsHandle newpath;
    VovTrace* tracePtr;
} VovLocalMvlibData;


void vov_mvlib_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    
    VovLocalMvlibData& data = *(VovLocalMvlibData*)ctrl.getData();
    VovTrace& trace = *data.tracePtr;
    VovDialogEditText oldPathDlg( data.oldpath );
    VovDialogEditText newPathDlg( data.newpath );
    
    String oldpath = oldPathDlg.getTextItem();
    String newpath = newPathDlg.getTextItem();
    if ( oldpath == 0 || newpath == 0 ) {
	vovConfirm( "ERROR", "Old and new path cannot be null.", "Continue", 0 );
    } else if ( oldpath == newpath ) {
	vovConfirm( "ERROR", "Old and new path must be different.", "Continue", 0 );
    } else {
	VovSelectRule selRule( VOV_PLACE_MASK | VOV_TRANSITION_MASK, VOV_ALL_PLACE_TYPE_MASK,
			      VOV_ALL_STATUS_MASK,
			      String(".*") + oldpath + ".*");
	String setName = "%%% MVLIB %%%";
	
	String msg;
	VovNodeSet set( setName , selRule );
	VOV_ASSERT( trace.findOrCreate( VovSuperObject(set) ) );
	trace.mvlib( set, oldpath, newpath, msg );
	VOV_ASSERT( trace.destroy( set ) );
	vovConfirm( "MVLIB INFO", msg, "Continue", 0 );
    }
} 

void  vis_mvlib( ddsHandle item )
{
    
    VovDialog top_dialog("Move Library Menu", DDS_VERTICAL, 3);

    VovDialogEditText oldPathDlg( top_dialog, "OLD PATH", "", 1 , 50 );
    VovDialogEditText newPathDlg( top_dialog, "NEW PATH", "", 1 , 50 );
    
    
    static VovLocalMvlibData* dataPtr = 0 ;
    if ( dataPtr == 0 ) dataPtr = new VovLocalMvlibData;
    VovLocalMvlibData& data = *dataPtr;
    data.oldpath = oldPathDlg.handle();
    data.newpath = newPathDlg.handle();
    data.tracePtr = GlobalTracePointer;

    top_dialog.addControl( "Ok", vov_mvlib_ccb, (ddsData)&data , "Meta<Key>Return" );
    top_dialog.addHelp("\
This menu is useful if you need to move data around the\n\
file system and you want the server to keep track of the data.\n\
\n\
Type the OLD path name and the NEW path name and then click\n\
the \"Ok\" button.\n\
");
    top_dialog.postUnmoded();
}


void vis_debug( ddsHandle item )
{
    VovTrace& trace = *GlobalTracePointer;
    VOVdebugCommunication( trace );
}


void vis_info( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovDialogList& list = *(VovDialogList*)ctrl.getData();

    VovTrace& trace = *GlobalTracePointer;
    int mask = list.reduceOR();

    String text = trace.getInfo( mask, 1 );
    Date date;
    String title = date.format();
    
    vovLongMessage( title, text, 20, 95 );
}




typedef struct {
    ddsHandle setNameHandle;
    VovTrace* tracePtr;
} VovLocalSet;


VovStatus vis_findset( VovTrace& trace, VovNodeSet& set, String& name )
{
    VovNodeSet tmpset( name );
    if ( trace.find( tmpset ) == VOV_NOT_FOUND ) {
	String l;
	if ( trace.list( VOV_NODE_SET, name, l ) > 0 ) {
	    vovShortMessage( form("SET NAMES MATCHING:`%s'",(char*)name ), (char*)l );
	} else {
	    vovShortMessage( "CANNOT FIND A SET BY NAME", form( "\
Set called \"%s\" does not exist.\nNo set with similar name was found.\n", (char*)name ) );
	}
	return VOV_ERROR;
    } else {
	set.id() = tmpset.id();
	trace.getById( set );
	return VOV_OK;
    }
}

void vov_testsetname_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalSet& data = *(VovLocalSet*)ctrl.getData();
    VovDialogEditText nameDlg( data.setNameHandle );
    String name = nameDlg.getTextItem();
    VovTrace& trace = *data.tracePtr;
    VovNodeSet set;
    if ( vis_findset( trace, set, name ) == VOV_OK ) {
	vovShortMessage( "Set name is good" , (char*)name );
    }
}

void vov_forgetset_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalSet& data = *(VovLocalSet*)ctrl.getData();
    
    VovDialogEditText nameDlg( data.setNameHandle );
    String name = nameDlg.getTextItem();
    VovNodeSet set;
    VovTrace& trace = *data.tracePtr;
    if ( vis_findset( trace, set, name ) == VOV_OK ) {
	if ( trace.destroy( set ) == VOV_OK ) {
	    // vovShortMessage( form("set %s", (char*)name), "Has been deleted." );
	}
    }
}

void vov_editset_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalSet& data = *(VovLocalSet*)ctrl.getData();
    VovDialogEditText nameDlg( data.setNameHandle );
    String name = nameDlg.getTextItem();
    VovNodeSet set;
    VovTrace& trace = *data.tracePtr;
    if ( vis_findset( trace, set, name ) == VOV_OK ) {
	vovEditSet( trace, set );
    }
}


void vis_editset( ddsHandle item )
{
    VovTrace& trace = *GlobalTracePointer;
    VovDialog top_dialog( "Set editing dialog", DDS_VERTICAL, 3 );
    
    static VovLocalSet* dataPtr = 0;
    if ( dataPtr == 0 ) dataPtr = new VovLocalSet;

    VovLocalSet& data = *dataPtr;

    data.tracePtr = &trace;
    static String* staticSetName = 0;
    if ( staticSetName == 0 ) {
	staticSetName = new String;
	*staticSetName = utcc_getenv( "USER" ) + "_set";
    }

    String& defaultSetName = *staticSetName;

    VovDialogEditText setNameDlg( top_dialog, "set name", defaultSetName, 1, 40 );

    data.setNameHandle = setNameDlg.handle();
    
    top_dialog.addControl( "Test name",     vov_testsetname_ccb, (ddsData)&data );
    top_dialog.addControl( "Forget" ,       vov_forgetset_ccb,   (ddsData)&data );
    top_dialog.addControl( "List elements", vov_editset_ccb,     (ddsData)&data );
    
    top_dialog.addHelp( "\
`Test name': test a name to see if it is a proper name of an existing set.\n\n\
`Forget': delete the set from the trace.\n\n\
`List elements': list the elements of this set.\n\n\
");

    top_dialog.postUnmoded();

}

// -----------------------------------------------
// -----------------------------------------------
// -----------------------------------------------
// -----------------------------------------------

typedef struct {
    VovTrace* tracePtr;
    VovSelectRule rule;
    ddsHandle nameDlgPtr;
    VovDialogHandles* handlesPtr;
} VovLocalCreateSetData;

void vov_createset_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalCreateSetData& data = *(VovLocalCreateSetData*)ctrl.getData();

    VovTrace& trace = *data.tracePtr;
    VovDialogEditText nameDlg( data.nameDlgPtr );

    data.rule.readDialog( *data.handlesPtr );
    String name = nameDlg.getTextItem( );
    VovNodeSet set( name, data.rule ); 
    VovNodeSet test_set( name, data.rule );
    VovStatus fs = trace.find( test_set );
    if ( fs == VOV_OK ) {
	if ( vovConfirm( "ATTENTION", form("Set `%s' exists already",(char*)name), "Override", "Don't do it" ) == "Override" ) {
	    trace.destroy( test_set );
	} else {
	    return;
	}
    }

    VOV_ASSERT( trace.create( set ) );
    dds_destroy( item );
}


void  vis_createset( ddsHandle item )
{
    VovTrace& trace = *GlobalTracePointer;

    VovDialog top_dialog( "SELECT NODES IN TRACE", DDS_VERTICAL, 3 );
    VovDialog selRuleDlg( top_dialog, "SELECTION_RULE", DDS_VERTICAL, 3 );

    static VovLocalCreateSetData* dataPtr = 0;
    if ( !dataPtr ) {
	dataPtr = new VovLocalCreateSetData;
    }
    VovLocalCreateSetData& data = *dataPtr;
    
    VovDialogHandles handles( 10 );

    data.rule.set( VOV_PLACE_MASK | VOV_TRANSITION_MASK, VOV_ALL_PLACE_TYPE_MASK, VOV_ALL_STATUS_MASK, SELECT_ALL );
    data.rule.setupDialog( selRuleDlg, handles );

    String dfltSetName = utcc_getenv( "USER" ) + "_set";
    VovDialogEditText dialog( top_dialog, "SET NAME", dfltSetName, 1, 20 );

    top_dialog.addControl( "Ok", vov_createset_ccb, (ddsData)&data );
    top_dialog.addHelp("\
This dialog allows you to create a set of nodes in the trace.\n\
You have to specify a NAME and a SELECTION RULE.\n\n\
");
	
    data.tracePtr = &trace;
    data.nameDlgPtr = dialog.handle();
    data.handlesPtr = &handles;
    top_dialog.postModed();
 
}


void vis_ok( ddsHandle item )
{
    VovTrace& trace = *GlobalTracePointer;
    String msg;
    trace.OK( msg );
    vovShortMessage( "TRACE DOUBLE CHECKED" , msg );
}

void vis_sync( ddsHandle item )
{
    VovTrace& trace = *GlobalTracePointer;
    if ( trace.saveOntoDisk() == VOV_ERROR ) {
	vovShortMessage( "ALARM!!" , "\
Saving to disk has failed!!.\n\n\
Make sure there is enough space on the disk and save again." );
    }
}


//-------------------------------------------------------------
typedef struct {
    VovTrace* tracePtr;
    ddsHandle nameDlgHandle;
} VovLocalDumpData;

void vis_dump_ok( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalDumpData& data = *(VovLocalDumpData*)ctrl.getData();
    VovTrace& trace = *data.tracePtr;
    VovDialogEditText nameDlg( data.nameDlgHandle );
    
    String name = nameDlg.getTextItem();

    if ( trace.saveOntoDisk( 1, name ) == VOV_ERROR ) {
	vovConfirm( "ALARM!!" , "\
Dumping to disk has failed!!.\n\n\
Make sure there is enough space on the disk and save again.", "Continue", 0 );
    } else {
	vovConfirm( "INFO", "Trace has been saved in\n" + name + "\n\n", "Continue", 0 );
    }
    dds_destroy( item );
}

void vis_dump( ddsHandle item )
{
    VovTrace& trace = *GlobalTracePointer;
    VovDialog top( "TRACE ASCII DUMP", DDS_VERTICAL, 3 );

    String name = "/usr/tmp/" + utcc_getenv( "VOV_PROJECT_NAME" ) + "VovDump";
    VovDialogEditText nameDlg( top, "File", name, 1, 40 );

    VovLocalDumpData data;
    data.tracePtr = &trace;
    data.nameDlgHandle = nameDlg.handle();

    top.addHelp( "Create an ASCII dump for the trace.\nPlease choose the name of the file." );
    
    top.addControl( "Ok", vis_dump_ok, (ddsData)&data, "<Key>Return" );
    top.postModed();
}

void  vis_checkall( ddsHandle item )
{
    VovTrace& trace = *GlobalTracePointer;
    trace.checkPlaces( 0  );			// Check all places.
}

// Get notification about one event.
void  vis_synchnotify( ddsHandle item )
{
    VovDialogControl ctrl( item );
    
    int recentOnlyFlag = ( ctrl.name() == "Recent" );
    VovTrace& trace = *GlobalTracePointer;
    SHELL_synchNotify( trace, recentOnlyFlag );
}

void vis_viewtrace( ddsHandle item )
{
    VovTrace& trace = *GlobalTracePointer;
    String text = trace.getInfo( VOV_INFO_CWD, 0 );
    while ( text.length() &&  text.firstchar() != '/' && text.firstchar() != '$' ) {
	text = text.after( RXwhite );
    }
    text = vov_normalizeString( text );
    String facet = text + "/" + trace.getProjectName();
    String command = form( "vem -F %s.page.0:symbolic >& /dev/null &", (char*)facet );
    // vovMessage( command );
    util_csystem( (char*)command );
}

// Always return 0 from this routine.
// It is called by dmDoDialog.
//
int vov_iv_handler( XEvent* evt, ddsData data )
{
    vovMessage("vov_iv_handler::extraneous event" );
}

void addControl( ddsHandle dialog, char* name, char* acc, void (*cb)( ddsHandle ), ddsData data )
{
    ddsControl controlDef;
    controlDef.name  = name;
    controlDef.accelerator = acc;
    controlDef.callback = cb;
    controlDef.user_data = data;
    
    dds_component( dialog, DDS_CONTROL, (ddsData)&controlDef );
}



void vov_exit_ccb( ddsHandle item )
{
    exit( 0 );
}

void vov_interactive_shell( VovTrace& t, int a )
{
    GlobalTracePointer = &t;

    String title = "VOV_SH: " + vov_signature();

    VovDialog top( title, DDS_HORIZONTAL, 2  );
    VovDialog left( top, "MAIN COMMANDS", DDS_VERTICAL, 3 );

    VovDialog browserDlg( left, "BROWSING", DDS_VERTICAL, 3 );
    browserDlg.addControl( "Browse", vis_browser, 0 );

    VovDialog retraceDlg( left, "RETRACING", DDS_VERTICAL, 3 );

    if ( a ) retraceDlg.addControl( "Global", vis_retrace, 0 );
    retraceDlg.addControl( "Check", vis_getscript, 0 );
    retraceDlg.addControl( "Monitor", vis_monitorSlaves, 0 );

    VovDialog setDlg( left, "SETS", DDS_VERTICAL, 3 );
    setDlg.addControl( "Create", vis_createset, 0 );
    setDlg.addControl( "Edit", vis_editset, 0 );

    VovDialog eventDlg( left, "EVENTS", DDS_VERTICAL, 3 );
    eventDlg.addControl( "Recent", vis_synchnotify, 0 );
    eventDlg.addControl( "All", vis_synchnotify, 0 );

    VovDialog miscDlg( left, "VARIOUS", DDS_VERTICAL, 3 );    
    miscDlg.addControl( "Nice placement", vis_placetrace, 0 );
    if ( a ) miscDlg.addControl( "Mvlib", vis_mvlib, 0 );
    if ( a ) miscDlg.addControl( "Check all", vis_checkall, 0 );
    if ( a ) miscDlg.addControl( "Debug I/O", vis_debug, 0 );



    VovDialog diskDlg( left, "TRACE", DDS_VERTICAL, 4 );
    diskDlg.addControl( "Clean", vis_ok, 0 );
    diskDlg.addControl( "To disk", vis_sync, 0 );
    diskDlg.addControl( "View", vis_viewtrace, 0 );
    if ( a ) diskDlg.addControl( "ASCII dump", vis_dump, 0 );
    
	
    VovDialog infoDlg( top, "INFORMATION", DDS_VERTICAL, 4 );
    VovDialogList infoFlagDlg( infoDlg, DDS_FALSE );
    infoDlg.addControl( "Get Info",  vis_info, (ddsData)&infoFlagDlg );

    int infoMask = VOV_INFO_TRACE | VOV_INFO_USER;
    infoFlagDlg.addItem( "TRACE", VOV_INFO_TRACE, (infoMask & VOV_INFO_TRACE) );
    infoFlagDlg.addItem( "SLAVE", VOV_INFO_SLAVE, (infoMask & VOV_INFO_SLAVE) );
    infoFlagDlg.addItem( "SERVER", VOV_INFO_SERVER, (infoMask & VOV_INFO_SERVER) );
    infoFlagDlg.addItem( "USER", VOV_INFO_USER, (infoMask & VOV_INFO_USER) );
    if ( a ) infoFlagDlg.addItem( "DATE", VOV_INFO_DATE, (infoMask & VOV_INFO_DATE) );
    if ( a ) infoFlagDlg.addItem( "STATS", VOV_INFO_STATS, (infoMask & VOV_INFO_STATS) );
    if ( a ) infoFlagDlg.addItem( "CWD", VOV_INFO_CWD, (infoMask & VOV_INFO_CWD) );
    infoFlagDlg.endOfList( 0, 0 );

    top.addHelp( "No help for this dialog yet." );
    top.addControl( "Dismiss", vov_exit_ccb, 0 );
    top.postUnmoded( 0 );	// Do not add dismiss button.

    dds_loop( 0 );		// This will not return.

}




void VC_monitorSlaves( VovTrace& trace, int delay )
{
    String title = "SLAVE ACTIVITY: " + vov_signature();
    VovDialog top_dialog( title, DDS_VERTICAL, 1);

    VovDialogEditText textDialog( top_dialog, "", "XXX", 8, 85 ); 
    
    int sdelay = 1;		// Service delay on this dialog.
    top_dialog.addHelp( "\
This window shows the activity of the slaves used\n\
for this design.\n\
");
    top_dialog.addHelp( form("This dialog is updated every %ds\nand serviced every %ds.\n\n",delay,sdelay));
    top_dialog.addControl( "Dismiss", vov_exit_ccb, 0 );
    top_dialog.postUnmoded( 0 ); // Do not add dismiss button
    
    
    // This infinite loop is only exited by hitting the  Dismiss button.
    while ( 1 ) {
	String text = trace.getInfo( VOV_INFO_SLAVE, 0 );
	textDialog.modifyText( text );
	int count = delay;
	while ( count-- ) {
	    dds_outstanding( 0 ) ;
	    sleep( sdelay );
	}
    }
}



