/* C++ includes */
#include "port_internal.h"

/* C includes */
extern "C"
{
#include "options.h"		// Octtools
#include "utility.h"		// Octtools
#include "errtrap.h"
#include <signal.h>
}

/* Local includes */
#include "vovlib.h"
#include "trace.h"
#include "place.h"
#include "retrace.h"
#include "voverror.h"
#include "path.h"
#include "dialog.h"
#include "version.h"
#include "ddsplus.h"


#define VHSHELL_NO_SERVER "No server found.\n\
\tCheck your environment variables VOV_HOST_NAME and VOV_PROJECT_NAME.\n\
\tIf they are OK, make sure the server you need is still alive.\n\
"

optionStruct optionList[] = {
    { "A", 0,		"Advanced user flag" },
    { "D", 0,		"Dump trace in ASCII format (used for emergency save)" },
    { "E", 0,		"Show all events during retracing" },
    { "X", 0,           "Exit immediately after starting a retracing" },
    { "F", 0,		"Set speed of retracing to FAST (default SLOW)" },
    { "H", "params",	"History parameters. Examples: -T30, +N2, -TP1000." },
    { "L", 0,		"Local trace (expert users only)." },
    { "K", 0,		"Kill server" },
    { "N", "string",    "New string for mvlib function" },
    { "O", "string",    "Old string for mvlib function" },
    { "R", 0,		"Retrace all in this directory (or everywhere with -A)" },
    { "S", 0,  		"Stop all retracing"},
    { "V", 0, 		"Print version number" },
    { "d", 0,		"Debug communication with server"},
    { "p","project",	"Specify a different project name than VOV_PROJECT_NAME." },
    { OPT_CONT, 0,      "" },	// Leave an empty line here.

    { "t", "name",	"Toggle status of place" },
    { "u", "name",	"Is the place used in the design?" },
    { "w", 0,		"Write trace onto disk (also force check of all places)"},
    { "f", "place",     "Forget place (Use with caution!!)" },
    { "e", "place",	"Edit place"},
    { "r", "place",	"Retrace  TO  specified place (retrace the place)"},
    { "k", "place",	"Retrace FROM specified place (kick the place)" },
    { "h", "name",	"Print history of place" },
    { "m", 0,		"Monitor slaves" },
    { "n", 0,		"Notification: get all unreported events" },
    { "i", 0,		"Interactive" },
    { "I", 0,		"Get info about design"},
    { 0, 0, 0 }
};

extern void vov_interactive_shell( VovTrace&, int advancedFlag );
extern void vovEditNodeStatus( VovTrace& , VovNode& );
extern String vov_normalizeString( const String&  );


//  cell:view	-> VOV_OCT_FACET	( contains a colon ':')
//  file	-> VOV_UNIX_FILE	( everything else )
VovPlaceType guessType( String& name )
{
    if (name.contains(":")) {
	return VOV_OCT_FACET;
    } else {
	return VOV_UNIX_FILE;
    }
}


SIGNAL_FN cleanExit()
{
    vovMessage(" Signal caught " );
}

void shellSignals()
{
    // extern int sigvec( int, struct sigvec* , struct sigvec* );
    struct sigvec sig, osig;
    sig.sv_handler = cleanExit;
    sig.sv_mask = 0;
    SIG_FLAGS(sig) = 0;
    sigvec( SIGHUP, &sig, &osig );
    return;
}

VovStatus SHELL_synchNotify( VovTrace& trace , int unreported)
{
    const int MAX_EVENTS = 400;
    VovEventQueue queue(trace, MAX_EVENTS);
    
    int count = VOVsynchNotify( trace, queue, MAX_EVENTS, unreported );
    
    if ( vovInitX() ) {
	if ( count == 0 ) {
	    vovConfirm( "MESSAGE", "No events in the queue.", "Dismiss", 0 );
	} else {
	    queue.edit();
	}
    } else {
	if ( count == 0 ) {
	    vovMessage( "No events in the queue." );
	} else {
	    VovEvent event;
	    while ( queue.length() ) {
		queue.deq( event );
		vovMessage( event.format() );
	    }
	}
    }
    return VOV_OK;
}

VovStatus VEF_showAllEvents( VovTrace& trace, VovEvent& event )
{
    vovMessage( event.getDescription() );
    return VOV_OK;
}

VovStatus VEF_monitorRecompilation( VovTrace& trace, VovEvent& event )
{
    VovEventFilter printCriteria ( VOV_PLACE_MASK, VE_CHANGE, ".std." );
    VovEventFilter msgCriteria( VOV_TRANSITION_MASK, VE_START | VE_ERROR, SELECT_ALL );
    VovEventFilter msg1Criteria(  VOV_RETRACE_MASK, VE_START | VE_END, SELECT_ALL );

    if ( printCriteria.filter( event ) == VOV_OK ) {
	// Cannot do a getById because the socket is busy getting events.
	String name = event.getDescription();
	name = vov_normalizeString( name );
	if ( name.contains(' ') ) {
	    name = name.before( " " ); 		// Just get the first element of the string,
	}
						// That should be the path name.
	utcc_cat( name );
    }
    if ( msgCriteria.filter( event ) == VOV_OK || msg1Criteria.filter( event ) == VOV_OK ) {
	cerr << event.getDescription();
    }
    return VOV_OK;
}

VovStatus VEF_monitorRetracing( VovTrace& trace,  VovEvent& event )
{
    cerr <<  event.getDescription();
    return VOV_OK;
}

void decodeHistoryParam( String& historyParam, int& mask, int&  dir, int& level )
{
    char* s = utcc_strsav( (char*)historyParam );
    int stop = 0;
    for ( char* c = s; *c != '\0' && ! stop ; c++ ) {
	switch ( *c ) {
	case '+':
	    dir = VOV_FORWARD;
	    break;
	case '-':
	    dir = VOV_BACKWARD;
	    break;
	case 't':
	case 'T':
	    mask |= VOV_TRANSITION_MASK;
	    break;
	case 'p':
	case 'P':
	    mask |= VOV_PLACE_MASK;
	    break;
	case 'N':
	case 'n':
	    mask |= VOV_NODE_MASK;
	    break;
	case '0'..'9':
	    level = atoi( c );
	    stop = 1;
	    break;
	default:
	    vovError( "Illegal history parameters string: " + historyParam );
	    vovMessage( "Format: {-}{T|t}{P|p}{N|n}<number>" );
	    stop = 1;
	}
    }
    // vovMessage( form( "%s -> mask %d dir %d , lev %d", s, mask, dir, level ) );
    return;
}

VovStatus findPlaceByName( VovTrace& trace, VovPlace& place, String& name )
{
    for ( VovPlaceType t = VOV_FIRST_PLACE_TYPE + 1; t != VOV_LAST_PLACE_TYPE; t++ ) {
	VovPlace p( t, name  );			// A temporary place.
	if ( trace.find( p ) == VOV_OK ) {
	    place = p;
	    return VOV_OK;
	} 
    }
    return VOV_NOT_FOUND;
}

void vovShellOpenTrace( VovTrace& trace, String& project, String& host, int type )
{
    static int done = 0;	// Execute only once.
    if ( done ) return;


    if ( trace.open( project, host, "a", type ) != VOV_OK ) {
	if ( trace.disabled() ) {
	    vovError( VHSHELL_NO_SERVER );
	} else {
	    vovError( "Cannot open trace" );
	}
	exit( -1 );
    }
    done = 1;
}



main(int argc, char* argv[])
{
    extern int optind;		// Declared in the option package.
    extern char* optarg;	// Declared in the option package.
    int option;			// The command line option.
    
    String theText;		// The text returned by several routines.
    
    String oldString;		// Used by mvlib.
    String newString;		// Used by mvlib.
    String historyParams = "-T8"; // Backwards, Transitions, 8 levels. 


    vovRegisterMainArgs( argc, argv ); // For efficiency, it might be better to skip this sometimes.


    errProgramName( argv[0] );	// Because some initializer can also fail.
    errCore( 1 );
    
    int tracetype = VOV_TRACE_TYPE_CLIENT;
    
    String host = utcc_getenv( "VOV_HOST_NAME" );
    String project = utcc_getenv( "VOV_PROJECT_NAME" );
    
    VovTrace trace;
    
    VOVcatchAllSignals();
    
    int retracingSpeed = VOV_RETRACE_SLOW;
    int advancedUser = 0;
    int exitAfterRetracing = 0;
    VovEventFilterFunction VEF_function = VEF_monitorRecompilation;
    
    while (( option = optGetOption( argc, argv ) ) != EOF) {
	switch (option) {
	case 'A':
	    advancedUser = 1;
	    break;
	case 'X':
	    exitAfterRetracing = 1;
	    break;
	case 'E':
	    VEF_function = VEF_monitorRetracing;
	    break;
	case 'p':
	    project = optarg;
	    break;
	case 'L':
	    advancedUser = 1;
	    tracetype = VOV_TRACE_TYPE_LOCAL;
	    host = "<local>";
	    break;
	case 'i':
	    vovShellOpenTrace( trace, project, host, tracetype );
	    vov_interactive_shell( trace, advancedUser );
	    break;
	case 'F':
	    if ( advancedUser ) {
		retracingSpeed = VOV_RETRACE_FAST;
	    } else {
		vovError( form("You must be an advanced user to use option %c.", option) );
	    }
	    break;
	case 'S':
	    vovShellOpenTrace( trace, project, host, tracetype );
	    trace.stopRetracing( theText );
	    if ( advancedUser ) trace.stopAllTransitions();	// 
	    vovMessage( theText );
	    break;
	case 'R':
	    { 
		VovStatus s;
		String wd = path_canonicalize(utcc_getwd());

		String retraceLabel = advancedUser? "All for " + utcc_getenv("USER" ) : "Dir " + wd;

		VovRetrace retrace( retraceLabel );

		if ( advancedUser ) { 
		    vovShellOpenTrace( trace, project, host, tracetype );
		    VOV_ASSERT( trace.create( retrace ) );
		    s = trace.evolve( retrace, retracingSpeed, theText );
		} else {
		    /* Retrace all stuff in current directory */
		    vovShellOpenTrace( trace, project, host, tracetype );


		    VovSelectRule selRule( VOV_PLACE_MASK, VOV_ALL_PLACE_TYPE_MASK,
					  VOV_ALL_STATUS_MASK, wd + ".*" );
		    VovNodeSet nodeSet( "%%%"+wd, selRule );
		    VOV_ASSERT( trace.findOrCreate( nodeSet ) );

		    VOV_ASSERT( trace.create( retrace ) );
		    s = trace.evolve( retrace, nodeSet, retracingSpeed, theText );
		}
		vovMessage( theText );
		if ( s == VOV_OK && !exitAfterRetracing) {
		    VovEventFilter stopCriteria( VOV_RETRACE_MASK, VE_END, retrace.id() );
		    VOVasynchNotify( trace, VEF_function , stopCriteria);
		}
	    }

	    break;
	case 'H':
	    historyParams = optarg;
	    break;
	case 'u':
	    {
		vovShellOpenTrace( trace, project, host, tracetype );
		VovPlace place;
		String name = path_localFull( optarg );
		char* type;
		if ( findPlaceByName( trace, place, name ) != VOV_OK ) {
		    type = "NOT FOUND";
		} else {
		    int mask = trace.testConnectivity( place );
		    if ( mask == 0 ) {
			type =  "ISOLATED";
		    } else if ( ( mask & VOV_INPUT_MASK ) == 0  ) {
			type =  "PRIM INPUT";
		    } else if ( ( mask & VOV_OUTPUT_MASK ) == 0  ) {
			type =  "PRIM OUTPUT";
		    } else {
			type =  "USEFUL";
		    }
		}
		cout.form( "%14s %s\n", type, optarg );
	    }
	    break;
	case 'r':
	case 'k':
	case 'h':
	case 't':
	case 'e':
	case 'f':
	    {
		vovShellOpenTrace( trace, project, host, tracetype );
		VovPlace place;
		String name = path_canonicalize( optarg );
		if ( findPlaceByName( trace, place, name ) == VOV_OK ) {
		    switch ( option ) {
		    case 'r':
		    case 'k':
			{
			    String dirString = "to ";
			    int dir = VOV_BACKWARD;
			    if ( option == 'k' ) {
				dirString = "from ";
				dir = VOV_FORWARD;
			    }
			    VovRetrace retrace( dirString + place.getCommonName() + " for " + utcc_getenv("USER" ) );
			    VOV_ASSERT( trace.create( retrace ) );
			    VovStatus s = trace.evolve( retrace, place, dir, retracingSpeed,  theText);
			    vovMessage( theText );
			    if ( s == VOV_OK && !exitAfterRetracing) {
				VovEventFilter stopCriteria( VOV_RETRACE_MASK, VE_END, retrace.id() );
				VOVasynchNotify( trace, VEF_function, stopCriteria );
			    }
			}
			break;
		    case 'h':
			{
			    int mask = 0;
			    int dir = VOV_BACKWARD;
			    int level = 2;
			    decodeHistoryParam( historyParams, mask, dir, level );
			    trace.history( place, mask, dir, level, theText);
			    cout << "\n" << theText << "\n";
			}
			break;
		    case 't':
			{
			    VovNodeStatus s = place.getStatus();
			    VovNodeStatus ns = s;
			    int flags = 0;
			    switch ( s ) {
			    case VOV_PLAN_NODE:
				ns = VOV_TRACE_NODE;
				flags = VOV_CHANGE_STATUS_SELF;
				break;
			    case VOV_TRACE_NODE:
				ns = VOV_PLAN_NODE;
				flags = VOV_CHANGE_STATUS_SELF_REC;
				break;
			    default:
				vovError("Use X interface to change status of this place." );
			    }
			    VOV_ASSERT( trace.changeStatus( place, ns, flags ) );
			    vovMessage( vov_getStatusString( s ) + " -> " + vov_getStatusString( ns ) );
			}
			break;
		    case 'e':
			trace.edit( place );
			break;
		    case 'f':
			VOV_ASSERT( trace.destroy( place ) );
			vovMessage( place.format()  + " forgotten." );
			break;
		    default:
			break;
		    }
		    break;
		} else {
		    vovError( name + " no place with that name." );
		}
	    }
	    break;
	case 'd':
	    if ( advancedUser ) {
		vovShellOpenTrace( trace, project, host, tracetype );
		VOVdebugCommunication( trace );
	    } else {
		vovError( form("You must be an advanced user to use option %c.", option) );
	    }
	    break;
	case 'm':
	    extern void VC_monitorSlaves( VovTrace& trace, int delay );
	    vovShellOpenTrace( trace, project, host, VOV_TRACE_TYPE_CLIENT );
	    VC_monitorSlaves( trace, 5 );
	    exit( 0 );
	    break;
	case 'n':
	    vovShellOpenTrace( trace, project, host, tracetype );
	    SHELL_synchNotify( trace, 1 );
	    break;
	case 'N':
	case 'O':
	    {
		static int bothargs = 0;
		if ( option == 'N' ) { newString = optarg; bothargs |= 01; }
		if ( option == 'O' ) { oldString = optarg; bothargs |= 02; }
		if ( bothargs == 03 ) {
		    vovShellOpenTrace( trace, project, host, tracetype );
		    VovSelectRule selRule( VOV_PLACE_MASK | VOV_TRANSITION_MASK, VOV_ALL_PLACE_TYPE_MASK,
					  VOV_ALL_STATUS_MASK,
					  String(".*") + oldString + ".*");
		    String setName = "%%% MVLIB %%%";
		    VovNodeSet set( setName , selRule );
		    VOV_ASSERT( trace.findOrCreate( set ) );
		    String msg;
		    trace.mvlib( set, oldString, newString, msg );
		    vovMessage( msg );
		    VOV_ASSERT( trace.destroy( set ) );
		    // SHELL_synchNotify( trace, 1 );
		}
	    }
	    break;
	case 'I':
	    {
		vovShellOpenTrace( trace, project, host, tracetype );
		int mask = VOV_INFO_SERVER | VOV_INFO_TRACE | VOV_INFO_USER ;
		if ( advancedUser ) {
		    mask |=  VOV_INFO_STATS;
		}
		theText = trace.getInfo( mask, advancedUser );
		cout << "\n";
		cout << theText;
		cout << "\n";
	    }
	    break;
	case 'D':
	    if ( advancedUser ) {
		vovShellOpenTrace( trace, project, host, tracetype );
		vovMessage( "Dumping trace to disk" );
		VOV_ASSERT( trace.saveOntoDisk( 1 ) );
		vovMessage( "The trace dump is in /usr/tmp/vovdump.<projectname>.\n\tMake sure to copy it somewhere else." );
	    } else {
		vovError( form("You must be an advanced user to use option %c.", option) );
	    }
	    break;
	case 'w':
	    {
		String msg;
		vovShellOpenTrace( trace, project, host, tracetype );
		trace.checkPlaces( 0 );
		trace.OK( msg );
		vovMessage( msg );
		if ( trace.saveOntoDisk() == VOV_ERROR ) {
		    for ( int i = 0 ; i < 5 ; i++ ) vovError( "Error Error Error" );
		    vovError("\tSaving to disk has failed.\n\tMake sure there is enough free space on disk\n\tand save again." );
		}
	    }
	    break;
	case 'K':
	    if ( advancedUser ) {
		vovShellOpenTrace( trace, project, host, tracetype );
		if ( VOVkillServer( trace ) == VOV_OK ) {
		    vovMessage( "Server got the message");
		} else {
		    vovMessage( "Kill aborted" );
		}
	    } else {
		vovError( form("You must be an advanced user to use option %c.", option) );
	    }
	    break;
	case 'V':
	    cout.form("VOV_SH: %s: %s@%s\n", VOV_VERSION, (char*)project, (char*)host );
	    break;
	default:
	    break;
	}
    }
    if ( argc==1 ) optUsage();	// Need some options to do something.
    
    if ( tracetype == VOV_TRACE_TYPE_LOCAL ) {
	VOV_ASSERT( trace.saveOntoDisk() );
    }
    
    exit( 0 );
}






