#include <stream.h>
#include <grp.h>


extern "C"
{
#include "options.h"
}

/* Local includes */
#include "trace.h"
#include "transition.h"
#include "place.h"
#include "bzocket.h"
#include "vovlib.h"
#include "utcc.h"
#include "VOVprotocol.h"
#include "ddsplus.h"

#include "voverror.h"
#include "comm.h"
#include "version.h"

extern int slave_exec( VovTrace& );
extern double vov_loadOnLocalHost();
extern double estimate_mips();

extern VC_interactiveSlave( VovTrace& trace );

String vov_slaveAffinity()
{
    static char* rtn = 0;
    if ( rtn == 0 ) {
	String s = utcc_gethostname() + " " +  utcc_machine();
	rtn = utcc_strsav( (char*)s );
    } 
    return rtn;
}

int vov_availableMIPS( double max_load, int coeff, int just_finished )
{
    static double est_power = 0.0; 		// Compute this just once.
    if ( est_power == 0 ) {			// 
	est_power = estimate_mips();		// 
    }						// 
    double load = vov_loadOnLocalHost();	// .. and this each time the routine is called.
    const double CREDIT = 0.9;			// 
    if ( just_finished && load > CREDIT) {	// 
	load -= CREDIT;				// Give some credit to the fact that we just finished a job.
    }						// 
    int power = 0;				// 
    if ( load <= max_load ) {			// 
	power = (int)(est_power/(load + 1.0));  // The conversion double->int maintains enough precision.
	power /= 1000;			        // Scale the power down by this factor.
    }						// 
    return power / coeff;			// 
}

VovStatus vov_updateMIPSinfo( VovTrace& trace, int effectiveMIPS, String& affinity )
{
    Bzocket& zoc = trace.getBzocket();
    zoc.pack( VOV_SlaveReady );
    zoc.pack( effectiveMIPS );
    zoc.pack( affinity );
    zoc.flush();

    // vovMessage(form("Updated slave info: %d %s", effectiveMIPS, (char*)affinity ));

    return VOV_OK;
}

void VC_slave( VovTrace& trace, int coeff, double max_load, String& affinity )
{
    Bzocket& zoc = trace.getBzocket();
    int alarmInterval = 120; 				// Delay between updates.
    int duration = alarmInterval;			// Duration of transitions since last update:
							//  initialized to a large value to force early timeout.
    while (1) {						// 
	int timeOutInterval = ( alarmInterval - duration ) >? 2;
	int s =  zoc.waitForPacket( timeOutInterval );  // 
	if ( s == BUF_ERROR ) {				// 
	    vovError( "Slave lost connection." );	// 
	    break;					// Slave exits.
	} else if ( s == BUF_TIMEOUT ) {		// 
	    int effectiveMIPS = vov_availableMIPS( max_load, coeff, 0 ) ;
	    vov_updateMIPSinfo(trace,effectiveMIPS,affinity);  
	    duration = 0;				// Reset duration.
	    continue;					// Loop again.
	}						// 
	if ( VC_get_reply( zoc ) != VOV_OK ) { 		// Get rest of packet.
	    break;					// Lost connection.
	}						// 
	int code = zoc.unpackInt();			// Get packed op-code.
	if ( code == VOV_Exec ) {			// 
	    // vovMessage( "exec request" );
	    duration += slave_exec( trace );		// 
	    int effectiveMIPS = vov_availableMIPS( max_load, coeff, (duration > alarmInterval) ) ;
	    vov_updateMIPSinfo( trace, effectiveMIPS, affinity ) ;
	    // vovMessage( "done" );
	    duration = 0;				// Reset duration.
	} else if ( code == VOV_KillSlave ) {
	    vovMessage( "Slave instructed to exit." );
	    break;
	} else {					// 
	    vovError( form( "Unexpected code %d.", code ) );
	    break;
	}
    }
}




optionStruct optionList[] = {
    { "d", 0,		"Debug communication slave <-> server" },
    { "g", "group name","Specify group name (for permissions)" },
    { "A", "affinity",  "Affinity resources for this slave" },
    { "M", "max_load",  "Maximum allowed load on the slave host (default 2)" },
    { "h", "host",      "Specify host (default is environment var.  VOV_HOST_NAME)" },
    { "p", "project",    "Specify project (default is environment var. VOV_PROJECT_NAME)" },
    { "c", "coefficient", "Slave coefficient in range [1,2,3,...]" },
    { "s", 0,		"Silent mode" },
    { "l", "logfile", 	"Logfile" },
    { "i", 0,		"Interactive slave" },
    { 0,   0,		0 }
};

main( int argc , char* argv[] )
{
    extern int optind;
    extern char* optarg;
    double max_load = 20.0;	// Max load allowed on host
    int coeff = 1;		// 
    int silent = 0;		//
    int interactive = 0;	// Flag for an interactive slave.
    int debugflag = 0;		// 
    String affinity;

    String host;
    String project;
    String logfile;

    char option;
    while (( option = optGetOption( argc, argv ) ) != EOF) {
	switch (option) {
	case 'h':
	    host = optarg;
	    util_setenv( "VOV_HOST_NAME", host );
	    break;
	case 'p':
	    project = optarg;
	    util_setenv( "VOV_PROJECT_NAME", project );
	    break;
	case 'M':
	    {
		double min_maxload = 0.2;   // Some bounds that make sense
		double max_maxload = 20.0;  // 
		max_load = atof( optarg );
		if ( max_load < min_maxload || max_load > max_maxload ) {
		    vovMessage( "max_load is weird. I am bringing it within reasonable limits.");
		    if ( max_load > max_maxload ) max_load = max_maxload;;
		    if ( max_load < min_maxload ) max_load = min_maxload;;
		}
	    }
	    break;
	case 'c':
	    coeff = atoi( optarg );
	    if ( coeff <= 0 ) {
		vovError("coefficient must  be positive and integer");
		exit ( -1 );
	    }
	    break;
	case 'A':
	    affinity += form( "%s ", optarg );
	    break;
	case 'g':
	    {
		struct group* Group = getgrnam( optarg );
		if ( Group == 0 ) {
		    vovFatalError( "No such group:" + String(optarg) );
		}
		if ( setregid( -1, Group->gr_gid ) == -1 ) {
		    utcc_printerrno("slave:main()");
		    vovFatalError("Cannot set group id" );
		}
	    }
	    break;
	case 'l':
	    logfile = optarg;
	    break;
	case 'd':
	    debugflag = 1;
	    break;
	case 's':
	    silent = 1;
	    break;
	case 'i':
	    interactive = 1;
	    break;
	default:
	    optUsage();
	    break;
	}
    }
    
    if (host == 0)     host = utcc_getenv("VOV_HOST_NAME");
    if (project == 0 ) project = utcc_getenv("VOV_PROJECT_NAME");
    
    if ( silent ) {
	if ( logfile == 0 ) {
	    String localhost = utcc_gethostname();
	    logfile = form("/usr/tmp/%s.%s.%s",(char*)localhost,(char*)project,(char*)host);
	}
	if (freopen( (char*)logfile, "w", stdout) == 0) {
	    utcc_printerrno( "Cannot redirect stdout" );
	}
	if (dup2( fileno(stdout) , fileno(stderr) ) == -1 ) {
	    utcc_printerrno("cannot dup2 stderr ");
	}
	if ( chmod( (char*)logfile, 0666 ) == -1 ) {
	    utcc_printerrno( "Cannot change mode to file" );
	}
    }

    // Print a banner for identification of version number.
    if ( ! interactive ) {
	cout << "VOV_SLAVE: " << VOV_VERSION << " " << project << "@" << host << "\n";
    }
    
    VOVcatchAllSignals();
    VovTrace trace( project, host, "a", VOV_TRACE_TYPE_CLIENT );
    
    if ( trace.disabled() ) {
	vovError( "No server. Slave must exit." );
	exit( 1 );
    } 
    if ( affinity == 0 ) affinity = vov_slaveAffinity();
    extern String vov_normalizeString( const String& line ); // Not a proper use of this function.
    affinity = vov_normalizeString( affinity );
    //    util_setenv( "VOV_NO_ECHO", "true" );

    if ( interactive ) {
	vovRegisterMainArgs( argc, argv );
	VC_interactiveSlave( trace );
    } else {
	if ( debugflag ) {
	    VOVdebugCommunication( trace );
	}
	VC_slave( trace, coeff, max_load, affinity );
    }

    exit( 0 );
}    
      
