#include <String.h>
#include "path.h"
#include "trace.h"
#include "voverror.h"
#include "utcc.h"
#include "utility.h"

extern "C"
{
#undef umask   /* Because of a conflict in declarations (not used anyway) */
#include <sys/types.h>
#include <sys/stat.h>
typedef struct stat stat_struct;
extern int stat();
}

int unix_info(char* path, int* exists, int* mode, Date* timeStamp)
{
    stat_struct buf;
    path = util_logical_expand( path );
    if (  stat( path, &buf ) == -1) {
	*exists = 0;
	*timeStamp = 0;
	*mode = 0;
    } else {
	*exists = 1;
	*mode = buf.st_mode;
	timeStamp->set( buf.st_mtime );
	// timeStamp->set( path_host( (String)path ) );
    }
    return *exists;
}


int unix_exists(char* path)
{
    stat_struct buf;
    path  = util_logical_expand( path );
    if (  stat( path, &buf ) == -1) {
	return 0;
    }
    return 1;
}


int unix_cp( char* from, char* to )
{
    // Do not copy if the source does not exist or if
    // the destination exists already.
    if ( unix_exists( to ) || ! unix_exists( from ) ) return -1;

    File fold( util_file_expand(from), io_readonly, a_use );
    File fnew( util_file_expand(to), io_writeonly, a_createonly );

    if ( fold.bad() ) return -1;
    if ( fnew.bad() ) return -1;
    fold.raw();
    fnew.raw();
    const int SIZE = 4096;
    char buf[SIZE];
    while ( 1 ) {
	int n = read( fold.filedesc(), buf, sizeof buf );
	if ( n == 0 ) break;
	if ( n < 0 ) {
	    utcc_printerrno( from );
	    return -1;
	} 
	if ( write( fnew.filedesc(), buf, n ) != n ) {
	    utcc_printerrno( to );
	    return -1;
	}
    } 
    return 0;
}


/* 
* Use rename rather than unlink, at least for the time being 
*/
int unix_delete(char* path)
// Returns 0 if deletion fails, 1 if it succeds.
{
    String garbageCan = "VOV_GARBAGE_CAN";
    String date = dec( getDate() );
    String newName = garbageCan + "/" + date + path_lastComponent( path ) ;

    if ( ! path_test( path ) ) {
	vovMessage( form("%s does not exist:\n\tcannot be deleted",path) );
	return 0;
    }
    if ( ! path_test( garbageCan ) ) {
	if ( mkdir( (char*)garbageCan, 0777 ) == -1 ) {
	    utcc_printerrno( "cannot create directory" );
	    return 0;
	}
    }
    
    // If regular rename fails, fake a renaming by hand.

    if ( rename( path, (char*)newName ) == -1 ) {
	vovMessage( String("Cannot rename file ") + path + " to " + newName );
	utcc_printerrno( "rename file" );
	if ( unix_cp( path, (char*)newName ) == -1 ) {
	    utcc_printerrno( "copying file to be deleted" );
	    return 0;
	}
	if ( unlink( path ) == -1 ) {
	    utcc_printerrno( "unlinking" );
	    return 0;
	}
    }
    return 1;
}




void unix_chmod(char* path , int m )
{
    int s = chmod( util_file_expand(path), m );
    if ( s == -1 ) {
	utcc_printerrno( path );
    }
}

int unix_valid(char* path)
{
    return unix_exists( util_file_expand(path) );
}


int unix_timestamp(char* path)
{
    stat_struct buf;
    path  = util_file_expand( path );
    if (  stat( path, &buf ) == -1) {
	utcc_printerrno( path );
	return 0;
    }
    return  buf.st_mtime;
}

int unix_size(char* path)
{
    stat_struct buf;
    path  = util_file_expand( path );
    if (  stat( path, &buf ) == -1) {
	utcc_printerrno( path );
	return 0;
    }
    return  buf.st_size;
}

char* unix_read_file( char* path )
/*
 * This is David Harrison's code, turned into C++.
 * Reads text from the given file. 
 * It is responsibility of the user to free the memory.
 */
{
    FILE *fp;

    int chunk = 2048;
    char* result = new char[chunk];
    char* ptr = result;
    int count = 0;

    if ( fp = fopen( util_file_expand( path ), "r") ) {
	char ch;
	while (( ch = getc(fp) ) != EOF ) {
	    count++;
	    if ((ptr-result) >= chunk) {
		int temp = ptr - result;
		chunk *= 2;
		result = (char*)realloc( result, chunk );
		ptr = result+temp;
	    }
	    *ptr++ = ch;
	}
	if ( (ptr-result) >= chunk ) {
	    int temp = ptr - result;
	    chunk *= 2;
	    result = (char*)realloc( result, chunk );
	    ptr = result+temp;
	}
	*ptr = '\0';		// Terminate the string.
	return result;
    } else {
	return utcc_strsav( form( "Can't open file `%s'", path) );
    }
}

int unix_write_file( char* path, char* text )
{
    FILE* fp;
    if ( ( fp = fopen( util_file_expand(path), "w") ) == 0 ) {
	utcc_printerrno( "opening for write" );
	return -1;
    }
    char* ptr = text;
    while ( *ptr != '\0'  ) {
	if ( fputc( *ptr++, fp ) == EOF ) {
	    utcc_printerrno( "putting chars" );
	    return -1;
	}
    }
    if ( fclose( fp ) == EOF ) {
	utcc_printerrno( form( "closing %s", path  ) );
    }
    return 0;
}
