// Easy interface to C++ programs.
// Make it easy for programmers to vov-ize their programs.

#include "String.h"
#include "port.h"
#include "utcc.h"
#include "ddsplus.h"
#include "voverror.h"
#include "vovlib.h"
#include "trace.h"
#include "transition.h"
#include "user.h"
#include "place.h"


extern "C"
{
#include <signal.h>
#undef ARGS			// Because there are problems including st.
#define ARGS(args) ()		// 
#include "st.h"
}


extern String vov_stdoutRedirect( String& commandline );

static int vovIsConnected = 0;		// Global to indicate that all is OK. 
static VovTrace* traceP_static;		// The trace we connect with.
static VovTransition* tranP_static;	// The current transition.

int vovApp_alreadyInput( VovPlace& place )
{
    static st_table* inST = 0;
    if ( inST == 0 ) {
	inST = st_init_table( strcmp, st_strhash );
    }
    char* key = utcc_strsav( place.getName() );	// Find possible key.
    if ( st_lookup( inST, key, 0 ) == 1 ) {	// If already in table
	delete key;				// delete key
	return 1;				// and return 1 to say that place is 
						// already an input.
    } else {					// else
	st_insert( inST, key , 0 );		// Add input into table
	return 0;				// Place not already an input.
    }
}

int VOV__input( VovPlaceType type , char* name )
{
    if ( ! vovIsConnected ) return oct_null_id;

    VovTrace& trace = *traceP_static;
    VovTransition& tran = *tranP_static;
    
    VovPlace place( type , name );
    if ( ! vovApp_alreadyInput( place ) ) {
	VOVaddInput( *traceP_static, tran, place );
    }
    return place.id();
}

int VOV__inputFacet( octObject* facet )
{
    return VOV__input( VOV_OCT_FACET, form( "%s:%s:%s", 
					 facet->contents.facet.cell,
					 facet->contents.facet.view,
					 facet->contents.facet.facet ) );
}

int VOV__output( VovPlaceType type , char* name )
{
    if ( ! vovIsConnected ) return oct_null_id;

    VovTrace& trace = *traceP_static;
    VovTransition& tran = *tranP_static;
    VovPlace place( type , name );
    VOVaddOutput( *traceP_static, tran, place );
    return place.id();
}

int VOV__outputFacet( octObject* facet )
{
    return VOV__output( VOV_OCT_FACET, form( "%s:%s:%s", 
					 facet->contents.facet.cell,
					 facet->contents.facet.view,
					 facet->contents.facet.facet ) );
}

int VOV__delete( VovPlaceType type , char* name )
{
    if ( ! vovIsConnected ) return oct_null_id;

    VovTrace& trace = *traceP_static;
    VovPlace place( type , name );			/*  */
    if ( trace.find( place ) == VOV_OK ) {		/*  */
	trace.destroy( place );				/* If it is there, destroy it. */
    }
    return 0;
}


int VOV__affinity( char* affinity )
{
    if ( ! vovIsConnected ) return oct_null_id;

    VovTrace& trace = *traceP_static;
    VovTransition& tran = *tranP_static;

    trace.getById( tran ) ;	/* Get a fresh copy */
    tran.setAffinity( affinity ); /* Change its affinity */
    trace.modify( tran );	/* Modify it. */

    return 0;
}



int VOV__begin( int argc, char* argv[])
{
    vovRegisterMainArgs( argc, argv ); // To be ready to initialize the DDS package.
    VOVcatchAllSignals();

    traceP_static = new VovTrace;
    VovTrace& trace = *traceP_static;
    String host = utcc_getenv( "VOV_HOST_NAME" );
    String project = utcc_getenv( "VOV_PROJECT_NAME" );
    if ( trace.open( project, host , "a", VOV_TRACE_TYPE_CLIENT ) != VOV_OK ) {
	// vovError("Connection to server failed." );
	return oct_null_id;			// Continue anyway.
    }

    String comm = argv[0];
    for (int i = 1; i < argc; i++ ) {
	comm += " ";
	comm += argv[i];
    }

    tranP_static = new VovTransition( comm , 0 );
    VovTransition& tran = *tranP_static;

    tran.setAffinity( utcc_machine() ); // By default, the transition can run on this type of machine.

    VOV_ASSERT( trace.create( tran ) );
    VOV_ASSERT( trace.initiate( tran ) );

    
    VovUser user( utcc_getenv("USER") );
    VOV_ASSERT( trace.findOrCreate( user ) );
    VOV_ASSERT( trace.activeDesigner( user , tran ) );

    vovIsConnected = 1;


    if ( utcc_getenv( "VOV_NO_ECHO", 0 ) != 0 ) {
	String stdoutfile =  vov_stdoutRedirect( comm );
	vovMessage( "Redirecting output to " + stdoutfile );
	VOV__output( VOV_UNIX_FILE, stdoutfile );
	if (freopen( (char*)stdoutfile, "w", stdout) == 0) {
	    utcc_printerrno( "Cannot redirect stdout" );
	}
	if (dup2( fileno(stdout) , fileno(stderr) ) == -1 ) {
	    utcc_printerrno("cannot dup2 stderr ");
	}
    }

    return tran.id();
}


void VOV__measure( int placeId, char* name, char* units, int value )
{
    VovTrace& trace = *traceP_static;
    VovTransition& tran = *tranP_static;

    // A measurement is first of all an output of the transition, just like any other output.
    String placeName = String(name) + "." + units;
    VovPlace measure( VOV_MEASURE, placeName );
    VOVaddOutput( trace, tran, measure ) ;


    // The measurement should also be associated to the right place.
    VovPlace place;
    place.id() = placeId;

    measure.setTimeStamp();		// Update timestamp of measure to now.
    measure.getValue() = value;		// Update value of this measurement.
    VOV_ASSERT( trace.measure( place, measure ) );

    return ;
}


void _VOV__end( int status )
{
    VovTrace& trace = *traceP_static;
    VovTransition& tran = *tranP_static;
    if ( vovIsConnected ) {
	trace.terminate( tran, VOV_OK, status, 0 );
    }
}

void VOV__end( int status )
{
    _VOV__end( status );
    exit( status );
}





// Catch all signals that might force clients to go in debug mode.

SIGNAL_FN genericCatcher( int sig,  int  code, struct sigcontext* scp)
{
    switch ( sig ) {
    case SIGCHLD:		// to parent on child stop or exit.
    case SIGTSTP:		// stop signal from tty ( ctrl-Z ).
	vovMessage( form("Caught signal %d: continuing.", sig ) );
	return;
    case SIGIOT:
	vovMessage( "Signal SIGIOT I must exit." );
	exit( 1 );		// Cannot print output message.
    default:
	vovError( form( "Caught signal %d: exiting.", sig) );
	exit( sig );
    }
} 

void VOVcatchAllSignals()
// Catch all strange/unexpected signals
{
    for ( int s = 1 ; s < 32 ; s++ ) {
	switch ( s ) {
	case SIGINT:
	case SIGCHLD:		// To parent on child stop or exit.
	case SIGTSTP:		// Stop signal from tty ( ctrl-Z ).
	case SIGCONT:		// Continue a stopped process.
	case SIGWINCH:		// Change of a window: ignore.
	    break;
	default:
	    signal( s, genericCatcher );
	}	
    }
}
