

#include "trace.h"
#include "annotation.h"
#include "place.h"
#include "transition.h"
#include "event.h"
#include "voverror.h"
#include "ddsplus.h"

extern String vov_normalizeString( const String& line );
extern String vov_prettyTime( int s );

extern char* unix_read_file( char* path );
extern int   unix_write_file( char* path, char* text );

static void newStringValue( VovProp& prop, String& s, int& changed )
{
    if ( prop.string() != s ) {
	changed = 1;
	prop = s;
    }
}

static void newIntValue( VovProp& prop, int v, int& changed )
{
    if ( prop.integer() != v ) {
	changed = 1;
	prop = v;
    }
}

void VovAnnotation::setupDialog( VovDialog& top_dialog, VovDialogHandles& data )
{
    VovDialogEditText note( top_dialog, "note", this->string(), 5, 60 );
    VovDialogEditText signature( top_dialog, "signature", signature.string(), 1, 30 );
    Date d = date.integer();

    VovDialogEditText date( top_dialog, "date", d.format(), 1, 60 );


    top_dialog.addHelp("You are preparing an annotation to be attached to an object.\n");

    data.add( note );
    data.add( signature );
    data.add( date );
}

VovStatus VovAnnotation::readDialog( VovDialogHandles& data )
{
    data.reset();
    VovDialogEditText noteDlg( data.next() );
    VovDialogEditText signatureDlg( data.next() );
    VovDialogShowText dateDlg( data.next() );

    int changed = 0;
    newStringValue( *this, noteDlg.getTextItem(), changed );
    newStringValue( signature, signatureDlg.getTextItem(), changed );

    return changed ? VOV_CHANGE : VOV_NO_CHANGE;
}


void VovSelectRule::setupDialog(  VovDialog& top_dialog, VovDialogHandles& data )
{
    // vovMessage( "in setupDialog: " + format() );

    VovDialog flagsDlg( top_dialog, "SELECTION FLAGS", DDS_HORIZONTAL, 3 );

    VovDialogList typeDialog( flagsDlg, DDS_FALSE );
    int tm = typemask.integer();
    typeDialog.addItem( "TRANSITION", VOV_TRANSITION_MASK, tm & VOV_TRANSITION_MASK );
    typeDialog.addItem( "PLACE",      VOV_PLACE_MASK, 	   tm & VOV_PLACE_MASK );
    typeDialog.addItem( "NO_INPUTS",  VOV_SOURCE_MASK,     tm & VOV_SOURCE_MASK );
    typeDialog.addItem( "NO_OUTPUTS", VOV_SINK_MASK,       tm & VOV_SINK_MASK );
    typeDialog.endOfList(0,0);
    


    VovDialogList statusDialog(  flagsDlg , DDS_FALSE);
    {
	int sm = statusmask.integer();
	for (int i=VOV_FIRST_STATUS_NODE+1;i<VOV_LAST_STATUS_NODE;i++) {
	    String label = vov_getStatusString( i );
	    statusDialog.addItem( label, (1 << i), sm & ( 1 << i ) );
	}
	statusDialog.endOfList( 0, 0 );
    }
    VovDialogList type2Dialog( flagsDlg, DDS_FALSE );
    {
	int t2m = type2mask.integer();
	for (int i=VOV_FIRST_PLACE_TYPE+1;i<VOV_LAST_PLACE_TYPE;i++) {
	    String label = vov_getPlaceTypeString( i );
	    type2Dialog.addItem( label, (1 << i), t2m & ( 1 << i ) );
	}
	type2Dialog.endOfList( 0, 0 );
    }


    String regEx = rx.string();
    VovDialogEditText rxDialog(top_dialog, "REG EX", regEx, 1, 20 );
    
    data.add( statusDialog );
    data.add( typeDialog );
    data.add( type2Dialog );
    data.add( rxDialog );

}

VovStatus VovSelectRule::readDialog( VovDialogHandles& data )
{
    int changed = 0;
    data.reset();
    VovDialogList statusDialog( data.next() );
    VovDialogList typeDialog( data.next() );
    VovDialogList type2Dialog( data.next() );
    VovDialogEditText rxDialog( data.next() );
    
    int tm  = typeDialog.reduceOR();
    int t2m = type2Dialog.reduceOR();
    int sm  = statusDialog.reduceOR();
    
    newStringValue( rx, rxDialog.getTextItem(), changed );
    newIntValue( typemask, tm, changed );
    newIntValue( type2mask, t2m, changed );
    newIntValue( statusmask, sm, changed );

    return changed ? VOV_CHANGE : VOV_NO_CHANGE;

}


void VovTransition::setupDialog( VovDialog& top, VovDialogHandles& handles, int viewonly = 0 )
{
    if ( viewonly ) {
	VovDialogShowText cwdDlg( top, "cwd", cwd.string(), 1, 60 );
	VovDialogShowText cmdDlg( top, "commandLine", commandLine.string(), 5, 60 );
	VovDialogShowText legalDlg( top, "legal exit status", okStatus.string(), 1, 60 );
	VovDialogShowText affinityDlg( top, "affinity", affinity.string(), 1, 60 );
    } else {
	VovDialogEditText cwdDlg( top, "cwd", cwd.string(), 1, 60 );
	VovDialogEditText cmdDlg( top, "commandLine", commandLine.string(), 5, 60 );
	VovDialogEditText legalDlg( top, "legal exit status", okStatus.string(), 1, 60 );
	VovDialogEditText affinityDlg( top, "affinity", affinity.string(), 1, 60 );
	handles.add( cwdDlg );
	handles.add( cmdDlg );
	handles.add( legalDlg );
	handles.add( affinityDlg );
    }
    String info = form("Exit status %d:  %s@%s: duration %s",
		       exitStatus.integer(),(char*)user.string(),
		       (char*)host.string(),(char*)vov_prettyTime(duration()) );
    VovDialogShowText infoDlg( top, "INFO", info, 1, 60 );
    VovDialogShowText statusDlg( top, "STATUS", vov_getStatusString( getStatus() ), 1, 60 );
}    

VovStatus VovTransition::readDialog( VovDialogHandles& handles )
{

    handles.reset();
    VovDialogEditText cwdDlg ( handles.next() );
    VovDialogEditText cmdDlg ( handles.next() );
    VovDialogEditText legalDlg ( handles.next() );
    VovDialogEditText affinityDlg ( handles.next() );
    int changed = 0;

    newStringValue( commandLine, cmdDlg.getTextItem( ), changed );
    if ( changed ) {
	commandLine = vov_normalizeString( commandLine.string() );
    }
    newStringValue( affinity, affinityDlg.getTextItem( ), changed );
    if ( changed ) {
	affinity = vov_normalizeString( affinity.string() );
    }
    newStringValue( okStatus, legalDlg.getTextItem( ), changed );
    if ( changed ) {
	okStatus = vov_normalizeString( okStatus.string() );
    }
    newStringValue( cwd, cwdDlg.getTextItem( ), changed );
    return changed ? VOV_CHANGE : VOV_NO_CHANGE;
}



void VovPlace::setupDialog( VovDialog& top_dialog, VovDialogHandles& handles, int viewOnly = 0 )
{
    VovStatus rtn = VOV_OK;
    
    VovDialogShowText typeDialog( top_dialog,  "PLACE TYPE", vov_getPlaceTypeString( getPlaceType() ), 1 , 40 );
    VovDialogShowText statusDlg( top_dialog, "STATUS", vov_getStatusString( getStatus() ), 1 , 40 );
    VovDialogShowText timeDlg( top_dialog, "TIMESTAMP", getTimeStamp().format(), 1, 40 );

    switch ( getPlaceType() ) {
    case VOV_UNIX_FILE:
	{
	    VovDialogShowText nameDlg( top_dialog, "NAME", getName(), 1, 60 );
	    if ( viewOnly ) {
		VovDialogShowText dialog( top_dialog, getName() );
	    } else {
		char* text = unix_read_file ( (char*)getName() );
		VovDialogEditText dialog( top_dialog,"", text, 24, 80 );
		delete text;
		handles.add( dialog );
	    }
	}
	break;
    case VOV_OCT_FACET:
	{
	    String msg = "\n\tThis is an OCT facet. Please use VEM to edit it.";
	    msg += "\n\tHere is the VEM command you can use:\n";
	    msg += form("\"%s\"o", (char*)getName() );
	    VovDialogShowText dialog( top_dialog, "FACET", msg, 6, 60 );
	}
	break;
    case VOV_PROGRAM_OPTION:
	{
	    VovDialogEditText dialog( top_dialog, "name", getName(), 1, 60 );
	    VovDialogEditText value ( top_dialog, "value", value.string(), 2, 60 );
	    handles.add( dialog );
	    handles.add( value );
	}
	break;
    case VOV_MEASURE:
	{
	    VovDialogEditText nameDlg( top_dialog, "NAME", getName(), 1, 60 );
	    VovDialogEditText valueDlg( top_dialog, "VALUE", dec( value.integer() ), 1, 60 );
	    handles.add( nameDlg );
	    handles.add( valueDlg );
	}
	break;
    case VOV_BOOLEAN:
    case VOV_EXIT_STATUS:
    case VOV_EXECUTABLE:
	{ 
	    VovDialogShowText nameDlg( top_dialog, "NAME", getName(), 1, 60 );
	    String msg = "\n\n\t" + vov_getPlaceTypeString( getPlaceType() ) + " cannot be edited.\n";
	    VovDialogShowText dialog( top_dialog, "Warning", msg , 6, 60 );
	}
	break;
    default:
	vovFatalError("Invalid place type in vovPlace::setupDialog");
	break;
    }
}

VovStatus VovPlace::readDialog( VovDialogHandles& handles )
{
    int changed = 0;		// Has the place been changed?

    handles.reset();

    switch ( getPlaceType() ) {
    case VOV_UNIX_FILE:
	{
	    VovDialogEditText textDlg( handles.next() );
	    char* oldtext = unix_read_file( (char*)getName() );
	    char* newtext = textDlg.getTextItem();
	    char* op = oldtext;
	    char* np = newtext;
	    while ( *op != '\0' ) {
		if ( *np++ != *op++ ) {
		    changed = 1;
		    break;
		}
	    }
	    if ( changed ) {
		// vovMessage( "Writing file " + getName() );
		unix_write_file( (char*)getName(), newtext );
	    }
	    delete oldtext;
	}
	break;
    case VOV_BOOLEAN:
    case VOV_OCT_FACET:
    case VOV_EXIT_STATUS:
    case VOV_EXECUTABLE:
    case VOV_MEASURE:
	break;
    case VOV_PROGRAM_OPTION:
	{
	    VovDialogEditText nameDlg( handles.next() );
	    VovDialogEditText valueDlg( handles.next() );
	    newStringValue( value, valueDlg.getTextItem(  ), changed );
	    if ( changed ) {
		value = vov_normalizeString( value.string() );
	    }
	    newStringValue( name, nameDlg.getTextItem( ), changed );
	}
	break;
    default:
	vovFatalError("Invalid place type in vovPlace::info");
	break;
    }
    return changed ? VOV_CHANGE : VOV_NO_CHANGE;
}


void VovNodeSet::setupDialog( VovDialog& top_dialog, VovDialogHandles& data )
{
    vovMessage( "nodeset::setupDialog(): not implemented (should not be used either)" );
#ifdef SKIPIT   
    int old_dialog_length = top_dialog.length();
    VovDialog& dialog = top_dialog.getNewDialog( DM_WHICH_ONE );

    const int MAX_ITEMS = 20;
    int countItems = 0;

    
    dialog.setTitle( VovBag::format() );
    VovContentsGenerator genNode( this, VOV_NODE_OCTMASK );
    VovNode node;
    String label;
    while ( octGenerate( &genNode, &node ) == OCT_OK ) {
	node.get();
	if ( node.nodeType() == VOV_PLACE ) {
	    VovPlace place = node;
	    label = place.getName();
	} else {
	    VovTransition tran = node;
	    label = tran.getCommandLine();
	}
	dialog.addItem(label, node.id() );
	if ( countItems++ > MAX_ITEMS ) break;
    }
    dialog.setItem( label );

    if ( countItems == 0 ) {
	vovMessage("The set is empty" );
	return VOV_OK;
    }

    return top_dialog.length() - old_dialog_length ;
#endif
}

VovStatus VovNodeSet::readDialog( VovDialogHandles& handles )
{
    vovMessage( "NodeSet:: Null routine" );
    return  VOV_NO_CHANGE;
}



