#include "trace.h"
#include "page.h"
#include "utcc.h"
#include "node.h"
#include "voverror.h"
#include "placematrix.h"
#include "version.h"

extern void vovUnpackDefaults( VovObject* f, char* m, char* s);
extern octStatus vovUnpackFacetName( VovObject* facet, char* name);

VovPage::VovPage()
    : 
    ( OCT_FACET ),
    nodes( ";;;NODES;;;" ), 
    places(";;;PLACES;;;" ),
    transitions(";;;TRANSITIONS;;;" ),
    placesToCheck(";;;PLACES_TO_CHECK;;;" ),
    placesToEvolve(";;;PLACES_TO_EVOLVE_TO;;;" ),
    transitionsToFire(";;;TRANSITIONS_READY;;;" ),
    transitionsToKill(";;;TRANSITIONS_TO_KILL;;;" ),

    retracings("RETRACINGS"),
    sets("SETS"),
    goals("GOALS"),
    users("USERS"),
    goodBackup( "GOOD_BACKUP", OCT_INTEGER ),
    lastSave( "LASTSAVE", OCT_INTEGER )
{
    matrix = 0;				// No placement matrix yet.
    contents.facet.view = "symbolic";	// 
    contents.facet.facet = "contents";	// 
    contents.facet.mode = "r";		// Default mode.
    lastDoubleSave = 0;			// Date of last save.
}

VovPage::VovPage( VovPage& p )
    : 
    ( p ),
    nodes( p.nodes ),
    places( p.places ),
    transitions( p.transitions ),
    retracings( p.retracings ),
    sets( p.sets ),
    goals( p.goals ),
    users( p.users ),
    placesToCheck( p.placesToCheck ),
    placesToEvolve( p.placesToEvolve ),
    transitionsToFire( p.transitionsToFire ),
    transitionsToKill( p.transitionsToKill ),
    goodBackup( p.goodBackup ),
    lastSave( p.lastSave )
{
    matrix = p.matrix;
    vovMessage( "VovPage(VovPage&) I do not expect this constructor to be ever used!!" );
}


VovPage::~VovPage()
{
}


VovStatus VovPage::initPlacementMatrix()
{
    if ( matrix == 0 ) {
	matrix = new VovPlacementMatrix( 40 , 10 );
	matrix->init( nodes );
    }
    return VOV_OK;
}

VovStatus VovPage::open( const String& name, const String& mode )
{
    contents.facet.cell = utcc_strsav( (char*)name );
    contents.facet.mode = utcc_strsav( (char*)mode );
    contents.facet.version = OCT_CURRENT_VERSION;
    octStatus s = octOpenFacet( this );
    if ( s < OCT_OK ) {
	vovError( octErrorString() );
	vovError( "Maybe the backup trace is OK." );
	return VOV_NOT_OK;
    }
    if ( mode == "a" && s == OCT_NEW_FACET ) {
	vovError("Cannot append to a non existing trace." );
	return VOV_NOT_OK;
    }
	
    VOV_ASSERT( get() );


    if ( mode != "r" ) {
	lastSave = getTimeStamp();
	goodBackup = 0;
	octCreateOrModify( this, &goodBackup );
	initPlacementMatrix();
    }

    lastDoubleSave = 0;			// Date of last save.


    return VOV_OK;
}

VovStatus VovPage::open( VovObject& facet )
{
    ohGetById( this, facet.id() );
    VOV_ASSERT( get() );
    // vovMessage( form( "Opened: %s", ohFormatName( this ) ) );
    if ( String(contents.facet.mode) != "r" ) {
	initPlacementMatrix();
    }
    return VOV_OK;
}

VovStatus VovPage::get()
{
    if ( octGetByName( this, &goals ) == OCT_OK ) {
	// vovMessage( "Getting an old page" );
	OH_ASSERT( octGetByName( this, &users ) );
	if ( octGetByName( this, &retracings ) != OCT_OK ) {
	    OH_ASSERT( octCreate( this, &retracings ) );
	}
	OH_ASSERT( octGetByName( this, &sets ) );
	OH_ASSERT( octGetByName( &sets, &nodes ) );
	OH_ASSERT( octGetByName( &sets, &places ) );
	OH_ASSERT( octGetByName( &sets, &transitions ) );
	OH_ASSERT( octGetByName( &sets, &placesToCheck ) );
	OH_ASSERT( octGetByName( &sets, &placesToEvolve ) );
	OH_ASSERT( octGetByName( &sets, &transitionsToFire ) );
	nodes.get();
	places.get();
	transitions.get();
	placesToCheck.get();
	placesToEvolve.get();
	transitionsToFire.get();
	if (octGetByName( &sets, &transitionsToKill ) == OCT_OK ) { 
	    transitionsToKill.get();
	} else {
	    OH_ASSERT( octCreate( &sets, &transitionsToKill ) );
	    transitionsToKill.create();
	}
    } else {
	vovMessage("Creating a new page." );
	OH_ASSERT( octCreate( this, &goals ) );
	OH_ASSERT( octCreate( this, &users ) );
	OH_ASSERT( octCreate( this, &retracings ) );
	OH_ASSERT( octCreate( this, &sets ) );
	OH_ASSERT( octCreate( &sets, &nodes ) );
	OH_ASSERT( octCreate( &sets, &places ) );
	OH_ASSERT( octCreate( &sets, &transitions ) );
	OH_ASSERT( octCreate( &sets, &placesToCheck ) );
	OH_ASSERT( octCreate( &sets, &placesToEvolve ) );
	OH_ASSERT( octCreate( &sets, &transitionsToFire ) );
	OH_ASSERT( octCreate( &sets, &transitionsToKill ) );
	nodes.create();
	places.create();
	transitions.create();
	placesToCheck.create();
	placesToEvolve.create();
	transitionsToFire.create();
	transitionsToKill.create();
    }
    

    
    // Properties required to view the facet in VEM.
    VovObject prop;
    ohGetOrCreatePropStr( this, &prop, "TECHNOLOGY", "scmos");
    ohGetOrCreatePropStr( this, &prop, "VIEWTYPE", "SYMBOLIC");
    ohGetOrCreatePropStr( this, &prop, "EDITSTYLE", "SYMBOLIC");

    return VOV_OK;
}


VovStatus VovPage::saveOntoDisk()
{

    if ( getTimeStamp() > lastSave.integer() ) {
	VovProp prop( "VOV_VERSION", OCT_STRING);
	prop = String(VOV_VERSION);
	OH_ASSERT( octCreateOrModify( this, &prop )) ;
	if ( octFlushFacet( this ) == VOV_ERROR ) {
	    vovError( form( "Problems saving the trace!!\n\t%s.", octErrorString()) );
	    if ( ! goodBackup.integer() ) { // Even the backup is corrupted. Emergency!!
		VovObject savedFacet( OCT_FACET );
		savedFacet.contents.facet = this->contents.facet;
		char* emergencytrace = "/usr/tmp/vovemergencytrace:panic";
		OH_ASSERT( vovUnpackFacetName( &savedFacet, emergencytrace ));
		if ( octWriteFacet( &savedFacet, this ) == OCT_ERROR ) {
		    vovError( form( "\
Emergency trace saving also failed!!\n\
\tWe are in truble.!!\n\
\t%s.", octErrorString()) );
		} else {
		    vovMessage( form("Emergency trace saved in %s", emergencytrace ));
		}
	    }
	    return VOV_ERROR;	// In any case, we have a serious error.
	}
	lastSave =  getTimeStamp();
    }

    // Is it time to save the backup too ?
    if ( getDate() > ( lastDoubleSave + DATE_HOUR / 2 ) ) {
	vovMessage( "Saving backup trace" );
	VovObject savedFacet( OCT_FACET );
	savedFacet.contents.facet = this->contents.facet;
	OH_ASSERT( vovUnpackFacetName( &savedFacet, "::contents:vov_saved" ) );
	vovMessage( ohFormatName( &savedFacet ) );
	vovMessage( ohFormatName( this ) );
	if ( octWriteFacet( &savedFacet, this ) == OCT_ERROR ) {
	    vovError( form( "Problems saving backup copy of trace!!\n\t%s.", octErrorString()) );
	    goodBackup.modify( 0 );
	    return VOV_ERROR;
	} else {
	    goodBackup.modify( 1 );
	    lastDoubleSave = getDate();
	}
    }
    vovMessage( "Saving Done" );
    return VOV_OK;
}



VovStatus VovPage::close()
{
    vovMessage( String("Closing ") + ohFormatName( this ));
    OH_ASSERT( octFreeFacet( this ) );
    return VOV_OK;
}


int32 VovPage::getTimeStamp()
{
    octFacetInfo info;
    OH_ASSERT( octGetFacetInfo( this, &info ) );
    return info.timeStamp;
}


String VovPage::format()
{
    String s = ( String("PAGE ") + contents.facet.cell );
    s += form( "\n\tPLACES: %5d  TRANSITIONS: %5d  SETS: %5d  RETRACINGS: %3d",
	      places.length(VOV_NODE_OCTMASK), 
	      transitions.length(VOV_NODE_OCTMASK),
	      sets.length(VOV_NODE_SET_OCTMASK),
	      retracings.length(VOV_NODE_SET_OCTMASK) );
    return s;
}

VovStatus VovPage::OK()
{
    vovMessage("VovPage::OK() is not yet implemented");
    return VOV_OK;
}
