#include "slaveset.h"
#include "transition.h"
#include "date.h"
#include "voverror.h"


extern String vov_prettyTime( int );

VovStatus vov_compatibleAffinity( String& slave, String& tran )
{
    if ( tran.length() == 0 ) {			// 
	return VOV_OK;				// No affinity requirement for tran, all slaves are OK.
    }						// 
    String words = tran;			// 
    while ( words.length() ) {			// Get one word at a time from tran.
	String word = words;			// 
	if ( words.contains( RXwhite ) ) {	// If more than one word, get first.
	    word = words.before( RXwhite );	// 
	}					// 
	if ( word != 0 ) {			// 
	    if ( slave.contains( word ) ) {	// 
		return VOV_OK;			// OK, slave is OK.
	    }					// 
	}					// 
	words = words.after( RXwhite );		// Eliminate first word from words.
    }						// 
    return VOV_NOT_OK;				// 
}



SlaveSet::SlaveSet( SlaveSet& s )
{
    numberOfSlaves = s.numberOfSlaves;
    numberOfWorkingSlaves = s.numberOfWorkingSlaves;
    for (int i = 0 ; i < numberOfSlaves ; i++ ) {
	slaves[i] = s.slaves[i];
    }
}

SlaveSet::~SlaveSet( )
{
    delete slaves;
}

int SlaveSet::index( Client* c) 
{
    int found = -1;
    for (int i = 0; i < numberOfSlaves ; i++ ) {
	if ( slaves[i].client == c ) {
	    found = i;
	    break;
	}
    }    
    return found;
}

int SlaveSet::first()
{
    return ( numberOfSlaves ) ? 0 : -1;
}

void SlaveSet::next( int& i ) 
{
    if ( i < numberOfSlaves -1 ) {
	i = i + 1;
    } else {
	i = -1;
    }
}

Client* SlaveSet::operator() (int i )
{
    if ( i >= 0 && i < numberOfSlaves ) {
	return slaves[i].client;
    } else {
	vovError( form("Illegal index %d in Slaveset()", i ) );
	return 0;
    }
}

int SlaveSet::availableSlaves()
{
    if ( numberOfSlaves == 0 ) {
	return 0;
    } else {
	return numberOfSlaves - numberOfWorkingSlaves;
    }
}
void SlaveSet::slaveBegin( Client* c, int duration )
{
    int i = index( c );
    if (i == -1 ) {
	vovError("not a slave ");
	return;
    }
    slaves[i].status = SLAVE_BUSY;
    numberOfWorkingSlaves++;

    slaves[i].estDuration = duration;
    slaves[i].startDate = getDate(); // Transition start date.
}

void SlaveSet::slaveEnd( Client* c)
{
    int i = index( c );
    if (i == -1 ) {
	vovError("not a slave ");
	return;
    }
    if ( slaves[i].status == SLAVE_BUSY ) {
	slaves[i].status = SLAVE_DONE; /* Not working, but not ready either. */
	numberOfWorkingSlaves--;
    } else {
	vovError("Slave did not need to be stopped.");
    }
}

// Adds or updates slave info.
VovStatus SlaveSet::add( Client* c, int power, const String& affinity)
{
    // print( "In add slave");
    int i = index( c );

    if ( i >=  0) {
	slaves[i].power  = power;
	slaves[i].affinity = affinity;
	if ( slaves[i].status == SLAVE_DONE ) {
	     slaves[i].status = SLAVE_IDLE;
	} else if ( slaves[i].status != SLAVE_IDLE ) {
	    vovMessage("Unexpected message from slave on " + slaves[i].client->host );
	}
    } else {
	slaves[ numberOfSlaves ].client = c;
	slaves[ numberOfSlaves ].power  = power;
	slaves[ numberOfSlaves ].affinity  = affinity;
	slaves[ numberOfSlaves ].status = SLAVE_IDLE;
	slaves[ numberOfSlaves ].pid = c->pid;
    	numberOfSlaves++;
    }
    c->slaveset = this;
    return VOV_OK;
}

// Linear search of the disappearing slave.
//
VovStatus SlaveSet::del( Client* c )
{
    // vovMessage( "In delete slave\n:" + format() );
    int i = index( c ) ;
    if ( i == -1) {
	vovMessage( "Not a valid slave" );
	return VOV_ERROR;    }
    if (slaves[i].status == SLAVE_BUSY ) {
	numberOfWorkingSlaves--;
    }
    int l = --numberOfSlaves;	/* last slave */
    slaves[i].client  = slaves[l].client;
    slaves[i].power   = slaves[l].power;
    slaves[i].status = slaves[l].status;
    slaves[i].startDate = slaves[l].startDate;
    slaves[i].estDuration = slaves[l].estDuration;
    slaves[i].affinity= slaves[l].affinity;
    return VOV_OK;	/* deletion successful */
}

// Get the best available slave, subject to the affinity constraint.
VovStatus SlaveSet::bestSlave( Client** c, const String& affinity )
{
    if ( numberOfSlaves  == 0) {    
	vovMessage("No slaves connected");
	return VOV_ERROR;
    }

    // Make sure that all the slaves are  still alive.
    int allSlavesOk = 1;
    do {
	allSlavesOk = 1;
	for (int i = 0; i < numberOfSlaves ; i++ ) {
	    if (slaves[i].client->check() != BUF_OK ) {
		vovMessage("Slave on " + slaves[i].client->host + " died prematurely." );
		del( slaves[i].client );
		allSlavesOk = 0;
		break;
	    }
	}
    } while ( ! allSlavesOk );



    if ( numberOfSlaves - numberOfWorkingSlaves == 0) {  // Not enough slaves.
	return VOV_NOT_OK;				 // Must wait a while.
    }

    int bestindex = 0;
    int bestpower = 0;
    for (int i = 0; i < numberOfSlaves ; i++ ) {
	if ( slaves[i].status == SLAVE_IDLE &&
	    vov_compatibleAffinity( slaves[i].affinity, affinity ) == VOV_OK && 
	    slaves[i].power > bestpower) {
            bestpower = slaves[i].power;
	    bestindex = i;
	}
    }

    if ( bestpower == 0 ) {
	int count = 0;		// Count compatible slaves.
	for (int i = 0; i < numberOfSlaves ; i++ ) {
	    if ( vov_compatibleAffinity( slaves[i].affinity, affinity ) == VOV_OK ) count++;
	}
	if ( count ) {
	    // vovMessage( "All compatible slaves are now busy" );
	    return VOV_NOT_OK;
	} else {
	    // vovError( "No compatible slaves" );
	    return VOV_ERROR;	
	}
    } else {
	*c = slaves[bestindex].client;
	return VOV_OK;
    }
}

// Put in string text some information on what is going on with the slaves
String SlaveSet::format(  )
{
    String text ;
    if ( numberOfSlaves == 0 ) {
	text = "NO SLAVES CONNECTED TO THE SERVER.\n";
    } else {
	for (int i = 0; i < numberOfSlaves ; i++ ) {
	    String slave_status = form("    IDLE (%s)", (char*)slaves[i].affinity );
	    String hostname = slaves[i].client->host;
	    String numberString;
	    int w = slaves[i].status;
	    if ( w == SLAVE_IDLE ) {
		if ( slaves[i].power == 0 ) {
		    slave_status = "    INACTIVE BECAUSE HOST IS OVERLOADED";
		} else {
		    numberString = form( "%d", slaves[i].power );
		}
	    } else if ( w == SLAVE_DONE ) {
		slave_status = "    JUST FINISHED A JOB";
	    } else {
		VovTransition tr = slaves[i].client->theTransition;
		
		slave_status = tr.getCommandLine();
		if ( slave_status.length() > 45 ) {
		    slave_status = slave_status.at(0,45) + " ...";
		}
		if ( slave_status.contains("\n") ) {
		    slave_status = slave_status.before( "\n" );
		}
		int curDuration = getDate() - slaves[i].startDate;
		numberString = vov_prettyTime( curDuration );


		int percent = -1;
		if ( slaves[i].estDuration > 0 ) {
		    percent = (curDuration * 100) / slaves[i].estDuration;
		    numberString += form("(%3d%%)",percent);
		} else {
		    numberString += ":    ";
		}
		
	    }

	    text += form( "%10s:%10s: %s\n", (char*)hostname, (char*)numberString, (char*)slave_status);
	}
    }
    return text;
}
