/* -*- C++ -*- */
/*
 * trace.h
 *
 * This file describes the data structures to represent the design trace
 */
#ifndef TRACE_H
#define TRACE_H
#pragma once

#include <String.h>
#include "base.h"
#include "date.h"
#include "vovinternal.h"
#include "event.h"
#include "page.h"
#include "VOVprotocol.h"

extern class Bzocket;
extern class Client;
extern class VovTransition;
extern class VovPlace;
extern class VovUser;
extern class VovPlacementMatrix;
extern class VovSelectRule;
extern class VovPage;
extern class VovRetrace;
extern class VovPlaceList;


#define VOV_TRACE_TYPE_DISABLED 0x00
#define VOV_TRACE_TYPE_LOCAL    0x01
#define VOV_TRACE_TYPE_SERVER   0x02
#define VOV_TRACE_TYPE_CLIENT   0x04

#define VOV_SERVER_ENTRY 1
#define VOV_SERVER_EXIT  0
/*
 * The trace is partitioned into pages.
 * A page is a set of VovNodes.
 * Pages are stored independently.
 * Each page has a name. Each page is stored independently in the file system.
 */
#define DEFAULT_PAGES_IN_TRACE 3

class VovTrace {
    String      projectName;	// This is the name of the project.
    String      hostName;	// Host on which the server is running.
    int		readonly;	// Read only flag.
    
    int	type;			// Server, client, local mask.
    Bzocket* zocketPointer;	// Used by traces of type VOV_TRACE_TYPE_CLIENT


    VovEventQueue workQ;	// 
    VovEventQueue unreportedQ;	// 

    /* Organize the trace in pages */
    int		numberOfPages;
    VovPage*	pageArray;	// Array of pages.

    int		retraceFastMode;

    /* Private functions */
    VovPage& workPage();
    int getType() { return type; }
    int isClient();
    int serverEntryCount;
    int isServer( int entry_flag );
    int isLocal();

    VovStatus initGenerator(  int type, VovObject& obj, octObjectMask mask, VovGenerator& gen );

 public:
    VovTrace();
    VovTrace( String& project, String& host, String& mode, int type );
    VovTrace( VovTrace& );
    ~VovTrace( );
    
    int disabled() { return ( type == VOV_TRACE_TYPE_DISABLED) ; }
    VovStatus init();
    VovStatus open( VovObject& facet );
    VovStatus open(String& project, String& host, String& mode, int type);
    VovStatus close();

    VovStatus startProject();
    String getProjectName(){ return projectName; }
    String signature(){ return projectName + "@" + hostName; }
    String getInfo( int mask, int detailed );
    String userFormat( int extended = 0 );		// Information about users' activity.
    String format();					// 

    Date   getProjectStartDate();
    int    accumulateStatistics( String& name, int );
    double accumulateStatistics( String& name, double );
    String formatStatistics();

    Bzocket& getBzocket();
    inline void     setBzocket( Bzocket& z ) { serverEntryCount = 0; zocketPointer = &z; }
    VovStatus serverDispatch( Client& client, VovCode code,  octId& id );

    inline VovEventQueue& getQueue() { return workQ; }
    inline VovEventQueue& getUnreportedQueue() { return unreportedQ; }

    VovStatus copyTrace( VovTrace& , VovNodeSet&, VovNodeSet& );
    VovStatus create( VovSuperObject& );
    VovStatus getById( VovSuperObject& );
    int       list( int objtype, String& rx, String& list );
    VovStatus find( VovSuperObject& );
    VovStatus findOrCreate( VovSuperObject& );
    VovStatus modify( VovSuperObject& );
    VovStatus edit( VovSuperObject&, int viewonly = 0 );

    VovStatus lock( VovObject& obj, String& desc, int force, octId& lockId, String& msg );
    VovStatus unlock( VovObject& obj, octId lockId  );

    VovStatus destroy( VovSuperObject& );
    VovStatus freeResources( VovPlace& );

    VovStatus dispatchTransition( Client&, VovTransition& );
    VovStatus initiate( VovTransition& );
    VovStatus terminate( VovTransition&, VovStatus , int exit_status, String& );
    
    VovStatus setOperation( VovNodeSet& set, VovNode& obj,  VovSetOperation op );
    VovStatus setOperation( VovNodeSet& acc, VovNodeSet& set, VovSetOperation op );

    VovStatus initGenContents(  VovObject& obj, octObjectMask mask, VovGenerator& gen );
    VovStatus initGenContainers( VovObject& obj, octObjectMask mask, VovGenerator& gen );
    VovStatus generate( VovGenerator& gen, VovObject& obj );
    VovStatus freeGenerator( VovGenerator& gen );
    VovStatus subselect( VovNodeSet& set, VovSelectRule& rule );

    VovStatus declareInput( VovTransition&, VovPlace&, int );
    VovStatus declareOutput( VovTransition& ,VovPlace&, int, int, String& errmsg );
    VovStatus declareIOs( VovTransition&, VovPlaceList& );
    VovStatus disconnect( VovNode& n1, VovNode& n2 );
    
    VovStatus addAnnotation( VovSuperObject&, VovAnnotation& );
    int	      numberOfAnnotations( VovSuperObject& );
    int       getAnnotations( VovSuperObject&, VovAnnotation*& array, int length);

    VovStatus history( VovNode&, int flags, int direction, int level, String& );
    VovStatus changeStatus( VovNode&, VovNodeStatus s, int flags );
    
    VovStatus mvlib(VovNodeSet&,String& oldPath,String& newPath,String& msg ); // 

    VovStatus measure( VovPlace& place, VovPlace& measure );			// 

    VovStatus checkPlaces( int );						// 
    VovStatus checkPlace(VovPlace&,int* exist_par,Date* timeStamp_par);		// 
    VovStatus evolve( VovRetrace&, int fastMode, String& msg );			// 
    VovStatus evolve( VovRetrace&, VovNodeSet&, int fastMode, String& msg );	// 
    VovStatus evolve( VovRetrace&, VovNode&, int direction, int fastMode, String& msg ); // 
    VovStatus stopRetracing( String& msg );					//
    VovStatus stopRetracing( VovTransition& tran, String& msg );		//
    VovStatus getTransitionsToFire( VovPlace& , VovBag& );			// 
    VovStatus getAllReadyTransitions( VovPlace&, VovBag&, String& msg );	// 
    VovStatus getAllReadyTransitions( VovBag&, int );				// 
    VovStatus prepareExecutionScript( String& script );				// 
    VovStatus prepareScript( VovNodeSet& set, int flags, String& script );	// 
    VovStatus getIntermediateNodes( VovNodeSet&,VovNodeSet&,VovSelectRule&,VovNodeSet&);
    VovStatus stopAllTransitions( );
    VovStatus killSlave( int index = -1 );
    VovStatus prune();					// 
    void      doPlacement();				// Place the trace nicely.

    VovStatus activeDesigner( VovUser&, VovTransition& );// 
    int       testConnectivity( VovNode& );		 // 
    //     VovStatus level( VovNode& );			// No longer used.
    VovStatus readFromDisk(  );				// 
    VovStatus saveOntoDisk( int asciiDump = 0, String& asciiFile = 0 );	// 
							// 
    VovStatus OK( String& msg );			// Check consistency of trace.
							// 
    void journalLog( const String& msg );		// Log something in the journal.
};


#endif	/* TRACE_H */
