#include "voverror.h"
#include "trace.h"
#include "place.h"
#include "transition.h"
#include "annotation.h"
#include "VOVprotocol.h"
#include "travfunc.h"
#include "retrace.h"
#include "clients.h"
#include "comm.h"

extern "C"
{
#include <signal.h>
}
extern String vov_prettyTime( int seconds );


VovStatus VovTrace::stopRetracing( VovTransition& tran, String& msg)
{
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    
    if ( isClient() ) {
	zoc.pack( VOV_EvolveStop );
	zoc.pack( tran.id() );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    tran.id() = zoc.unpackInt();
	    if ( tran.id() ) {
		if ( tran.get() != VOV_OK ) {
		    msg = "bogus id in stop retracing.";
		    rtn = VOV_ERROR;
		}
	    }
	}
	if ( rtn == VOV_OK ) {
	    VovPage& page = workPage();
	    if ( tran.id() ) {				// If id not null, eliminate transition
		page.transitionsToFire.del( tran );	// 
		VovContentsGenerator gen( &page.retracings, VOV_NODE_SET_OCTMASK );
		VovRetrace retrace;
		while ( octGenerate( &gen, &retrace ) == OCT_OK ) {
		    VOV_ASSERT( retrace.get() );
		    if ( retrace.terminate( tran ) == VOV_DELETE_NODE ) {
			getQueue().enq( VOV_RETRACE_MASK, VE_END, retrace.format() + "\n", retrace.id() , 0 );
			retrace.del();
		    }
		}		

		VovContentsGenerator genOut( &tran, VOV_PLACE_OCTMASK );
		VovPlace p = 0;
		while ( octGenerate( &genOut, &p ) == OCT_OK ) {
		    VOV_ASSERT( p.get() );
		    page.placesToEvolve.del( p );
		}
	    } else {					// completely stop retracing.
		int count = page.transitionsToFire.length( VOV_TRANSITION_OCTMASK );
		page.placesToEvolve.clear( VOV_PLACE_OCTMASK );
		page.transitionsToFire.clear( VOV_TRANSITION_OCTMASK );
		VovContentsGenerator gen( &page.retracings, VOV_NODE_SET_OCTMASK );
		VovRetrace retrace;
		while ( octGenerate( &gen, &retrace ) == OCT_OK ) {
		    VOV_ASSERT( retrace.get() );
		    if ( retrace.stop( msg ) == VOV_OK ) {
			getQueue().enq( VOV_RETRACE_MASK, VE_END, msg + "\n", retrace.id() , 0 );
		    }
		    retrace.del();
		}
	    }
	}
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    zoc.pack( msg );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc ) ;
	rtn = zoc.unpackInt();
	msg = zoc.unpackString();
    }
    return rtn;
}

VovStatus VovTrace::stopRetracing( String& msg )
{
    VovTransition t;
    t.id() = 0;				// Null id, to stop all retracing.
    return stopRetracing( t, msg );	// 
}



int retraceableNode( VovNode& node, int direction )
{
    VovNodeStatus s = node.getStatus();
    if ( direction==VOV_BACKWARD && (s==VOV_PLAN_NODE||s==VOV_NONEXISTING_NODE) ) return 1;
    if ( direction==VOV_FORWARD && s==VOV_TRACE_NODE) return 1;
    return 0;
}



// VovTrace::evolveToNode()
// Evolve to or from a single node.
//
VovStatus VovTrace::evolve( VovRetrace& retrace, VovNode& node, int direction, int speed, String& msg )
{
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    
    if ( isLocal() ) {
	vovError("Cannot evolve a trace which is local");
	return VOV_ERROR;
    }
    
    if ( isClient() ) {
	zoc.pack( VOV_EvolveNode );
	zoc.pack( retrace.id() );
	zoc.pack( node.id() );
	zoc.pack( speed );
	zoc.pack( direction );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    retrace.id() = zoc.unpackInt();
	    node.id() = zoc.unpackInt();
	    speed = zoc.unpackInt();
	    direction = zoc.unpackInt();
	    if ( node.get() != VOV_OK || retrace.get() != VOV_OK ) {
		rtn = VOV_ERROR;
		vovError("corrupt id in evolveToNodelace");
	    }
	}
	
	// vovMessage( "In new evolve::node " );
	if ( rtn == VOV_OK ) {
	    // First check all places up from this node.
	    VOV_ASSERT( node.bfs(VOV_BACKWARD,VT_checkPlace,0,0,0,this) );
	    VOV_ASSERT( node.get() );	// Must get again, because node might have changed !!
	    //
	    // Update page.placesToEvolve
	    //
	    retraceFastMode = speed;
	    VovPage& page = workPage();

	    if ( retraceableNode(node,direction) ) {
		VOV_ASSERT( node.bfs( direction,VT_collectPlanTransition,0,0,0,(void*)&retrace) );
		int eventtype = VE_START;
		VovStatus s = retrace.start( msg );
		msg = retrace.format() + "\n" + msg + "\n" ;
		getQueue().enq( VOV_RETRACE_MASK, eventtype, msg, retrace.id() , 0 );
		if ( s == VOV_OK ) {
		    page.transitionsToFire |= retrace;
		    if ( node.nodeType() == VOV_PLACE ) {
			page.placesToEvolve.add( node );
		    }
		} else {
		    eventtype = VE_ERROR;
		    rtn = VOV_ERROR;
		    retrace.del();
		}
	    } else {
		msg = "No retracing, because\n" +  node.format() + "\nis " + node.getStatusString() + ".";
		rtn = VOV_ERROR;
	    }
	}
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    retrace.pack( zoc );
	    zoc.pack( msg );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc ) ;
	rtn = zoc.unpackInt();
	retrace.unpack( zoc );
	msg = zoc.unpackString();
    }
    return rtn;
}

// Evolve to or from a set.
//
VovStatus VovTrace::evolve( VovRetrace& retrace, VovNodeSet& set, int speed, String& msg )
{
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    if ( isLocal() ) {
	vovError("Cannot evolve a trace which is local");
	return VOV_ERROR;
    }
    
    if ( isClient() ) {
	zoc.pack( VOV_EvolveSet );
	zoc.pack( retrace.id() );
	zoc.pack( set.id() );
	zoc.pack( speed );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    retrace.id() = zoc.unpackInt();
	    set.id() = zoc.unpackInt();
	    speed = zoc.unpackInt();
	    if ( set.get() != VOV_OK || retrace.get() != VOV_OK ) {
		rtn = VOV_ERROR;
		vovError("corrupt id in evolveToPlace");
	    }
	}
    
	retraceFastMode = speed;
	// Update page.placesToEvolve

	int count = 0;		// Count the new places that we evolve to.
	VovPage& page = workPage();
	VovContentsGenerator genNode( &set, VOV_NODE_OCTMASK );
	VovNode node;
	while ( octGenerate( &genNode, &node ) == OCT_OK ) {
	    VOV_ASSERT( node.get() );
	    if ( node.getStatus() != VOV_PLAN_NODE ) continue;

	    if ( retraceableNode(node,VOV_BACKWARD ) ) {
		VOV_ASSERT( node.bfs( VOV_BACKWARD,VT_collectPlanTransition,0,0,0,(void*)&retrace) );
		page.placesToEvolve.add( node );

	    }
	}

	if (  retrace.length( VOV_TRANSITION_OCTMASK ) == 0 ) {
	    msg = "Nothing has to be retraced.";
	    rtn = VOV_ERROR; // Nothing to evolve to.
	} else {
	    page.transitionsToFire |= retrace;
	    if ( retrace.start( msg ) != VOV_OK ) {
		retrace.del();
		msg = retrace.format() + msg;;
	    }

	    getQueue().enq( VOV_RETRACE_MASK, VE_START, msg + "\n", retrace.id() , 0 );
	}
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    retrace.pack( zoc );
	    zoc.pack( msg );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
	retrace.unpack( zoc );
	msg = zoc.unpackString();
    }
    return rtn;
}

VovStatus VovTrace::evolve(  VovRetrace& retrace, int speed, String& msg)
{
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    if ( isLocal() ) {
	vovError("Cannot evolve a trace which is local");
	return VOV_ERROR;
    }
    
    if ( isClient() ) {
	zoc.pack( VOV_EvolveAll );
	zoc.pack( retrace.id() );
	zoc.pack( speed );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    retrace.id() = zoc.unpackInt();
	    speed = zoc.unpackInt();
	    if ( retrace.get() != VOV_OK ) {
		rtn = VOV_ERROR;
		vovError("corrupt id in evolveToPlace");
	    }
	}
    
	checkPlaces( 0 ); 			// Check all places.

	retraceFastMode = speed;
	// Update page.placesToEvolve

	int countplace = 0;		// Count the new places that we evolve to.
	int counttrans = 0;		// Count transitions to be retraced.
	int totalduration = 0;		// Total estimated duration of transitions.
	int unknownduration = 0;	// Number of transitions whose duration is unknown.
	VovPage& page = workPage();
	VovContentsGenerator genPlace( &page.places, VOV_PLACE_OCTMASK );
	VovNode node;		// A node is faster to get than a place.
	while ( octGenerate( &genPlace, &node ) == OCT_OK ) {
	    node.get();
	    if ( node.getStatus() == VOV_PLAN_NODE && node.hasInputs() == VOV_OK ) {
		page.placesToEvolve.add(node );
		countplace++;
	    }
	}
	VovContentsGenerator genTran( &page.transitions, VOV_TRANSITION_OCTMASK );
	while ( octGenerate( &genTran, &node ) == OCT_OK ) {
	    VOV_ASSERT( node.get() );
	    if ( node.getStatus() == VOV_PLAN_NODE ) {
		retrace.add( node );
	    }
	}
	if ( retrace.start( msg ) != VOV_OK ) {
	    msg = retrace.format() + "\n" + msg + "\n";
	    getQueue().enq( VOV_RETRACE_MASK, VE_END, msg, 0 , 0 );
	    retrace.del();
	    rtn = VOV_ERROR;	// Nothing to evolve to.
	} else {
	    page.transitionsToFire |= retrace;
	    msg = retrace.format() + "\n" + msg + "\n";
	    getQueue().enq( VOV_RETRACE_MASK, VE_START, msg, retrace.id() , 0 );
	}
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    retrace.pack( zoc );
	    zoc.pack( msg );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
	retrace.unpack( zoc );
	msg = zoc.unpackString();
    }
    return rtn;
}

VovStatus VovTrace::prepareExecutionScript( String& script )
{
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    String errmsg;

    if ( isClient() ) {
	zoc.pack( VOV_GetExecutionScript );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	}	
	VovPage& page = workPage();

	script = "## PLACES TO BE RETRACED:\n";
	prepareScript( page.placesToEvolve, VOV_PLACE_MASK, script );
	script += "## TRANSITIONS SCHEDULED TO BE RETRACED:\n";
	prepareScript( page.transitionsToFire, VOV_TRANSITION_MASK|VOV_DETAIL_MASK, script );

	if ( isServer(VOV_SERVER_EXIT) ) {
	    VS_sendText( zoc, script );
	    VS_sendText( zoc, 0 );
	    zoc.pack( rtn );
	    zoc.pack( errmsg );
	    zoc.flush();
	}
#endif
    }
    if (isClient() ) {
	VC_get_reply( zoc );
	VC_get_text( zoc, script );
	rtn = zoc.unpackInt();
	errmsg = zoc.unpackString();
    }
    return rtn;
}


class VovScriptData {
public:
    String script;
    int flags;
    int totalDuration;
    int unknownDuration;

    VovScriptData( int f ) { flags = f; totalDuration = 0; unknownDuration = 0; }
    void summary() {
	if ( flags & VOV_TRANSITION_MASK &&   totalDuration > 0 ) {
	    script += form("# Total estimated duration: %s.\n", (char*)vov_prettyTime( totalDuration ) );
	}
    }
};

VovStatus VT_prepareScript( VovNode& node , void* data )
{
#ifndef VOVCLIENT
    VovScriptData& D = *(VovScriptData*)data;
    int nodeDone = 0;
    if ( node.nodeType() == VOV_TRANSITION ) {
	if ( D.flags & VOV_TRANSITION_MASK ) {
	    VovTransition t( node );
	    VovTime duration = t.duration();
	    if ( duration > 0 ) {
		D.totalDuration += duration;
	    } else {
		D.unknownDuration++;
	    }
	    D.script += t.getCommandLine();

	    if ( D.flags & VOV_DETAIL_MASK ) {
		D.script += form( "\n# %-10s", (char*)t.getStatusString() );
		if ( duration > 0 ) {
		    D.script += form( "Estimated duration: %s.", (char*)vov_prettyTime( duration ) );
		}
	    }
	    D.script += "\n";
	    nodeDone = 1;
	}
    } else {						// Node is a place.
	if ( D.flags & VOV_PLACE_MASK ) {			// Should we do places ?
	    VovPlace place( node );			// 
	    D.script += "## " + place.getName() + "\n";	// 
	    nodeDone = 1;
	}
    }
    if ( nodeDone && D.flags & VOV_ANNOTATION_MASK ) { 		// Notes as well ?
	int n = node.countNotes();			// 
	if ( n ) {					// If there actually are some notes.
	    static VovAnnotation* staticNotes = 0;	// Avoid bug in ds3100 g++. SHOULD GET FIXED!!
	    if ( staticNotes == 0 ) {			// 
		staticNotes = new VovAnnotation[VOV_MAX_ANNOTATIONS];	// 
	    }						// 
	    if ( n > VOV_MAX_ANNOTATIONS ) n = VOV_MAX_ANNOTATIONS;			// HACK!!
	    n = node.getNotes( staticNotes, n );	// 
	    for ( int i = 0 ; i < n ; i++ ) {		// 
		String s = "#\t";			// 
		s += staticNotes[i].format();		// 
		s.gsub( "\n", "\n#\t" );		// Make sure it looks like a shell comment.
		D.script += s + "\n";			// New line after note.
	    }
	}
    }
#endif
    return VOV_OK;
}



VovStatus VovTrace::prepareScript( VovNodeSet& set, int flags, String& script )
{
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    String errmsg;
    
    if ( isClient() ) {
	zoc.pack( VOV_GetScript );
	set.pack( zoc );
	zoc.pack( flags );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    set.unpack( zoc );
	    flags = zoc.unpackInt();
	    set.get();
	}
	
	VovPage& page = workPage();
	
	VovScriptData data( flags );
	VOV_ASSERT( set.cover( VT_prepareScript, &data ) );
        data.summary();
	script += data.script;

	if ( isServer(VOV_SERVER_EXIT) ) {
	    VS_sendText( zoc, script );
	    VS_sendText( zoc, 0 );
	    zoc.pack( rtn );
	    zoc.pack( errmsg );
	    zoc.flush();
	}
#endif
    }
    if (isClient() ) {
	VC_get_reply( zoc );
	VC_get_text( zoc, script );
	rtn = zoc.unpackInt();
	errmsg = zoc.unpackString();
    }
    return rtn;
}


//
//
// Collect all the transitions that
// contribute towards the evolution of the places 
// in 'placesToEvolve' and that
// are immediately ready to fire.
//
VovStatus VovTrace::getAllReadyTransitions( VovBag& bag , int maxTransitions )
{
#ifndef VOVCLIENT
    static int activeRetracing = 0;
    
    VovPage& page = workPage();
    OH_ASSERT( octGetOrCreate( &page, &bag ));
    bag.clear( VOV_TRANSITION_OCTMASK );
    
    VovContentsGenerator genTrans( &page.transitionsToFire, VOV_TRANSITION_OCTMASK );
    VovTransition trans = 0;
    int countReadyTransition = 0;
    int countTransitionsInBag = 0;
    String msg;
    while ( octGenerate( &genTrans, &trans ) == OCT_OK ) {
	trans.get();
	if ( trans.isReadyToFire( msg ) == VOV_OK  ) {
	    // vovMessage( "Adding: " + trans.getCommandLine() );
	    bag.add( trans );
	    countReadyTransition++;
	} else if ( trans.getStatus() == VOV_TRACE_NODE ) {
	    // vovMessage( "eliminating: " + trans.getCommandLine() );
	    page.transitionsToFire.del( trans );
	    countReadyTransition++;
	} else if ( trans.getStatus() == VOV_RETRACING_NODE ) {
	    countReadyTransition++;
	} else {
	    // vovMessage( "not ready yet: " + trans.getCommandLine() ) ;
	}
	countTransitionsInBag++;			// Count transitions in bag.
    }

    static int last_countTransitionsInBag = 0;	// latest # of transitions in the bag: 0 or > 0 is what matters.

    if (bag.length( VOV_TRANSITION_OCTMASK ) > 0 ) {
	last_countTransitionsInBag = countTransitionsInBag;
	return VOV_OK;
    }

    if ( countReadyTransition == 0 && countTransitionsInBag > 0 ) {
	// Retracing cannot be completed.
	// It is important to see which transitions did not get retraced.
	// It is also better to not generate too many events, which might confuse the user.
	VovContentsGenerator genTrans( &page.transitionsToFire, VOV_TRANSITION_OCTMASK );
	VovTransition trans;
	while ( octGenerate( &genTrans, &trans ) == OCT_OK ) {
	    VOV_ASSERT( trans.get() );
	    String msg = trans.getCommandLine() + "\ncannot be retraced.\n";
	    trans.isReadyToFire( msg );
	    getQueue().enq( VOV_TRANSITION_MASK, VE_ERROR, msg, trans.id(), 0 );
	    stopRetracing( trans, msg ); // msg is discarded
	}
	page.placesToEvolve.clear( VOV_PLACE_OCTMASK );
	page.transitionsToFire.clear( VOV_TRANSITION_OCTMASK );
	countTransitionsInBag = 0;
    } else if ( (last_countTransitionsInBag > 0  || countReadyTransition > 0)  && countTransitionsInBag == 0 ) {
	// Retracing has now ended.
	// Yes, this can happen.
	page.placesToEvolve.clear( VOV_PLACE_OCTMASK );
	page.transitionsToFire.clear( VOV_TRANSITION_OCTMASK );
	// getQueue().enq( VOV_RETRACE_MASK, VE_END, "", 0 , 0 );
    }

    last_countTransitionsInBag = countTransitionsInBag;
#endif
    return  VOV_NOT_OK;
}

VovStatus VovTrace::getAllReadyTransitions( VovPlace& place, VovBag& readySet , String& msg )
{
    VovStatus rtn = VOV_OK;
#ifndef VOVCLIENT
    switch ( place.getStatus() ) {
    case VOV_TRACE_NODE:
	msg +=  place.getCommonName() + " is already OK." ;
	return VOV_NOT_OK;
    case VOV_TRACING_NODE:
    case VOV_RETRACING_NODE:
	msg += place.getCommonName() + " is being generated now.";
	return VOV_NOT_OK;
    case VOV_PLAN_NODE:
	{
	    VovPage& page = workPage();
	    VovBag bag("EVOLVE_TO_ONE_PLACE");
	    OH_ASSERT( octGetOrCreate( &page, &bag ) );
	    bag.clear( VOV_TRANSITION_OCTMASK );
	    if (place.bfs(VOV_BACKWARD,VT_usefulTransition,0,0,0,(void*)&bag)!=VOV_OK) {
		vovFatalError("in getAllreadyTransitions" );
	    } 
	    int l = bag.length( VOV_TRANSITION_OCTMASK );
	    if ( l == 0 ) {
		getQueue().enq( VOV_PLACE_MASK, VE_ERROR, place.getName() + "\ncannot be retraced", place.id(), 0 );
		msg += "No active transitions in the path to " + place.format();
		return VOV_ERROR;
	    } else {
		readySet |= bag;
		return VOV_OK;
	    }
	}
	break;
    default:
	vovError( "Unexpected type in getTransitionstofire:" );
	rtn =  VOV_ERROR;
	break;
    }
#endif
    return rtn;
}



#ifndef VOVCLIENT
//
// Collect in set all transitions which are 'PLAN'
// 
VovStatus VT_collectPlanTransition( VovNode& node, void* data )
{
    VovBag set = *(VovBag*)data;
    if ( node.nodeType() == VOV_PLACE) {
	return VOV_OK;
    }

    VovTransition trans = node.id();
    if ( trans.getStatus( ) == VOV_PLAN_NODE ) {
	set.add( trans );
    }
    return VOV_OK;
}

VovStatus VT_checkPlace( VovNode& node, void* data )
{
    VovTrace& trace = *(VovTrace*)data;
    if ( node.nodeType() == VOV_PLACE) {
	VovPlace place( node );
	int exists;
	Date timestamp;
	trace.checkPlace( place, &exists, &timestamp );
	return VOV_OK;
    }
    return VOV_OK;
}

VovStatus VT_readyTransition( VovNode& node, void* data )
{
    VovBag set = *(VovBag*)data;
    if ( node.nodeType() == VOV_PLACE) {
	return VOV_OK;
    }

    VovTransition trans = node.id();
    String msg;
    if ( trans.isReadyToFire( msg ) == VOV_OK ) {
	set.add( trans );
    }
    return VOV_OK;
}

VovStatus VT_usefulTransition( VovNode& node, void* data )
{
    VovBag set = *(VovBag*)data;
    if ( node.nodeType() == VOV_PLACE) {
	return VOV_OK;
    }

    VovTransition trans = node.id();
    String msg;
    if ( trans.isReadyToFire(msg) == VOV_OK || trans.getStatus() == VOV_RETRACING_NODE ) {
	set.add( trans );
    }
    return VOV_OK;
}
#endif


VovStatus VovTrace::stopAllTransitions( )
{
    Bzocket& zoc = getBzocket();
    
    if ( isClient() ) {
	zoc.pack( VOV_StopAllTransitions );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    
	}
	
	VovPage& page = workPage();
	
	VovContentsGenerator genNode( &page.nodes, VOV_NODE_OCTMASK );
	VovNode node;
	while ( octGenerate( &genNode, &node ) == OCT_OK ) {	
	    VOV_ASSERT( node.get() );

	    VovNodeStatus s = node.getStatus();
	    if ( s == VOV_TRACING_NODE || s == VOV_RETRACING_NODE ) {
		if ( node.nodeType() == VOV_TRANSITION ) {
		    VovTransition tran( node );
		    if ( tran.id() != node.id() ) {
			vovError( node.format() + "\n\tproblems with this node. deleting it." );
			node.del();
			continue;
		    }
		    int signalToSend = (s==VOV_TRACING_NODE)?SIGKILL:SIGUSR1;

		    if ( tran.kill( signalToSend ) != VOV_OK ) {
			vovMessage( "Kill failed" );
		    }
		    if ( terminate( tran, VOV_ERROR, 0, 0 ) != VOV_OK ) {
			VOV_ASSERT( destroy( tran ) );
		    }
		} else {
		    VovPlace place( node );
		    if ( changeStatus( place, VOV_PLAN_NODE, VOV_CHANGE_STATUS_SELF_REC ) != VOV_OK ) {
			VOV_ASSERT( destroy( place ) );
		    }
		}
	    } 
	}
	isServer(VOV_SERVER_EXIT);
#endif
    }
    return VOV_OK ;
}

VovStatus VovTrace::killSlave( int index = -1 )
{
    Bzocket& zoc = getBzocket();
    
    if ( isClient() ) {
	zoc.pack( VOV_KillSlave );
	zoc.pack( index );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    index = zoc.unpackInt();
	}

	vovMessage("Kill slave not yet implemented" );

	isServer(VOV_SERVER_EXIT);
#endif
    }
    return VOV_OK ;
}


#ifndef VOVCLIENT
VovStatus VovTrace::dispatchTransition( Client& slave, VovTransition& trans )
{
    VovContainersGenerator genInputs( &trans, VOV_PLACE_OCTMASK );
    VovPlace place;
    int changed = 0;
    while ( octGenerate( &genInputs, &place ) == OCT_OK ) {
	VOV_ASSERT( place.get() );
	if ( place.changeTransition( trans, 1 ) == VOV_CHANGE ) {
	    vovMessage( "Temporary fix for OPTION handling" );
	    changed = 1;
	}
    }

    VOV_ASSERT( trans.initiate( VOV_RETRACING_NODE ) );
    trans.setPid( slave.pid );

    //
    // Send to the slave the transition.
    // 
    vovMessage( "Sending " + trans.getCommandLine() + " to " + slave.format() );
    slave.pack( VOV_Exec ); 		// Code.
    trans.pack( slave );		// Transition.
    slave.pack( retraceFastMode );	// Speed.
    slave.flush();			// 
    return VOV_OK;
}
#endif
