// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of GNU CC.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the GNU CC General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
GNU CC, but only under the conditions described in the
GNU CC General Public License.   A copy of this license is
supposed to have been given to you along with GNU CC so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies.  
*/

#include "dateVHMap.h"

/* codes for status fields */

#define EMPTYCELL   0
#define VALIDCELL   1
#define DELETEDCELL 2


StringintVHMap::StringintVHMap(int  dflt, int sz = DEFAULT_INITIAL_CAPACITY)
     :(dflt)
{
  tab = new String[size = sz];
  cont = new int[size];
  status = new char[size];
  for (int i = 0; i < size; ++i) status[i] = EMPTYCELL;
}

StringintVHMap::StringintVHMap(StringintVHMap& a) : (a.def)
{
  tab = new String[size = a.size];
  cont = new int[size];
  status = new char[size];
  for (int i = 0; i < size; ++i) status[i] = EMPTYCELL;
  count = 0;
  for (Pix p = a.first(); p; a.next(p)) (*this)[a.key(p)] = a.contents(p);
}


/* 
 * hashing method: double hash based on high bits of hash fct,
 * followed by linear probe. Can't do too much better if table
 * sizes not constrained to be prime.
*/


static inline doublehashinc(unsigned int h, int s)
{
  return ((h / s) % s) >? 1;
}

Pix StringintVHMap::seek(String& key)
{
  unsigned int hashval = StringHASH(key);
  int h = hashval % size;
  for (int i = 0; i <= size; ++i)
  {
    if (status[h] == EMPTYCELL)
      return 0;
    else if (status[h] == VALIDCELL && StringEQ(key, tab[h]))
      return Pix(&tab[h]);
    if (i == 0)
      h = (h + doublehashinc(hashval, size)) % size;
    else if (++h >= size)
      h -= size;
  }
  return 0;
}


int& StringintVHMap::operator [](String& item)
{
  if (size <= count + 1)
    resize();

  int bestspot = -1;
  unsigned int hashval = StringHASH(item);
  int h = hashval % size;
  for (int i = 0; i <= size; ++i)
  {
    if (status[h] == EMPTYCELL)
    {
      ++count;
      if (bestspot < 0) bestspot = h;
      tab[bestspot] = item;
      status[bestspot] = VALIDCELL;
      return cont[bestspot] = def;
    }
    else if (status[h] == DELETEDCELL)
    {
      if (bestspot < 0) bestspot = h;
    }
    else if (StringEQ(tab[h],item))
      return cont[h];

    if (i == 0)
      h = (h + doublehashinc(hashval, size)) % size;
    else if (++h >= size)
      h -= size;
  }

  ++count;
  status[bestspot] = VALIDCELL;
  tab[bestspot] = item;
  return cont[bestspot] = def;
}


void StringintVHMap::del(String& key)
{
  unsigned int hashval = StringHASH(key);
  int h = hashval % size;
  for (int i = 0; i <= size; ++i)
  {
    if (status[h] == EMPTYCELL)
      return;
    else if (status[h] == VALIDCELL && StringEQ(key, tab[h]))
    {
      status[h] = DELETEDCELL;
      --count;
      return;
    }
    if (i == 0)
      h = (h + doublehashinc(hashval, size)) % size;
    else if (++h >= size)
      h -= size;
  }
}


void StringintVHMap::clear()
{
  for (int i = 0; i < size; ++i) status[i] = EMPTYCELL;
  count = 0;
}

void StringintVHMap::resize(int newsize = 0)
{
  if (newsize <= count)
  {
    newsize = DEFAULT_INITIAL_CAPACITY;
    while (newsize <= count) newsize <<= 1;
  }
  String* oldtab = tab;
  int* oldcont = cont;
  char* oldstatus = status;
  int oldsize = size;
  tab = new String[size = newsize];
  cont = new int[size];
  status = new char[size];
  for (int i = 0; i < size; ++i) status[i] = EMPTYCELL;
  count = 0;
  for (i = 0; i < oldsize; ++i) 
    if (oldstatus[i] == VALIDCELL) 
      (*this)[oldtab[i]] = oldcont[i];
  delete [oldsize] oldtab;
  delete [oldsize] oldcont;
  delete oldstatus;
}

Pix StringintVHMap::first()
{
  for (int pos = 0; pos < size; ++pos)
    if (status[pos] == VALIDCELL) return Pix(&tab[pos]);
  return 0;
}

void StringintVHMap::next(Pix& i)
{
  if (i == 0) return;
  int pos = ((unsigned)i - (unsigned)tab) / sizeof(String) + 1;
  for (; pos < size; ++pos)
    if (status[pos] == VALIDCELL)
    {
      i = Pix(&tab[pos]);
      return;
    }
  i = 0;
}
  

int StringintVHMap::OK()
{
  int v = tab != 0;
  v &= status != 0;
  int n = 0;
  for (int i = 0; i < size; ++i) 
  {
    if (status[i] == VALIDCELL) ++n;
    else if (status[i] != DELETEDCELL && status[i] != EMPTYCELL)
      v = 0;
  }
  v &= n == count;
  if (!v) error("invariant failure");
  return v;
}
