#include "port_internal.h"
#include "intmask.h"
#include "utcc.h"

Intmask::Intmask( int length  = INT_SIZE )
{
    if ( length < 1 ) {
	utcc_fatalerror( "length cannot be < 1" );
    }
    size = ( length + INT_SIZE - 1 ) / INT_SIZE ;
    maskptr = new int[size];
    clear();
}

Intmask::Intmask( Intmask& m )
{
    size = m.size;
    maskptr = new int[size];
    for ( int b = 0 ; b < size ; b++ ) maskptr[b] = m.maskptr[b];
    firstbit = m.firstbit;
    lastbit = m.lastbit;
    count = m.count;
}

Intmask::~Intmask()
{
    delete maskptr;
}

void Intmask::resize( int length ) 
{
    int newsize = ( length + INT_SIZE - 1 ) / INT_SIZE ;
    if ( newsize > size ) {
	int* tmp = new int[newsize];
	for ( int b = 0 ; b < size ; b++ ) tmp[b] = maskptr[b];
	delete maskptr;
	maskptr = tmp;
	size = newsize;
    }
}

void Intmask::print()
{
    cerr << "\nMask ";
    for ( int b = 0 ; b < size ; b++ ) cout.form( "%08x" , maskptr[b] );
    cerr.form(" Count %d First %d Last %d\n", count, firstbit, lastbit);
}

void Intmask::error( char* s )
{
    cerr.form( "Intmask:ERROR:%s\n", s );
    print();
}

void  Intmask::add(int i)
{
    if ( test( i ) ) {
	error(" bit already set ");
    } else {
	int b = i / INT_SIZE;			// 
	int ii = i % INT_SIZE;			// 
	maskptr[b] |= ( 1 << ii );		// Set the bit.
	if ( count++ == 0 ) {			// Is this the first bit set ?
	    firstbit = lastbit = i;		// 
	} else {				// 
	    if ( i < firstbit ) firstbit = i;	// Adjust first and last bit indices.
	    if ( i > lastbit )  lastbit = i ;	// 
	}					// 
    }						// 
}

void  Intmask::del(int i)
{
    if ( test(i) ) {				// 
	int b = i / INT_SIZE;			// 
	int ii = i % INT_SIZE;			// 
	maskptr[b] &= ~( 1 << ii );		// Set the bit.
	if (--count) {
	    if ( i == firstbit ) {
		while ( b < size && maskptr[b] == 0 ) b++;
		firstbit =  b * INT_SIZE + ffs(maskptr[b]) - 1;
	    }
	    if (i == lastbit) {
		while ( b >= 0 && maskptr[b] == 0 ) b--;
		ii = INT_SIZE - 1;
		int ml = ( 1 << ii ); 		// mask with the last bit.
		while ( ! ( maskptr[b] & ml ) ) {
		    ml >>= 1; 
		    ii--;
		}
		lastbit = b * INT_SIZE + ii;
	    }
	} else {		/* mask == 0 */
	    firstbit = 0;
	    lastbit = 0;
	}
    } else {
	error("deleting: bit already zero");
    }
}

void Intmask::clear()
{
    for( int i = 0; i< size ; i++ ) maskptr[i] = 0;
    count = 0;
    firstbit = 0;
    lastbit = 0;
}


/* 
 * Somebody else has changed maskptr but all the rest might be wrong. Fix this.
 */
void Intmask::fix()
{
    count = 0;
    for ( int b = 0 ; b < size ; b++ ) {
	if ( maskptr[b] != 0 ) {
	    for ( int ii = 0 ; ii < INT_SIZE ; ii++ ) {	// Test all bits in theis word.
		if ( maskptr[b] & ( 1 << ii ) ) {	// This bit is set.
		    lastbit = b * INT_SIZE + ii;	// Update lastbit.
		    if ( count == 0 ) firstbit = b * INT_SIZE + ii;
		    count++;				// 
		}
	    }
	}
    }
    if ( ! count ) {
	firstbit = lastbit = 0;
    }
}

/*
 * Make sure that everything is fine with the Intmask.
 * This is just a debugging routine.
 */
void Intmask::OK()
{
    int newcount = 0;
    for (int i = 0; i < size * INT_SIZE; i++) {
	if ( test( i ) ) {
	    newcount++;
	    if ( newcount == 1) {
		if (firstbit != i) {
		    print();
		    utcc_error("firstbit is not correct ");
		}
	    } 
	    if (newcount == count) {
		if (lastbit != i) {
		    print();
		    utcc_error( "lastbit is not correct " );
		}
	    }
	}
    }
    if ( newcount != count) {
	print();
	utcc_error( form("count is not correct: is %d , should be %d", count , newcount ) );
	count = newcount;
    }
}




RoundRobinSelector::RoundRobinSelector()
{
    rr = 0;
}

RoundRobinSelector::~RoundRobinSelector()
{
}

void RoundRobinSelector::print()
{
    cout << "\nMask ";
    for ( int b = 0 ; b < size ; b++ ) cout.form( "%08x" , maskptr[b] );
    cout.form(" Count %d RR %d First %d Last %d\n",count,rr,firstbit,lastbit);
}

void  RoundRobinSelector::add(int i)
{
    Intmask::add(i);
    if (count == 1) rr = firstbit;
}

void  RoundRobinSelector::del(int i)
{
    Intmask::del(i);
    if (count) {
	if (rr == i) next();
    } else {			
	rr = 0;
    }
}

int  RoundRobinSelector::index()
{
    if (count == 0) {
	utcc_error (" cannot call rr::index if count is 0 ");
    } 
    if ( test( rr ) == 0 ) {
	utcc_error(" rr should always be correct ");
    }
    return rr;
}

int  RoundRobinSelector::next()
{
    if (count == 0) {
	error ("Calling rr:next with size 0");
    }
    if (rr >= lastbit) {
	rr = firstbit;
    } else {
	while ( rr <= lastbit &&  test( rr ) == 0 ) rr++;
    }
    return index();
}

void RoundRobinSelector::clear()
{
    Intmask::clear();
    rr = 0;
}

void RoundRobinSelector::OK()
{
    Intmask::OK();
}


// Tests.
test_intmask()
{
    Intmask tt;
    tt.clear(); tt.print();
    tt.OK();
    tt.OK();
    tt.add(1); tt.print();
    tt.OK();
    tt.add(2); tt.print();
    tt.OK();
    tt.add(5); tt.print();
    tt.OK();
    tt.del(2); tt.print();
    tt.OK();
    tt.add(6); tt.print();
    tt.OK();
    tt.add(7); tt.print();
    tt.OK();
    tt.OK();
    tt.OK();
    tt.OK();
    tt.del(1); tt.print();
    tt.OK();
    tt.OK();
    tt.OK();
    tt.add(17); tt.print();
    tt.OK();
    tt.OK();
    tt.del(6); tt.print();
    tt.OK();
    tt.del(7); tt.print();
    tt.OK();
    tt.OK();
    tt.del(5); tt.print();
    tt.OK();
    tt.OK();
    tt.del(17); tt.print();
    tt.OK();
}
test_rrobin()

{
    RoundRobinSelector tt;
    tt.clear(); tt.print();
    tt.OK();
    tt.OK();
    tt.add(1); tt.print();
    tt.index(); tt.print();
    tt.OK();
    tt.add(2); tt.print();
    tt.index(); tt.print();
    tt.OK();
    tt.add(5); tt.print();
    tt.index(); tt.print();
    tt.OK();
    tt.del(2); tt.print();
    tt.index(); tt.print();
    tt.OK();
    tt.add(6); tt.print();
    tt.index(); tt.print();
    tt.OK();
    tt.add(7); tt.print();
    tt.OK();
    tt.index(); tt.print();
    tt.OK();
    tt.next(); tt.print();
    tt.OK();
    tt.next(); tt.print();
    tt.OK();
    tt.del(1); tt.print();
    tt.OK();
    tt.index(); tt.print();
    tt.OK();
    tt.index(); tt.print();
    tt.OK();
    tt.add(17); tt.print();
    tt.OK();
    tt.next(); tt.print();
    tt.OK();
    tt.del(6); tt.print();
    tt.OK();
    tt.del(7); tt.print();
    tt.OK();
    tt.index(); tt.print();
    tt.OK();
    tt.del(5); tt.print();
    tt.OK();
    tt.index(); tt.print();
    tt.OK();
    tt.del(17); tt.print();
    tt.OK();
}

