/* -*- C++ -*- */
#include "naming.convention"
/*
 * node.h
 *
 * This file describes the data structures to represent the design node
 */
#ifndef NODE_H
#define NODE_H
#pragma once

#include "vovinternal.h"
#include "base.h"
#include "select.h"

extern class Bzocket;
extern class VovTrace;		/* forward reference */
extern class VovPage;		/* forward reference */
extern class VovNode;		/* forward reference */

extern class VovPlacementMatrix;

#define VOV_NODE_OCTMASK      OCT_INSTANCE_MASK
#define VOV_NODE_SET_OCTMASK  OCT_BAG_MASK 

typedef enum _VovSetOperation {
    VOV_SET_ADD, 
    VOV_SET_DEL, 
    VOV_SET_CONTAINS,
    VOV_SET_RECOMPUTE,
    VOV_SET_UNION,
    VOV_SET_INTERSECTION,
    VOV_SET_COMPLEMENT
  } VovSetOperation;
		

/*
 * Node sets are actually bags.
 */
class VovNodeSet : public VovBag {
    VovProp owner;		// 
    VovProp created;		// Date.
    VovSelectRule selectRule;	// 
public:
    VovNodeSet();
    VovNodeSet( String& name );
    VovNodeSet( String& name, VovSelectRule& rule);
    VovNodeSet( VovNodeSet& );
    ~VovNodeSet();
    void operator=(VovNodeSet&);
    VovStatus create( );
    VovStatus get();
    VovStatus modify();
    VovStatus del();

    String getName() { return contents.bag.name; }
    VovSelectRule& getRule() { return selectRule; }

    int select( VovNodeSet&  );
    int subselect( VovSelectRule& rule );

    VovStatus visit( VovNode& node, VovStatusFunc visitFunction, void*  );
    VovStatus cover( VovStatusFunc visitFunction, void*  );

    String format( );
    VovStatus edit();
    void setupDialog( VovDialog& top_dialog, VovDialogHandles& data );
    VovStatus readDialog( VovDialogHandles& data );

    void pack( Bzocket& );
    void unpack( Bzocket& );
    
};

#define LABEL_REGION_HEIGHT 140
#define LABEL_TEXT_HEIGHT   70
#define LABEL_OFFSET        40
#define LABEL_COLOR	    "COM2"

class VovNode : public VovObject {
    VovObject   label;		// A documentation label for the node.
    VovProp	level;		// An integer.
    VovProp	status;		// A vov status (integer).
    VovProp     node_type;	// Another integer.
    VovProp	searchId;	// Used by the graph traveral algorithms.
 public:
    VovNode();
    VovNode( octId i );
    VovNode( char* cell, VovType nt );
    VovNode( VovNode& );
    ~VovNode();
    void operator=( VovNode& );

    char*    getViewName();

    virtual void setName( String& );
    virtual String getName();

    VovStatus create( );
    VovStatus get( );
    VovStatus modify();
    VovStatus del();

    virtual String format() ;
    VovType nodeType() { return node_type.integer(); }

    VovStatus bfs(int, 
		  VovStatusFunc, VovStatusFunc,
		  VovStatusFunc, VovStatusFunc,
		  void*);
    VovStatus forallInputs( VovStatusFunc , void* ); 	// 
    VovStatus forallOutputs( VovStatusFunc, void* );	// 
    VovStatus foreachInput( VovStatusFunc, void* );	// 
    VovStatus foreachOutput( VovStatusFunc, void* );	// 
    VovNode* existsInput( VovStatusFunc, void* );	// 
    VovNode* existsOutput( VovStatusFunc, void* );	// 

    VovStatus connectInput(VovNode& );
    VovStatus connectOutput(VovNode& );
    VovStatus disconnectInput( VovNode& );
    VovStatus disconnectOutput( VovNode& );
    void disconnect( int mask = VOV_IOS_MASK); 		// Eliminate links to this node.
							// 
    int  numberOfInputs();				// 
    int  numberOfOutputs();				// 
    VovStatus   hasInputs();				// Faster than testing numberOfInputs() != 0
    VovStatus   hasOutputs();				// Faster than testing numberOfOutputs() != 0
    VovStatus   isInput( VovNode& );			// 
    VovStatus   isOutput( VovNode& );			// 
							// 
    inline VovNodeStatus getStatus(){return status.integer();}	// 
    String getStatusString();				// 

    VovStatus changeStatus( VovNodeStatus s, int flags, String& msg );
    void setStatus(VovNodeStatus s);			// 
							// 
    int  newVisit();					// 
    void markVisited();					// 
    VovStatus alreadyVisited();				// 

    void levellize( int );				// 
    int  getLevel() { return level.integer();}		// 
    void setLevel(int l);				// 
							// 
    void doPlacement();					// 
    void move( octPoint p );				// 
    void updateLabel();					// 
    octPoint bestPlace();				// 
    VovPlacementMatrix* getOrAddPlacementMatrix(VovPlacementMatrix* m);		// 
    
    void pack( Bzocket& );				//
    void unpack( Bzocket& );				// 
    void packInputs( Bzocket& );
    void packOutputs( Bzocket& );
};

#endif	/* NODE_H */


